/* oqtbroadcaster.c
 *
 * Part of OpenQuicktime Broadcaster
 * Copyright (C) 2002 University of Southampton
 *
 * Makes use of OpenQuicktime, a free QuickTime library.
 * Based (very loosly) on rtp.c from the LAME 3.70 sources.
 *
 * OQTBroadcaster is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OQTBroadcaster is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Nicholas Humfrey <njh@ecs.soton.ac.uk>
 * http://www.openquicktime.org/broadcaster/
 *
 * Looks nice in BBEdit with Tab Spacing = 4
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"

#ifdef USESHOUT
#include <shout/shout.h>
#endif

#include "oqtbroadcaster.h"



#define PROGRAM_NAME      	"oqtbroadcaster"
#define OQTB_VERSION		"0.2"
#define MAX_CODEC_PARAMS   	64
#define MAX_STRLEN			512
#define PSEUDO_FILE_LEN		8192

#define QT_HEADER_VERS		1
// 'soun'
#define SOUND_MEDIA_TYPE	(0x736F756E)
// 'sd'
#define SOUND_DESCRIPTION	(0x7364)
// 'ti'
#define TRACK_DESCRIPTION	(0x7469)



// File I/O
FILE *infile = NULL;
FILE *outfile = NULL;

// RTP Variables
struct sockaddr_in rtp_si;
int rtp_seq_num=0;
int rtp_timestamp=0;
int rtp_ts_diff=1;
int rtp_head_len=0;
oqtb_head_t *rtp_head=NULL;
char* rtp_packet_buf=NULL;
int rtp_packets_per_desc=15;

// Globals structure
struct global_settings gf;
struct codec_param *param_list=NULL;
int param_count=0;
int track_num=-1;
volatile sig_atomic_t keep_going = 1; /* controls program termination */

#ifdef USESHOUT
shout_conn_t shout;
#endif

// Command line arguments
int opt_verbose = 0;
int opt_background = 0;
char* opt_conffile = NULL;
char* opt_infile = NULL;
char* opt_outfile = NULL;

// Audio caculations
int bytes_per_frame=0;
int frames_per_packet=0;
int samples_per_packet=0;
int bytes_per_packet=0;
int sample_byte_size=0;
float packet_duration=0.0;

// Audio encoding buffers
int out_buff_size;
oqt_uint8_t *in_buff;
oqt_uint8_t *out_buff;




static void
init_settings(void)
{
	gf.verbose_on = 0;
	gf.background_on = 0;
	gf.libshout_on = 0;
	gf.rtp_on = 1;
	
	gf.audio_source = strdup("/dev/dsp");
	gf.record_file = NULL;
	gf.audio_samplerate = 44100;
	gf.audio_bits = 16;
	gf.audio_channels = 2;

	gf.rtp_ip = strdup("127.0.0.1");
	gf.rtp_port = 41287;
	gf.rtp_ttl = 1;
	gf.rtp_stream_id = 0x01;
	gf.rtp_samp_desc_int = 1000;
	gf.rtp_packet_size_limit = 1450;
	gf.rtp_packet_dur_limit = 200;
	
	gf.codec[0] = '.';
	gf.codec[1] = 'm';
	gf.codec[2] = 'p';
	gf.codec[3] = '3';
	
#ifdef USESHOUT
	// Libshout settings...
	shout_init_connection(&shout);
	shout.ip = strdup("127.0.0.1");
	shout.port = 8000;
	//shout.mount = NULL;
	shout.password = strdup("letmein");
	shout.icy_compat = 1;
	shout.name = strdup("OQTBroadcaster Stream");
	shout.genre = strdup("None");
	shout.description = strdup("Unconfigured stream using OQTBroadcaster");
	shout.url = strdup("http://www.openquicktime.org/broadcaster/");
#endif
	
	
	// Attempt to generate random numbers
	srand(getpid() ^ time(0));
}




static void
init_rtp_packet( oqt_t *file, char * snddesc_ptr, int snddesc_len )
{
	// Initalise the sequence number and timestamp
	rtp_seq_num= rand() & 65535;
	rtp_timestamp = rand();
	rtp_ts_diff = file->atracks[track_num].track->mdia.minf.stbl.stsd.table[0].samplesPerPacket
					 * frames_per_packet;

	// Calculate total length of full packet
	rtp_head_len = sizeof(oqtb_head_t)-sizeof(unsigned long) + snddesc_len;
	rtp_head = malloc(rtp_head_len);
	
	// The standard RTP header
	rtp_head->r.v=RTP_VERSION;
	rtp_head->r.p=0;
	rtp_head->r.x=0;
	rtp_head->r.cc=0;
	rtp_head->r.m=1;
	rtp_head->r.pt=96;		// Dynamic
	rtp_head->r.ssrc=0;
	
	// The short QT header
	rtp_head->q.ver=QT_HEADER_VERS;
	rtp_head->q.pck=1;		// packetization scheme
	rtp_head->q.s=1;		// sync sample
	rtp_head->q.q=0;		// payload desc included 
	rtp_head->q.l=0;		// sample specific desc included
	rtp_head->q.id=htons(gf.rtp_stream_id);	// Payload ID (unique server RTP stream ID)
	
	rtp_head->q.res=0;		// Reserved bits
	//rtp_head->q.res2=0;	// Reserved bits
	//rtp_head->q.res3=0;	// Reserved bits
	
	//rtp_head->q.k=1;		// samples are sync
	//rtp_head->q.f=1;		// samples are sparse
	//rtp_head->q.a=1;		// contains payload desc
	//rtp_head->q.z=1;		// end of payload desc
	rtp_head->q.kfaz=htons(0xFFFB); // All of the above
	
	rtp_head->q.dlen=0;	// Length of Payload description
	
	
	// Extended header (with payload information)
	rtp_head->q.mtype=htonl(SOUND_MEDIA_TYPE);	// Media Type (Sound)
	rtp_head->q.timescale=htonl(oqt_get_audio_samplerate(file,track_num));	// Timescale (Fixed Decimal)
	
	
	// 'ti' TLV - The track number
	rtp_head->q.ti_len = htons(sizeof(rtp_head->q.ti_value));
	rtp_head->q.ti_type = htons(TRACK_DESCRIPTION);
	rtp_head->q.ti_value = htonl(1); // Fixed at 1 at the moment

	// 'sd' TLV - The Sound Format Description
	rtp_head->q.sd_len = htons(snddesc_len);
	rtp_head->q.sd_type = htons(SOUND_DESCRIPTION);
	
	// Copy the SoundDesc accross to the end of header
	memmove(&rtp_head->q.sd_value, snddesc_ptr, snddesc_len);
	
}

static
int send_rtp_packet(oqt_t *file, int rtp_socket, char* audio_buffer, int audio_len) {
	int packet_len, head_len;
	int channels = oqt_get_audio_channels(file, track_num);
	
	// Set values
	rtp_head->r.ts=htonl(rtp_timestamp+=rtp_ts_diff);
	rtp_head->r.seq=htons(++rtp_seq_num);
	
	// Send the full header every X packets
	if ((rtp_seq_num%rtp_packets_per_desc)==0) {
		head_len = rtp_head_len;
		packet_len = audio_len+head_len;
		rtp_head->q.dlen = htons(head_len-16);
		rtp_head->q.q = 1;
		rtp_head->q.kfaz = htons(0xB000);
		if (gf.verbose_on && !gf.background_on) fprintf(stderr, "*");
	} else {
		// Just send a short header...
		head_len = sizeof(rtp_hdr_t) + 4;
		packet_len = audio_len+head_len;
		rtp_head->q.dlen = htons(0);
		rtp_head->q.q = 0;
		rtp_head->q.kfaz = htons(0xFFFB);
	}
	
	// Allocate enough memory for the complete packet
	// and copy the header and data into it
	rtp_packet_buf = realloc(rtp_packet_buf, packet_len);
	memmove(rtp_packet_buf, rtp_head, head_len);
	memmove(rtp_packet_buf+head_len, audio_buffer, audio_len);
	
	return sendto(rtp_socket, rtp_packet_buf, packet_len, 0, (struct sockaddr *)&rtp_si, sizeof(rtp_si));
}


// create a sender socket.
static int
make_rtp_socket(char *szAddr, unsigned short port, int TTL, struct sockaddr_in *sSockAddr)
{
  int          iRet, iLoop = 1;
  struct       sockaddr_in sin;
  char         cTtl = (char)TTL;
  char         cLoop=0;
  unsigned int tempaddr;

  int iSocket = socket( AF_INET, SOCK_DGRAM, 0 );
  if (iSocket < 0) {
    fprintf(stderr,"socket() failed.\n");
    exit(1);
  }

  tempaddr=inet_addr(szAddr);
  sSockAddr->sin_family = sin.sin_family = AF_INET;
  sSockAddr->sin_port = sin.sin_port = htons(port);
  sSockAddr->sin_addr.s_addr = tempaddr;

  iRet = setsockopt(iSocket, SOL_SOCKET, SO_REUSEADDR, &iLoop, sizeof(int));
  if (iRet < 0) {
    fprintf(stderr,"setsockopt SO_REUSEADDR failed\n");
    exit(1);
  }

  if ((ntohl(tempaddr) >> 28) == 0xe) {
    // only set multicast parameters for multicast destination IPs
    iRet = setsockopt(iSocket, IPPROTO_IP, IP_MULTICAST_TTL, &cTtl, sizeof(char));
    if (iRet < 0) {
      fprintf(stderr,"setsockopt IP_MULTICAST_TTL failed.  multicast in kernel?\n");
      exit(1);
    }

    cLoop = 1;	// !?
    iRet = setsockopt(iSocket, IPPROTO_IP, IP_MULTICAST_LOOP,
		      &cLoop, sizeof(char));
    if (iRet < 0) {
      fprintf(stderr,"setsockopt IP_MULTICAST_LOOP failed.  multicast in kernel?\n");
      exit(1);
    }
  }

  return iSocket;
}

static void
calculate_audio_sizes(oqt_t *file)
{
	int frames_per_packet_size, frames_per_packet_dur, bpf;

	
	// Caculate the duration of a frame (ms)
	float frame_dur = 1000 / ( oqt_get_audio_samplerate(file,track_num) /
					file->atracks[track_num].track->mdia.minf.stbl.stsd.table[0].samplesPerPacket);

	// Number of bytes in a frame off audio
	bytes_per_frame = file->atracks[track_num].track->mdia.minf.stbl.stsd.table[0].bytesPerFrame;
	//bytes_per_frame = oqt_get_audio_frames_to_bytes(file,track_num, 1);

	// Calculate the number of frames per packet (based on size)
	frames_per_packet_size = gf.rtp_packet_size_limit/bytes_per_frame;
	
	// Calculate the number of frames per packet (based on duration)
	frames_per_packet_dur = gf.rtp_packet_dur_limit/frame_dur;
	
	// Use the smaller number
	if (frames_per_packet_size < frames_per_packet_dur)
			frames_per_packet = frames_per_packet_size;
	else	frames_per_packet = frames_per_packet_dur;
	
	// Can't have less than 1 frame per packet ;)
	if (frames_per_packet < 1) frames_per_packet = 1;

	// Duration and size of a packet
	packet_duration = frames_per_packet * frame_dur;
	bytes_per_packet = frames_per_packet * bytes_per_frame;
	
	// Calculate the number of samples per RTP packet
	samples_per_packet = (file->atracks[track_num].track->mdia.minf.stbl.stsd.table[0].samplesPerPacket)
						  * frames_per_packet * gf.audio_channels;
	sample_byte_size = (oqt_get_audio_bits(file,track_num)/8);
	
	// Calculate the number of packets per description
	rtp_packets_per_desc = (gf.rtp_samp_desc_int/frame_dur) / frames_per_packet;
	
	// apple default values:
	//gf.rtp_samp_desc_int = 1000;
	//gf.rtp_packet_size_limit = 1450;
	//gf.rtp_packet_dur_limit = 100;
	
	//fprintf(stderr, "Length of frame = %d bytes.\n", bytes_per_frame);
	//fprintf(stderr, "Duration of frame = %f ms.\n", frame_dur);
}




/*
*/

static int
check_boolean(char* bool) {
	if (!strcasecmp(bool, "on"))			return 1;
	else if (!strcasecmp(bool, "off"))		return 0;
	else if (!strcasecmp(bool, "true"))		return 1;
	else if (!strcasecmp(bool, "false"))	return 0;
	else if (!strcasecmp(bool, "yes"))		return 1;
	else if (!strcasecmp(bool, "no"))		return 0;
	else if (!strcasecmp(bool, "1"))		return 1;
	else if (!strcasecmp(bool, "0"))		return 0;

	fprintf(stderr, "Unknown boolean value: %s\n", bool);
	return 0;	
}

static void
process_param(char* name, char* value) {

	// General Settings
	if (!strcasecmp(name, "verbose"))			gf.verbose_on = check_boolean(value);
	else if (!strcasecmp(name, "shout")) {
		#ifdef USESHOUT
			gf.libshout_on = check_boolean(value);
		#else
			gf.libshout_on = 0;
			fprintf(stderr, "libshout support not compiled into %s.\n", PROGRAM_NAME);
		#endif
	}		
	else if (!strcasecmp(name, "rtp"))			gf.rtp_on = check_boolean(value);
	else if (!strcasecmp(name, "background"))	gf.background_on = check_boolean(value);
	else if (!strcasecmp(name, "record_file")) {
		if (gf.record_file) free(gf.record_file);
		gf.record_file = strdup(value);
	}
	
	// Audio Settings
	else if (!strcasecmp(name, "audio_source")) {
		if (gf.audio_source) free(gf.audio_source);
		gf.audio_source = strdup(value);
	}
	else if (!strcasecmp(name, "audio_samplerate"))
		gf.audio_samplerate = atoi(value);
	else if (!strcasecmp(name, "audio_bits"))
		gf.audio_bits = atoi(value);
	else if (!strcasecmp(name, "audio_channels"))
		gf.audio_channels = atoi(value);
	
	// RTP Settings
	else if (!strcasecmp(name, "rtp_ip")) {
		if (gf.rtp_ip) free(gf.rtp_ip);
		gf.rtp_ip = strdup(value);
	}
	else if (!strcasecmp(name, "rtp_port"))
		gf.rtp_port = atoi(value);
	else if (!strcasecmp(name, "rtp_ttl"))
		gf.rtp_ttl = atoi(value);
	else if (!strcasecmp(name, "rtp_samp_desc_int"))
		gf.rtp_samp_desc_int = atoi(value);
	else if (!strcasecmp(name, "rtp_packet_size_limit"))
		gf.rtp_packet_size_limit = atoi(value);
	else if (!strcasecmp(name, "rtp_packet_dur_limit"))
		gf.rtp_packet_dur_limit = atoi(value);
	else if (!strcasecmp(name, "rtp_stream_id"))
		gf.rtp_stream_id = atoi(value);
	
	// Codec settings
	else if (!strcasecmp(name, "codec")) {
		// Is it four characters long ?
		if (strlen(value)==4) {
			memcpy(gf.codec, value, 4);
		} else if (!strncasecmp(value, "0x", 2)) {
			unsigned int  num;
			sscanf(value, "0x%x", &num);
			#ifdef WORDS_BIGENDIAN
			gf.codec[0] = (num & 0x000000FF);
			gf.codec[1] = (num & 0x0000FF00) >> 8;
			gf.codec[2] = (num & 0x00FF0000) >> 16;
			gf.codec[3] = (num & 0xFF000000) >> 24;
			#else
			gf.codec[3] = (num & 0x000000FF);
			gf.codec[2] = (num & 0x0000FF00) >> 8;
			gf.codec[1] = (num & 0x00FF0000) >> 16;
			gf.codec[0] = (num & 0xFF000000) >> 24;
			#endif
		} else {
			fprintf(stderr, "Value of codec's signature is in unknown format (%s)\n", value);
		}
	}
	else if (!strcasecmp(name, "codec_param")) {
		char *param_value=NULL;
		int n;
		
		for(n=0; value[n]; ++n) {
			if (isspace(value[n])) {
				value[n] = 0;
				param_value = &value[n+1];
				break;
			}
		}
		
		if (param_value) {
			param_list = realloc(param_list,
						(++param_count*sizeof(struct codec_param)));
			param_list[param_count-1].name = strdup(value);
			param_list[param_count-1].value = strdup(param_value);
		} else {
			fprintf(stderr, "Failed to extract codec paramete name and value for %s", value);
		}
	}

#ifdef USESHOUT
	// Shoutcast settings
	else if (!strcasecmp(name, "shout_ip")) {
		if (shout.ip) free(shout.ip);
		shout.ip = strdup(value);
	}
	else if (!strcasecmp(name, "shout_port"))
		shout.port = atoi(value);
	else if (!strcasecmp(name, "shout_password")) {
		if (shout.password) free(shout.password);
		shout.password = strdup(value);
	}
	else if (!strcasecmp(name, "shout_icy_compat"))
		shout.icy_compat = check_boolean(value);
	else if (!strcasecmp(name, "shout_name")) {
		if (shout.name) free(shout.name);
		shout.name = strdup(value);
	}
	else if (!strcasecmp(name, "shout_genre")) {
		if (shout.genre) free(shout.genre);
		shout.genre = strdup(value);
	}
	else if (!strcasecmp(name, "shout_desc")) {
		if (shout.description) free(shout.description);
		shout.description = strdup(value);
	}
	else if (!strcasecmp(name, "shout_url")) {
		if (shout.url) free(shout.url);
		shout.url = strdup(value);
	}
#endif	
	
	else {
		fprintf(stderr, "Unknown configuration option %s.\n", name);
	}
	
	
}



// Returns 0 if successful
static int
parse_conf_file(char* filename) {
	char line[MAX_STRLEN];
	FILE* file;
	
	file = fopen(filename, "r");
	if (!file) return 1;
	
	//fprintf(stderr, "Loading settings from : %s\n", filename);

	while(fgets(line, MAX_STRLEN, file)) {
		char *name=NULL, *value=NULL;
		int n, name_end=0;
		
		// Skip the line if it start with a hash
		// or has little or no length
		if(line[0]=='#' || strlen(line)<=1) continue;
		
		// Find the name and value
		for(n=0;n<MAX_STRLEN; ++n) {
			if ( line[n] == '\n' ||
				 line[n] == '\r' )
			{ line[n] = 0;
			  break;
			} else if (!name) {
				if (!isspace(line[n])) name = &line[n];
			} else if (!name_end) {
				if (isspace(line[n])) {
					name_end = 1;
					line[n]=0;
				}
			} else if (!value) {
				if (!isspace(line[n])) value = &line[n];
			}
		}
		
		// Did we find a name and value ?
		if (name_end && name && value) 
			process_param(name, value);
	}
	
	fclose(file);
	
	return 0;
}


static void
display_settings(oqt_t *file)
{
	const oqt_codec_param_t* params;
	oqt_audio_codec_t *codec = file->atracks[0].codec;
	int i;
	
	// Display audio settings
	if (gf.verbose_on) {
		fprintf(stderr, "\nAudio Input and (Output) Settings:\n");
		fprintf(stderr, "   Source: %s", gf.audio_source);
		if (gf.record_file) fprintf(stderr," (%s)\n",gf.record_file);
		else fprintf(stderr, "\n");
		fprintf(stderr, "   Channels: %d (%d)\n", gf.audio_channels,
				oqt_get_audio_channels(file,track_num));
		fprintf(stderr, "   Sample Rate: %.3f kHz (%.3f kHz)\n", ((float)gf.audio_samplerate)/1000,
				((float)oqt_get_audio_samplerate(file,track_num))/1000);
		fprintf(stderr, "   Bits: %d (%d)\n", gf.audio_bits, oqt_get_audio_bits(file,track_num));
	}
	
	fprintf(stderr, "Using audio codec \"%s\" [%.4s]\n", codec->info.name, codec->info.fourcc);
	
	// Display the codec settings
	if (gf.verbose_on) {
		params = oqt_list_codec_params(&codec->info, &param_count);
		for(i=0;i<param_count;i++) {
			char data[OQT_PARAMETER_MAX_VALUE_LEN];
			int type = oqt_get_audio_param(file, track_num, params[i].name, &data);
			fprintf(stderr, "   %s: ", params[i].name);
			
			switch(type) {
				case OQT_PARAMETER_STRING: {
					fprintf(stderr, "%s", data);
				break;}
				case OQT_PARAMETER_BOOLEAN:
					if ((char)data[0])	fprintf(stderr, "True");
					else				fprintf(stderr, "False");
				break;
				case OQT_PARAMETER_INTEGER:
					fprintf(stderr, "%d", *((int*)&data));
				break;
				case OQT_PARAMETER_UNSIGNED_INTEGER:
					fprintf(stderr, "%d", *((unsigned int*)&data));
				break;
				default:
					fprintf(stderr, "Unknown Param Type");
				break;
			}
			
			fprintf(stderr, "\n");
		}
		
		// Display the networking settings
		fprintf(stderr, "Network Settings\n");
		
		#ifdef USESHOUT
			if (gf.libshout_on) fprintf(stderr, "  Libshout Server: %s:%d\n", shout.ip, shout.port);
		#endif
		
		if (gf.rtp_on) {
			fprintf(stderr, "   RTP Socket: %s:%d/%d\n", gf.rtp_ip, gf.rtp_port, gf.rtp_ttl);
			fprintf(stderr, "   Packets per Sample Description: %d\n", rtp_packets_per_desc);
			fprintf(stderr, "   Timestamp Difference: %d\n", rtp_ts_diff);
		}
	
		fprintf(stderr, "   Bytes Per Frame: %d.\n", bytes_per_frame);
		fprintf(stderr, "   Frames Per Packet: %d.\n", frames_per_packet);
		fprintf(stderr, "   Duration of packet: %f ms.\n", packet_duration);
		fprintf(stderr, "   Length of packet: %d bytes.\n\n", bytes_per_packet);
	}
}

static
void set_codec_options(oqt_t *file, int track)
{
	int n;
	for(n=0;n<param_count;n++) {
		char buff[OQT_PARAMETER_MAX_VALUE_LEN];
		int type = oqt_get_audio_param(file, track, param_list[n].name, &buff);
		int value_si;
		unsigned int value_ui;
		double value_d;
		
		switch(type) {
			case OQT_PARAMETER_INTEGER:
				value_si = atoi(param_list[n].value);
				oqt_set_audio_param(file, track, param_list[n].name, &value_si);
			break;
			case OQT_PARAMETER_UNSIGNED_INTEGER:
				value_ui = atol(param_list[n].value);
				oqt_set_audio_param(file, track, param_list[n].name, &value_ui);
			break;
			case OQT_PARAMETER_BOOLEAN:
				value_si = check_boolean(param_list[n].value);
				oqt_set_audio_param(file, track, param_list[n].name, &value_si);
			break;
			case OQT_PARAMETER_STRING:
				oqt_set_audio_param(file, track, param_list[n].name, param_list[n].value);
			break;
			case OQT_PARAMETER_DOUBLE:
				value_d = atof(param_list[n].value);
				oqt_set_audio_param(file, track, param_list[n].name, &value_d);
			break;
			case OQT_PARAMETER_UNKNOWN:
				fprintf(stderr, "Ignoring unknown codec parameter %s\n", param_list[n].name);
			break;
		}
	}
}



static void
termination_handler (int signum)
{
  keep_going = 0;
  switch(signum) {
  	case SIGTERM:	fprintf(stderr, "\nGot termination signal.\n"); break;
  	case SIGINT:	fprintf(stderr, "\nGot interupt signal.\n"); break;
  }
  signal(signum, termination_handler);
}


static int
encode_audio(oqt_t *file, oqt_uint8_t *buffer, int bufffer_size) {
	int samples, encoded_bytes;
	
	samples = fread(in_buff, sample_byte_size, samples_per_packet, infile);
	//fprintf(stderr, "Read in %d/%d samples.\n", samples, samples_per_packet);
	
	// Now encode it...
	encoded_bytes = file->atracks[track_num].codec->encode
	  (file, 
	   track_num,
	   samples/gf.audio_channels,	// Number of samples in input buffer
	   in_buff, 			// Input Buffer
	   bufffer_size,		// Size of output buffer
	   buffer);				// Output Buffer
	   
	//fprintf(stderr, "%d/%d bytes in output buffer.\n", encoded_bytes, out_buff_size);
	
	// Dump it to recording file ?
	if (outfile) {
		fwrite(buffer, encoded_bytes, 1, outfile);
	}

	return encoded_bytes;
}


static oqt_uint64_t
oqtb_get_utime(void) {
	struct timeval mtv;
	gettimeofday(&mtv, NULL);
	return (oqt_uint64_t)(mtv.tv_sec) * 1000 + (oqt_uint64_t)(mtv.tv_usec) / 1000;
}



// ****** THIS DOESN'T WORk !!! ********
static void oqtb_sleep(void)
{
	static oqt_uint64_t start_time=0;
	static oqt_uint64_t sent_time=0;
	//static last=0;
	int sleep;
	
	// Is this our first time ?
	if (start_time<=0) start_time = oqtb_get_utime();
	
	// How long ago did we start
	//ago = oqtb_get_utime() - start_time;
	
	// We just sent another packet...
	sent_time += packet_duration;
	sleep = sent_time-(oqtb_get_utime() - start_time);
	
	//fprintf(stderr, "\ns=%lld  ", sent_time);
	//fprintf(stderr, "\r l=%f ", ((float)sleep/1000));
	//fprintf(stderr, "d=%f  ", ((float)(sleep-last)/1000));
	
	// Are we late ?
	if (sleep<0)
		fprintf(stderr, "Error: encoder can't keep up. You machine is too slow !\n");
	else if (sleep>0) {
		// We are early... sleepy, sleepy for a bit...
		//fprintf(stderr,"!");
		usleep(sleep);
	}
	
	//last=sleep;
}



static void
start_streaming(oqt_t *file, int rtp_socket) {
	int child_process;
	char* buf= NULL;
	int buf_len=0, buf_used = 0;

	if( !gf.background_on || ((child_process=fork()) == 0) )    {
		// This is the child process that does the broadcasting...
		
		// Establish signal handler to clean up before termination:
		if (signal(SIGTERM, termination_handler) == SIG_IGN) signal(SIGTERM, SIG_IGN);
		if (signal(SIGINT, termination_handler) == SIG_IGN)  signal(SIGINT, SIG_IGN);
		signal(SIGHUP, SIG_IGN);
		
		while(keep_going && !feof(infile)) {
			// Do stuff
			buf = out_buff + buf_used;
			buf_len = out_buff_size - buf_used;
			//fprintf(stderr,"Free=%5.5d  Used=%5.5d", buf_len, buf_used);
			
			// Make sure there is room for a packet of audio
			if (buf_len < bytes_per_packet)
				fprintf(stderr,"start_streaming: warning no room in buffer for full packet of audio.\n");
		
			// Encode some (more?) audio
			buf_used += encode_audio(file, buf, buf_len);
			//fprintf(stderr,"   Encoded=%5.5d   Wanted=%5.5d", buf_used, bytes_per_packet);
			
			// If there is enough data to send...
			if (buf_used>=bytes_per_packet) {
				//fprintf(stderr, "Start = %x (%d)\n", *((oqt_uint32_t*)out_buff), bytes_per_packet);
				
				// Delay until it is time to send the next packet...
				oqtb_sleep();
				//shout_sleep( &shout );
			
				// Send to RTP multicast socket...
				if (gf.rtp_on) send_rtp_packet(file, rtp_socket, out_buff, bytes_per_packet);
				
				#ifdef USESHOUT
					// Send to libshout server
					if (gf.libshout_on) {
						int ret = shout_send_data(&shout, out_buff, bytes_per_packet);
						if (!ret) {
							fprintf(stderr,"libshout: Send error: %i...\n", shout.error);
							break;
						}
						
					}
				#endif
				
				// Move remaining audio data to start of buffer
				memmove(out_buff, (out_buff+bytes_per_packet), buf_used-bytes_per_packet);
				buf_used -= bytes_per_packet;
				
				   if (gf.verbose_on && !gf.background_on) fprintf(stderr,".");
			} else if (gf.verbose_on && !gf.background_on) fprintf(stderr,",");
		}
		
		
	} else {
		// This is the parent process...
		fprintf(stderr, "Started background process (%d)...\n",child_process);
		exit(0);
	}

	if (gf.verbose_on) fprintf(stderr, "\n");
}



static void show_usage(void) {
    fprintf(stderr,"usage: %s [-v] [-b] [-f <conffile>] [-i <infile>] [-o <outfile>]\n", PROGRAM_NAME);
    exit(1);
}


int
main(int argc, char **argv)
{
	oqt_t *file;
	int rtp_socket=0;
	
	fprintf(stderr, "OQTBroadcaster version %s (oqtlib v%s)\n", OQTB_VERSION, oqt_version());

	// Initalise global settings
	init_settings();
	
	
	// Parse the command line arguments
	while(1) {
		int c = -1;
		int option_index = 0;
		c = getopt(argc, argv, "vbf:i:o:h");
		
		if (c == -1)
		break;
		
		switch(c) {
			case 'f':
				opt_conffile = optarg;
			break;
			case 'v':
				opt_verbose = 1;
			break;
			case 'b':
				opt_background = 1;
			break;
			case 'i':
				opt_infile = optarg;
			break;
			case 'o':
				opt_outfile = optarg;
			break;
			case 'h':
			default:
				show_usage();
				exit(1);
			break;
		}
	}


	// If user passed configuration file, try and open it
	if (opt_conffile) {
		if (parse_conf_file(opt_conffile)) {
			fprintf(stderr, "Failed to open configuration file %s: ", opt_conffile);
			perror("");
			exit(1);
		}
	} else {
		parse_conf_file("/etc/oqtbroadcaster.conf");
		parse_conf_file("~/.oqtbroadcaster");
		parse_conf_file("./oqtbroadcaster.conf"); 
	}


	// Settings on command line override the conf files
	if (opt_verbose) gf.verbose_on = 1;
	if (opt_background) gf.background_on = 1;
	if (opt_infile) {
		if (gf.audio_source) free(gf.audio_source);
		gf.audio_source = opt_infile;
	}
	if (opt_outfile) {
		if (gf.record_file) free(gf.record_file);
		gf.record_file = opt_outfile;
	}
	
	
	// Open a pseudo oqt file (it is really a buffer in memory)
	file = oqt_open_memfile( NULL, PSEUDO_FILE_LEN );
	if (!file) {
		fprintf(stderr, "Failed to create pseudo OpenQuicktime file.\n");
		exit(1);
	}
	
	// Create an audio track
	track_num = oqt_add_audio_track(file, gf.audio_channels, gf.audio_samplerate, gf.audio_bits, gf.codec);
	if (track_num<0) {
		fprintf(stderr, "Failed to create audio track.\n");
		exit(1);
	}
	
	// Check the codec is supported..
	if (!oqt_supported_audio(file, track_num)) {
		fprintf(stderr, "Error: chosen codec (%.4s) is not supported.\n", gf.codec);
		exit(1);
	}
	
	// Configure the codec and prepare the encoder
	if(param_count && param_list) set_codec_options(file, track_num);
	file->atracks[track_num].codec->prepare_encoder(file, track_num);
	
	// Make sure it suppports SoundDesc v1
	if (file->atracks[track_num].track->mdia.minf.stbl.stsd.table[0].version != 1) {
		fprintf(stderr, "Error: error this program only works with codecs that support SoundDescription version 1.\n");
		exit(1);
	}
	
	// libshout only supports MP3... ?
	#ifdef USESHOUT
	if (gf.libshout_on) {
		if (!(gf.codec[0] == '.' && gf.codec[1] == 'm' &&
		      gf.codec[2] == 'p' && gf.codec[3] == '3' ) &&
		    !(gf.codec[0] == 'm' && gf.codec[1] == 's' &&
		      gf.codec[2] == 0x00 && gf.codec[3] == 0x55 ))
		 {
		     fprintf(stderr, "Disabling libshout because codec is not MPEG Audio.");
		     gf.libshout_on = 0;
		 }
	}
	#endif
	
	// Open the audio source
	if (strcmp("-", gf.audio_source)) {
		infile = fopen(gf.audio_source, "rb");
		if (!infile) {
			fprintf(stderr, "Failed to open audio source %s: ", gf.audio_source);
			perror("");
			exit(1);
		}
	} else {
		infile = stdin;
	}
	
	// Open the recording file
	if (gf.record_file) {
		if (strcmp("-", gf.record_file)) {
			outfile = fopen(gf.record_file, "wb");
			if (!outfile) {
				fprintf(stderr, "Failed to open recording file %s: ", gf.record_file);
				perror("");
				exit(1);
			}
		} else {
			outfile = stdout;
		}
	}
		
	
	// Calculate frames_per_packet and rtp_packets_per_des
	calculate_audio_sizes( file );

	// Connect to libshout server
	#ifdef USESHOUT
	if (gf.libshout_on) {
		int err;
		
		if ((err=shout_connect(&shout))) {
			if (gf.verbose_on) fprintf(stderr, "Connected to libshout server...\n");
		} else {
			char errstr[MAX_STRLEN];
			fprintf(stderr, "Failed to connect to libshout server (%s)\n", shout.ip);
			shout_strerror(&shout, err, errstr, MAX_STRLEN);
			fprintf(stderr, "Reason: %s\n", errstr);
			fprintf(stderr, "Turning off libshout.\n");
			gf.libshout_on=0;
		}
	}
	#endif
	
	// Get ready for RTP broadcast
	if (gf.rtp_on) {
		oqt_trak_t* trak = file->atracks[track_num].track;
		char* snddesc_ptr=NULL;
		int snddesc_len=0;
		
		// get pointer to start of SoundDescriptor
		snddesc_ptr = oqt_memfile_get_ptr(file);
		oqt_write_stsd_table(file, &(trak->mdia.minf), &(trak->mdia.minf.stbl.stsd.table[0]));
		snddesc_len = (int)file->oqt_flen(file);
		
		// Create an RTP multicast socket
		rtp_socket = make_rtp_socket(gf.rtp_ip, gf.rtp_port, gf.rtp_ttl, &rtp_si);
		if (!rtp_socket) {
			fprintf(stderr, "Failed to create a RTP socket.\n");
			exit(1);
		}
		
		// Initalise a 'generic' packet
		init_rtp_packet( file, snddesc_ptr, snddesc_len );
	}
	


	// Allocate input buffer
	in_buff = malloc(samples_per_packet * sample_byte_size);
	if (!in_buff) {
		fprintf(stderr, "Failed to allocate input buffer.\n");
		exit(1);
	}
	
	// Allocate output buffer
	//  - over twice as big as it needs to be so that we can store
	//    two packets of audio, if the first isn't quite big enough...
	out_buff_size = (bytes_per_packet*2) + 7200;
	out_buff = malloc(out_buff_size);
	if (!out_buff) {
		fprintf(stderr, "Failed to allocate output buffer.\n");
		exit(1);
	}
	
	// Display settings
	display_settings( file );
	
	
	// ------- This is where it all happens ! -------
	start_streaming( file, rtp_socket );
	// ------- This is where it all happens ! -------

	
	
	// Free up the buffers
	if (in_buff) free(in_buff);
	if (out_buff) free(out_buff);
	if (rtp_head) free(rtp_head);
	if (rtp_packet_buf) free(rtp_packet_buf);
	
	// Close the recording file
	if (outfile) fclose(outfile);
	
	// Close the input file
	fclose(infile);
	
	// Close connection to libshout server
	#ifdef USESHOUT
		if (gf.libshout_on) shout_disconnect(&shout);
	#endif
	
	// Close the pseudo OpenQuicktime file
	oqt_close(file);
	
	
	return 0;
}

