/* oqtencoder.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OQTEncoder, a free video encoder using the
 * OpenQuickTime library.
 *
 * OQTEncoder is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation.
 *
 * OQTEncoder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: oqtencoder.c,v 1.29 2003/03/15 04:30:57 jhatala Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>

#include <audiofile.h>

#include <openquicktime/openquicktime.h>
#include <openquicktime/colormodels.h>

#define DEFAULT_VIDEO_RATE 25
#define UNKNOWN_VIDEO_RATE -1
#define OPTION_NUMBER_MAX 100
#define NEW_VIDEO_TRACK	  -1
#define NEW_AUDIO_TRACK	  -1
#define NO_VIDEO_TRACK	  -2
#define NO_AUDIO_TRACK	  -2
#define BC_YUV4MPEG2 -4201
#define PROGRAM_NAME      "oqtencoder"

#define IMHERE() fprintf(stderr, "DEBUG: I'm at line %d\n", __LINE__)

typedef struct {
  char option[OQT_PARAMETER_MAX_VALUE_LEN];
  char value[OQT_PARAMETER_MAX_VALUE_LEN];
} codecoption_t;

static int keep_going = 1;
static oqt_t *out_file = NULL;

/* Video settings */
static codecoption_t videooptions[OPTION_NUMBER_MAX];
static int videooptionnumber = 0;
static char video_codec[5]    = { 'j','p','e','g', 0 };
static int video_track	= NEW_VIDEO_TRACK;
static float video_rate	= UNKNOWN_VIDEO_RATE;
static int video_width	= -1;
static int video_height	= -1;
static int video_cmodel = BC_RGB888;
static char *c_video_filename = NULL;
static char *pnm_input_dir = NULL;

/* Audio settings */
static codecoption_t audiooptions[OPTION_NUMBER_MAX];
static int audiooptionnumber = 0;
static char audio_codec[5]    = { 't','w','o','s', 0 };
static int audio_track = NO_AUDIO_TRACK;
static int audio_channels = 0;
static int audio_bits = 0;
static int audio_sample_size = 0;
static int audio_sample_rate;
static char *audio_filename = NULL;
static int audio_block_size = 1024;


static void usage(void)
{
  printf("\nUsage: %s [OPTION] <output quicktime file>\n", PROGRAM_NAME);
  printf("\nOPTIONS:\n");
  printf("  -d <dir>    Directory of PPM/PGM files for input\n");
  printf("  -V <codec>  Sets the video codec used (default [%c%c%c%c])\n", video_codec[0], video_codec[1], video_codec[2], video_codec[3]);
  printf("  -A <codec>  Sets the audio codec used (default [%c%c%c%c])\n", audio_codec[0], audio_codec[1], audio_codec[2], audio_codec[3]);
  printf("  -F <fps>    Specifies the framerate, (default %f),\n"
	 "              overrides framerate from YUV4MPEG and -c\n",
	 (double)DEFAULT_VIDEO_RATE);
  printf("  -v <o>=<v>  Specifies an option to pass to the video codec\n");
  printf("  -a <o>=<v>  Specifies an option to pass to the audio codec\n");
  printf("  -r <model>  Input color model: YUV420P (PGM), YUV420PACKED (PGM),\n"
	 "                                 YUV4MPEG2, RGB888 (PPM) (default)\n");
  printf("  -c <file>   Copy compressed video track from this qt movie instead.\n"
	 "              -r, -V, -v, -d and stdin will be ignored\n");
  printf("  -s <file>   Specify the audio file to use (default silent movie ;-)\n");
  printf("  -n          Do not create a video track: audio only\n");
  printf("  -H          Shows this help\n");
}

static void splitoption(char*string, char*option, char*value)
{
  int i = 0;

  if(!string || !option || !value)
    return;

  while(string[i] != '=' && i < strlen(string)) {
    i++;
  }

  if(string[i] =='=') {
    strncpy(option, string, i);
    strcpy(value, (&string[i+1]));
  }
}

typedef int
get_param_t(oqt_t *file, int track, const char* param, void* data);

typedef int
set_param_t(oqt_t *file, int track, const char* param, const void* data);

static void set_codec_options(get_param_t get_param,
			      set_param_t set_param,
			      codecoption_t *options,
			      int optionnumber)
{
  int i;
  for(i = 0; i < optionnumber; i++) {
    char buff[OQT_PARAMETER_MAX_VALUE_LEN];
    int type = get_param(out_file, 0, options[i].option, &buff);
    int value_si;
    unsigned int value_ui;
    double value_d;

    switch(type) {
    case OQT_PARAMETER_INTEGER:
      value_si = atoi(options[i].value);
      set_param(out_file, 0, options[i].option, &value_si);
      break;
    case OQT_PARAMETER_UNSIGNED_INTEGER:
      value_ui = atol(options[i].value);
      set_param(out_file, 0, options[i].option, &value_ui);
      break;
    case OQT_PARAMETER_BOOLEAN:
      value_si = 0;
      if(strcmp(options[i].value, "1") == 0 ||
	 strcmp(options[i].value, "TRUE") == 0 ||
	 strcmp(options[i].value, "true") == 0)
	value_si = 1;
      set_param(out_file, 0, options[i].option, &value_si);
      break;
    case OQT_PARAMETER_STRING:
      set_param(out_file, 0, options[i].option, options[i].value);
      break;
    case OQT_PARAMETER_DOUBLE:
      value_d = atof(options[i].value);
      set_param(out_file, 0, options[i].option, &value_d);
      break;
    case OQT_PARAMETER_UNKNOWN:
      fprintf(stderr, "Ignoring unknown parameter %s\n", options[i].option);
      break;
    }
  }
}


static void termination_handler(int signum)
{
  keep_going = 0;
  switch(signum) {
  case SIGTERM:	fprintf(stderr, "\nGot termination signal.\n"); break;
  case SIGINT:	fprintf(stderr, "\nGot interupt signal.\n"); break;
  }
  signal(signum, termination_handler);
}


/* midnight PPM/PGM parser helpers, see pgm(5) for format spec. */
static int is_white(char ch)
{
  return (ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r');
}

static void skip_to_eol(FILE *s)
{
  int ch;
  do {
    ch = fgetc(s);
  } while(ch != '\n' && ch != -1);
}

static int read_pnm_token(FILE *s, char *buf, int len)
{
  int ch = fgetc(s);
  if(ch == -1) return -1;

  while(is_white(ch)) {
    ch = fgetc(s);
    if(ch == -1) return -1;
  }

  while(ch == '#') {
    skip_to_eol(s);
    do {
      ch = fgetc(s);
      if(ch == -1) return -1;
    } while(is_white(ch));
  }

  do {
    *buf = ch;
    ++buf;
    --len;
    if(len <= 0) {
      *buf = 0;
      return 1;
    }

    ch = fgetc(s);
    if(ch == -1) return -1;

    if(ch == '#') {
      skip_to_eol(s);
      break;
    }
  } while(!is_white(ch));

  *buf = 0;
  return 0;
}


static int parse_pnm_header( FILE *input ) {
  static int frame_count = -1;
  char buffer[256];
  int res, max_value;

  /* Increment the frame counter */
  ++frame_count;

  /* read the ppm/pgm header.  We do not use fscanf here any more,
   * because of a subtle bug it introduced: the tricky fscanf
   * "P5\n\n%d %d\n255\n" would eat the first byte of frame data if
   * it was '\r'.  */

  /* P6 or P5 (for PPM or PGM respectively) */
  res = read_pnm_token(input, buffer, 255);
  if(res != 0) {
  parserr:
    fprintf(stderr, "Could not parse PPM/PGM header, frame #%d.\n", frame_count);
    return -1;
  }
  if(buffer[0] != 'P') goto parserr;
  if(video_cmodel == BC_RGB888 && buffer[1] != '6') {
    fprintf(stderr, "Expecting rawbits PPM in RGB mode\n");
    return -1;
  }
  if(video_cmodel != BC_RGB888 && buffer[1] != '5') {
    fprintf(stderr, "Expecting rawbits PGM in YUV mode\n");
    return -1;
  }
  if(buffer[2] != 0) goto parserr;


  /* width */
  res = read_pnm_token(input, buffer, 255);
  if(res != 0) goto parserr;
  res = sscanf(buffer, "%d", &video_width);
  if(res != 1) goto parserr;

  /* height */
  res = read_pnm_token(input, buffer, 255);
  if(res != 0) goto parserr;
  res = sscanf(buffer, "%d", &video_height);
  if(res != 1) goto parserr;

  /* max value, assumed to be 255 */
  res = read_pnm_token(input, buffer, 255);
  if(res != 0) goto parserr;
  res = sscanf(buffer, "%d", &max_value);
  if(res != 1 || max_value != 255) goto parserr;

  /* Adjust height value for composite */
  if(video_cmodel != BC_RGB888) {
    video_height = (video_height * 2) / 3;
  }

  return 0;
}



static int read_pnm_frame( FILE* file, oqt_uint8_t *framebuffer, oqt_uint8_t **rowbuffer, int byteframesize ) {
  int i;

  switch(video_cmodel) {
  case BC_YUV420P:
  case BC_RGB888:
    if(fread(framebuffer, 1, byteframesize, file) != byteframesize) {
      fprintf(stderr, "%s: Cannot read frame from input.\n", PROGRAM_NAME);
      return -1;
    }
    break;
  case BC_YUV420PACKED:
    fread (rowbuffer[0], video_width, video_height, file);
    for (i = 0; i < video_height >> 1; i++) {
      fread (rowbuffer[1]+i*video_width/2, video_width/2, 1, file);
      fread (rowbuffer[2]+i*video_width/2, video_width/2, 1, file);
    }
    break;
  default:
    fprintf(stderr,"%s: Don't know how to read frame for colour model %d.\n", PROGRAM_NAME, video_cmodel);
    break;
  }

  return 0;
}

static void create_video_track(void)
{
  /* Setting the video codec */
  if(video_rate == UNKNOWN_VIDEO_RATE)
    video_rate = DEFAULT_VIDEO_RATE;
  video_track = oqt_add_video_track(out_file, video_width, video_height, video_rate, video_codec);
  if(video_track<0) {
    fprintf(stderr,"%s: Cannot create video track with "
	    "compressor '%s' (%dx%d @ %.2f fps)\n",
	    PROGRAM_NAME, video_codec, video_width, video_height, video_rate);
    oqt_close(out_file);
    exit(1);
  }

  /* checks bellow are not needed for stream copy */
  if(c_video_filename)
    return;

  /* Is video codec supported ? */
  if (!oqt_supported_video(out_file, video_track)) {
    fprintf(stderr, "%s: Video codec '%.4s' is not supported.\n", PROGRAM_NAME, video_codec);
    oqt_close(out_file);
    exit(1);
  }

  /* Setting now the video codec option */
  set_codec_options(oqt_get_video_param, oqt_set_video_param,
		    videooptions, videooptionnumber);
}

static void create_audio_track(void)
{
  fprintf(stderr,"Creating audio track [%.4s] (%d channels, "
	  "%d bits, %d Hz, %d sample size)\n", audio_codec, audio_channels,
	  audio_bits, audio_sample_rate, audio_sample_size);

  audio_track = oqt_add_audio_track(out_file, audio_channels, audio_sample_rate, audio_bits, audio_codec);
  if(audio_track<0 || !oqt_supported_audio(out_file, audio_track)) {
    fprintf(stderr,"%s: Cannot create audio track with compressor '%s' "
	    "(%d channels, %d bits, %d Hz)\n", PROGRAM_NAME, audio_codec,
	    audio_channels, audio_bits, audio_sample_rate);
    oqt_close(out_file);
    exit(1);
  }

  /* Is audio codec supported ? */
  if (!oqt_supported_audio(out_file, audio_track)) {
    fprintf(stderr, "%s: Audio codec '%.4s' is not supported.\n", PROGRAM_NAME, audio_codec);
    oqt_close(out_file);
    exit(1);
  }

  set_codec_options(oqt_get_audio_param, oqt_set_audio_param,
		    audiooptions, audiooptionnumber);
}


/* Returns the length in bytes of a frame */
static int allocate_video_buffer( oqt_uint8_t **framebuffer, oqt_uint8_t ***rowbuffer )
{
  int framebuffer_size = video_width * video_height * 4;
  int rowbuffer_size = video_height * sizeof(oqt_uint8_t*);
  int byteframesize = -1, i = 0;

  /* We allocate here the frame buffer ... max size to support all the color space */
  (*framebuffer) = (oqt_uint8_t*)malloc(framebuffer_size);
  (*rowbuffer) = (oqt_uint8_t**)malloc(rowbuffer_size);
  if (!*framebuffer)	fprintf(stderr, "Failed to allocate %d byte buffer for framebuffer.\n", framebuffer_size);
  if (!*rowbuffer)	fprintf(stderr, "Failed to allocate %d byte buffer for rowbuffer.\n", rowbuffer_size);

  switch(video_cmodel) {
  case BC_YUV420P:
  case BC_YUV4MPEG2:
  case BC_YUV420PACKED:
    (*rowbuffer)[0] = (*framebuffer);
    (*rowbuffer)[1] = (*framebuffer) + (video_height * video_width);
    (*rowbuffer)[2] = (*framebuffer) + (video_height * video_width * 5 / 4);
    byteframesize = video_height * video_width * 3 / 2;
    break;
  case BC_RGB888:
    for(i = 0; i < video_height; i ++)
      (*rowbuffer)[i]=(*framebuffer)+(video_width*3*i);
    byteframesize = video_width * video_height * 3;
    break;
  default:
    fprintf(stderr,"%s: Don't know this color model %d\n", PROGRAM_NAME, video_cmodel);
    exit(1);
    break;
  }

  return byteframesize;
}



static char* parse_command_line( int argc, char** argv ) {
  int option_index = 0;
  int do_show_help = 0;
  int c = -1;

  while(1) {
    c = getopt(argc, argv, "d:A:V:s:bF:O:v:a:r:c:Hn");

    if (c == -1)
      break;

    switch(c) {
    case 'd':
      pnm_input_dir = optarg;
      break;
    case 'A':
      *(oqt_int32_t*)audio_codec = *(oqt_int32_t*)optarg;
      break;
    case 'V':
      *(oqt_int32_t*)video_codec = *(oqt_int32_t*)optarg;
      break;
    case 'F':
      video_rate = atof(optarg);
      break;
    case 'v':
      splitoption(optarg,
		  videooptions[videooptionnumber].option,
		  videooptions[videooptionnumber].value);
      videooptionnumber++;
      break;
    case 'a':
      splitoption(optarg,
		  audiooptions[audiooptionnumber].option,
		  audiooptions[audiooptionnumber].value);
      audiooptionnumber++;
      break;
    case 'r':
      if(strcasecmp(optarg, "YUV420P") == 0)
	video_cmodel = BC_YUV420P;
      else if(strcasecmp(optarg, "RGB888") == 0)
	video_cmodel = BC_RGB888;
      else if(strcasecmp(optarg, "YUV420PACKED") == 0)
	video_cmodel = BC_YUV420PACKED;
      else if(strcasecmp(optarg, "YUV4MPEG2") == 0 ||
	      strcasecmp(optarg, "Y4M") == 0)
	video_cmodel = BC_YUV4MPEG2;
      else {
	fprintf(stderr, "Unrecognised input color model: %s\n", optarg);
      }
      break;
    case 'c':
      c_video_filename = optarg;
      break;
    case 's':
      audio_filename = optarg;
      audio_track = NEW_AUDIO_TRACK;
      break;
    case 'n':
      video_track = NO_VIDEO_TRACK;
      break;
    case 'H':
      ++do_show_help;
      break;
    default:
      fprintf(stderr, "Unknown option specified %c\n", c);
      usage();
      exit(1);
      break;
    }
  }

  if(do_show_help) {
    usage();
    exit(0);
  }

  /* check that there is our quicktime output file as well  */
  if ((argc - optind) < 1) {
    usage();
    exit(1);
  }


  return argv[optind++];
}

/* Only look at 'files' ending .ppm or .pgm  */
static int scandir_select(struct dirent *entry) {
  char* suffix = (strlen(entry->d_name)-4) + entry->d_name;
  if (
#ifdef DT_REF
      (entry->d_type == DT_REG || entry->d_type == DT_UNKNOWN) &&
#endif
      (strcmp(suffix, ".ppm")==0 || strcmp(suffix, ".pgm")==0)
      ) {
    return 1;
  } else {
    /*fprintf(stderr, "moo: type=%d\n", entry->d_type); */
    return 0;
  }
}

static int myalphasort(const void* e1, const void* e2)
{
  char* d1 = *((char**) e1);
  char* d2 = *((char**) e2);
  return strcmp(d1,d2);
}

static int myscandir(char* dirname,
		     char *(*namelist[]))
{
  DIR* dir;
  struct dirent* cur;
  int nb = 0;
  int max = 100;
  dir = opendir(dirname);
  *namelist = NULL;
  if (!dir) return -1;
  *namelist = malloc(sizeof(void*)*max);
  cur = readdir(dir);
  while (cur) {
    if (scandir_select(cur)) {
      if (nb>=max) {
	max*=2;
	*namelist = realloc(*namelist, sizeof(void*)*max);
      }
      (*namelist)[nb] = strdup(cur->d_name);
      nb++;
    }
    cur = readdir(dir);
  }
  closedir(dir);
  qsort(*namelist, nb, sizeof(void*), myalphasort);
  return nb;
}


/**** audio sources */

/* fill audio_* variables, return 0 if OK, <0 on error. */
typedef int (*audsrc_set_up_t)(void);

/* return byte(!) count if OK or <0 on error. */
typedef int (*audsrc_get_frames_t)(oqt_uint8_t *buffer, int frames);

/* state variables used by the af audio source */
static AFfilehandle audio_input = AF_NULL_FILEHANDLE;
static int audio_frames_remaining = 0;
static int audio_frames_per_chunk;

/* this is the audio source based on lib audiofile */
static int audsrc_set_up_af(void)
{
  oqt_int32_t sampleFormat;

  /* Read audio from file */
  audio_input = afOpenFile(audio_filename, "r", AF_NULL_FILESETUP);

  if (audio_input == AF_NULL_FILEHANDLE) {
    fprintf(stderr, "%s: Cannot open audio file: %s.\n",
	    PROGRAM_NAME, audio_filename);
    oqt_close(out_file);
    exit(1);
  }

  /* Get information about the file audio_input from libaudiofile */
  audio_channels = afGetChannels(audio_input, AF_DEFAULT_TRACK);
  audio_sample_rate = afGetRate(audio_input, AF_DEFAULT_TRACK);
  afGetSampleFormat(audio_input, AF_DEFAULT_TRACK, &sampleFormat, &audio_bits);

  /* Make sure it is a sample format we support */
  if (sampleFormat != AF_SAMPFMT_TWOSCOMP) {
    fprintf(stderr, "%s: Sorry the only supported sample format is twos complement.\n",
	    PROGRAM_NAME);
    oqt_close(out_file);
    exit(1);
  }

  audio_sample_size = audio_channels * audio_bits / 8;

  /* Get the length of the audio file */
  audio_frames_remaining = afGetFrameCount(audio_input, AF_DEFAULT_TRACK);
  fprintf(stderr, "Total file length = %d audio frames.\n", audio_frames_remaining);

  return 0;
}

static int audsrc_get_frames_af(oqt_uint8_t *buffer, int frames)
{
  int frames_read;

  if(frames > audio_frames_remaining)
    frames = audio_frames_remaining;

  if(frames <= 0)
    return -1;

  frames_read = afReadFrames(audio_input, AF_DEFAULT_TRACK,
			     buffer, frames);

  if (frames_read <= 0 || frames_read > frames) {
    fprintf(stderr, "Bad read of audio track data.\n");
    return -1;
  }

  audio_frames_remaining -= frames_read;

  /* return bytes, not samples, so that the API can also work with
   * compressed formats */
  return frames_read * audio_sample_size;
}

/**** video sources */

/* fill video_* variables, return 0 if OK, <0 on error.
 * beware: this is called for each frame. */
typedef int (*vidsrc_next_header_t)(void);

/* get next frame from stream, return frame length if OK, <0 on error */
typedef int (*vidsrc_next_frame_t)(oqt_uint8_t *framebuffer, oqt_uint8_t **rowbuffer, int byteframesize, int *is_a_key);

/* state variable used by sources based on ppm/pgm files */
static FILE *pnm_file = NULL;

static int video_header_pnm(void)
{
  /* Parse the PPM/PGM header */
  if(parse_pnm_header(pnm_file)) {
    return -1;
  }
  return 0;
}

/* state variables used by the directory source (also uses pnm_file above) */
static char **dirlist = NULL;
static int dir_len = 0, dir_pos = 0;

static int video_header_pnm_dir(void)
{
  char *filepath;

  if(dirlist == NULL) {
    /* Remove trailing slashes */
    while (pnm_input_dir[0] && pnm_input_dir[strlen(pnm_input_dir)-1] == '/')
      pnm_input_dir[strlen(pnm_input_dir)-1] = 0;

    dir_len = myscandir(pnm_input_dir, &dirlist);
    if (dir_len==-1) {
      fprintf(stderr, "Failed to open directory: %s\n", pnm_input_dir);
      exit(1);
    }
    if (dir_len==0) {
      fprintf(stderr, "No PPM/PGM files found in directory: %s\n", pnm_input_dir);
      exit(1);
    }
  }

  if (dir_pos>=dir_len)
    return -1;

  /* Open the next file */
  filepath = (char*) malloc(MAXNAMLEN+strlen(pnm_input_dir)+1);
  sprintf(filepath, "%s/%s", pnm_input_dir, dirlist[dir_pos]);

  fprintf(stderr, "File [%d/%d]: %s\n", dir_pos, dir_len, filepath);

  pnm_file = fopen(filepath, "rb");

  free(filepath);

  return video_header_pnm();
}

static int video_frame_pnm(oqt_uint8_t *framebuffer, oqt_uint8_t **rowbuffer, int byteframesize, int *is_a_key)
{
  if (read_pnm_frame(pnm_file, framebuffer, rowbuffer, byteframesize)<0) {
    return -1;
  }
  *is_a_key = 1;
  return byteframesize;
}

static int video_frame_pnm_dir(oqt_uint8_t *framebuffer, oqt_uint8_t **rowbuffer, int byteframesize, int *is_a_key)
{
  int res = video_frame_pnm(framebuffer, rowbuffer, byteframesize, is_a_key);

  fclose(pnm_file);
  dir_pos++;
  return res;
}

/* state variables used by the compressed qt source */
static oqt_t *c_video_file = NULL;
static int c_video_nframes = 0;
static int c_video_frame = 0;
static int c_video_key = 0;

static int video_header_c_mov(void)
{
  if(c_video_file == NULL) {
    c_video_file = oqt_open(c_video_filename);

    if(!c_video_file) {
      fprintf(stderr, "%s: Failed to open file for video stream copy: %s\n", PROGRAM_NAME, c_video_filename);
      return -1;
    }

    if(oqt_read_headers(c_video_file)) {
      fprintf(stderr, "%s: Could not read movie headers from file: %s\n", PROGRAM_NAME, c_video_filename);
      return -1;
    }

    if(oqt_get_video_track_count(c_video_file) <= 0) {
      fprintf(stderr, "%s: File for video stream copy does not have a video track: %s\n", PROGRAM_NAME, c_video_filename);
      return -1;
    }

    video_width = oqt_get_video_width(c_video_file, 0);
    video_height = oqt_get_video_height(c_video_file, 0);
    if(video_rate == UNKNOWN_VIDEO_RATE)
      video_rate = oqt_get_video_framerate(c_video_file, 0);
    memcpy(video_codec, oqt_get_video_compressor(c_video_file, 0), 4);
    c_video_nframes = oqt_get_video_length(c_video_file, 0);
    c_video_frame = 0;
    video_cmodel = BC_RGB888;

    fprintf(stderr, "Copying compressed video %dx%d@%.3f [%c%c%c%c], %d frames\n", video_width, video_height, video_rate, video_codec[0], video_codec[1], video_codec[2], video_codec[3], c_video_nframes);
  }

  return 0;
}

static int video_frame_c_mov(oqt_uint8_t *framebuffer, oqt_uint8_t **rowbuffer, int byteframesize, int *is_a_key)
{
  if(c_video_frame >= c_video_nframes) {
    return -1;

  } else {
    int size = oqt_get_video_framesize(c_video_file, 0, c_video_frame);

    if(size > byteframesize)
      size = byteframesize;

    oqt_set_video_position(c_video_file, 0, c_video_frame);
    oqt_read_data(c_video_file, framebuffer, size);

    if(c_video_frame == c_video_key) {
      *is_a_key = 1;
      c_video_key = oqt_get_video_keyframe_after(c_video_file, 0, c_video_key+1);
    } else {
      *is_a_key = 0;
    }

    c_video_frame += 1;

    return size;
  }
}

/* state variable used for parsing YUV4MPEG2 */
static int seen_file_header = 0;

static int read_y4m_token(FILE *s, char *buf, int len)
{
  while(len > 0) {
    int ch = fgetc(s);
    if(ch == -1 || ch == ' ' || ch == '\n') {
      *buf = 0;
      return ch;
    }
    *buf = ch;
    ++buf;
    --len;
  }

  return 0;
}


static int video_header_y4m(void)
{
  static int frame_count = -1;
  char buffer[256];
  int res;

  if(!seen_file_header) {
    seen_file_header = 1;

    res = read_y4m_token(pnm_file, buffer, 255);
    if(res == 0 || res == -1 || strcmp(buffer, "YUV4MPEG2") != 0) {
      fprintf(stderr, "Expecting YUV4MPEG2 magic header\n");
      return -1;
    }

    while(res == ' ') {
      res = read_y4m_token(pnm_file, buffer, 255);
      if(res == 0 || res == -1) {
	fprintf(stderr, "Invalid YUV4MPEG2 header\n");
	return -1;
      }
      switch(buffer[0]) {
      case 'W':
	video_width = atoi(buffer+1);
	if(video_width <= 0) {
	  fprintf(stderr, "Invalid frame width from YUV4MPEG2 header: '%s'\n",
		  buffer+1);
	  return -1;	  
	}
	break;
      case 'H':
	video_height = atoi(buffer+1);
	if(video_height <= 0) {
	  fprintf(stderr, "Invalid frame height from YUV4MPEG2 header: '%s'\n",
		  buffer+1);
	  return -1;	  
	}
	break;
      case 'F':
	if(video_rate == UNKNOWN_VIDEO_RATE) {
	  char *colon = strchr(buffer+1, ':');
	  int p, q;

	  if(colon == NULL) {
	  fpserror:
	    fprintf(stderr, "Invalid frame height from YUV4MPEG2 header: '%s'\n",
		    buffer+1);
	    return -1;	  
	  }
	  p = atof(buffer+1);
	  q = atof(colon+1);

	  if(p <= 0 || q <= 0) goto fpserror;

	  video_rate = (0.0+p) / q;
	}
	break;
      }
    }

    if(video_rate == UNKNOWN_VIDEO_RATE) {
      fprintf(stderr, "Input does not specify framerate and -F is not given\n");
      return -1;
    }

  }

  ++frame_count;

  buffer[0] = 0;
  res = read_y4m_token(pnm_file, buffer, 255);
  if(res == 0 || res == -1 || strcmp(buffer, "FRAME") != 0) {
    fprintf(stderr, "Expecting YUV4MPEG2 FRAME header, frame #%d\n", frame_count);
    return -1;
  }

  while(res == ' ') {
    res = read_y4m_token(pnm_file, buffer, 255);
    if(res == 0 || res == -1) {
      fprintf(stderr, "Invalid YUV4MPEG2 FRAME header, frame #%d\n", frame_count);
      return -1;
    }
  }

  return 0;
}

static int video_frame_y4m(oqt_uint8_t *framebuffer, oqt_uint8_t **rowbuffer, int byteframesize, int *is_a_key)
{
  if(fread(framebuffer, 1, byteframesize, pnm_file) != byteframesize) {
    fprintf(stderr, "%s: Cannot read frame from input.\n", PROGRAM_NAME);
    return -1;
  }
  *is_a_key = 1;
  return byteframesize;
}


/**** main */

#define AUDIO_TS(frame_num) ((1.0*(frame_num))/audio_sample_rate)
#define VIDEO_TS(frame_num) ((1.0*(frame_num))/video_rate)

int main(int argc, char** argv)
{
  char *qtfilename = NULL;
  oqt_uint8_t *framebuffer = NULL;
  oqt_uint8_t **rowbuffer = NULL;
  oqt_uint8_t *audiobuffer = NULL;
  vidsrc_next_header_t vidsrc_next_header = NULL;
  vidsrc_next_frame_t vidsrc_next_frame = NULL;
  int is_video_input_compressed = 0;
  audsrc_set_up_t audsrc_set_up = NULL;
  audsrc_get_frames_t audsrc_get_frames = NULL;


  pnm_file = stdin;

  /* Establish signal handler to clean up before termination: */
  if (signal(SIGTERM, termination_handler) == SIG_IGN)
    signal(SIGTERM, SIG_IGN);
  if (signal(SIGINT, termination_handler) == SIG_IGN)
    signal(SIGINT, SIG_IGN);
  signal(SIGHUP, SIG_IGN);


  /* Parse the command line */
  qtfilename = parse_command_line( argc, argv );

  /* choose the video source handler */
  if(pnm_input_dir) {
    vidsrc_next_header = video_header_pnm_dir;
    vidsrc_next_frame = video_frame_pnm_dir;
  } else if(c_video_filename) {
    vidsrc_next_header = video_header_c_mov;
    vidsrc_next_frame = video_frame_c_mov;
    is_video_input_compressed = 1;
  } else if(video_cmodel == BC_YUV4MPEG2) {
    vidsrc_next_header = video_header_y4m;
    vidsrc_next_frame = video_frame_y4m;
  } else {
    vidsrc_next_header = video_header_pnm;
    vidsrc_next_frame = video_frame_pnm;
  }

  /* only one possible audio source for now */
  audsrc_set_up = audsrc_set_up_af;
  audsrc_get_frames = audsrc_get_frames_af;

  if(c_video_filename && pnm_input_dir) {
    fprintf(stderr, "%s: option -d is pointless when importing compressed video\n", PROGRAM_NAME);
  }

  if(video_cmodel == BC_YUV4MPEG2 && pnm_input_dir) {
    fprintf(stderr, "%s: option -d is pointless with YUV4MPEG2\n", PROGRAM_NAME);
  }

  /* Open the output file */
  if ( (out_file = oqt_open_file(qtfilename, OQT_MODE_WRITE) ) == NULL ) {
    fprintf(stderr, "%s: Cannot open output file: %s\n", PROGRAM_NAME, qtfilename);
    exit(1);
  }

  /* Open the audio input file */
  if(audio_track != NO_AUDIO_TRACK) {
    fprintf(stderr, "%s: Setting up audio %s\n", PROGRAM_NAME, audio_filename);
    audsrc_set_up();
  }


  /* Start the movie Data */
  oqt_write_moviedata_init(out_file);

  /* Does the audio track need creating? */
  if(audio_track == NEW_AUDIO_TRACK) {
    create_audio_track();

    audiobuffer = malloc(audio_block_size);
    audio_frames_per_chunk = audio_block_size/audio_sample_size;
    fprintf(stderr, "Reading in blocks of %d audio frames.\n", audio_frames_per_chunk);
  }

  { /* Here we go */
    int video_byteframesize = 0;
    int video_actualsize = 0;
    int audio_frames_read = 0;
    int is_a_key = 0;
    int qtcolormodel = BC_NONE;
    long video_frame_num = -1;
    /* 31 bits are only good for 13 hours of audio */
    long long audio_frame_num = -1;
    int have_video = (video_track != NO_VIDEO_TRACK);
    int have_audio = (audio_track != NO_AUDIO_TRACK);
    int video_buf_empty = have_video;
    int audio_buf_empty = have_audio;
    double next_video_ts;
    double next_audio_ts;

    /* This loop deals with the audio/video interleaving. */
    while(keep_going && (have_video || have_audio)) {

      /* read-ahead one video frame */
      if(video_buf_empty && have_video) {
	video_buf_empty = 0;

	if( vidsrc_next_header() ) {
	  /* no more video frames */
	  have_video = 0;

	} else {
	  /* Does the video track need creating? */
	  if(video_track == NEW_VIDEO_TRACK) {

	    create_video_track();

	    video_byteframesize = allocate_video_buffer( &framebuffer,
							 &rowbuffer );

	    qtcolormodel = video_cmodel;
	    if(video_cmodel==BC_YUV420PACKED || video_cmodel==BC_YUV4MPEG2)
	      qtcolormodel = BC_YUV420P;
	  }

	  /* read-in the frame */
	  video_actualsize = vidsrc_next_frame(framebuffer, rowbuffer,
					       video_byteframesize, &is_a_key);
	  if( video_actualsize <= 0 ) {
	    /* no more video frames */
	    have_video = 0;

	  } else {
	    video_frame_num += 1;
	    next_video_ts = VIDEO_TS(video_frame_num);
	  }
	}
      }


      /* read-ahead one audio buffer */
      if(audio_buf_empty && have_audio) {
	int res;
	audio_buf_empty = 0;

	res = audsrc_get_frames_af(audiobuffer, audio_frames_per_chunk);
	if (res <= 0) {
	  /* no more audio data */
	  have_audio = 0;

	} else {
	  audio_frames_read = res/audio_sample_size;
	  audio_frame_num += audio_frames_read;
	  next_audio_ts = AUDIO_TS(audio_frame_num);
	}
      }


      if(have_audio && (!have_video || next_audio_ts <= next_video_ts)) {
	/* Encode the audio to the file */
	int frames_written = oqt_encode_audio(out_file, audio_track,
					      audiobuffer, audio_frames_read);
	if (frames_written <= 0 || frames_written != audio_frames_read) {
	  fprintf(stderr, "Bad write of audio track data.\n");
	  break;
	}

	audio_buf_empty = 1;

      } else if(have_video) {

	if(is_video_input_compressed) {

	  /* Copy compressed frame */
	  oqt_write_video_data(out_file, video_track,
			       framebuffer, video_actualsize, is_a_key);

	} else {

	  /* Encode the video to the file */
	  oqt_encode_video(out_file, video_track, qtcolormodel, rowbuffer);
	}

	video_buf_empty = 1;
      }
    }
  }

  /* 'Flush' any remaining audio */
  if (audio_track != NO_AUDIO_TRACK) 
  	oqt_encode_audio(out_file, audio_track, NULL, 0);

  /* End of the movie data - and write 'moov' atom */
  oqt_write_moviedata_end(out_file);

  /* Time for closing our quicktime file */
  oqt_close(out_file);


  /* Free memory (actually useless) */
  if (framebuffer) free(framebuffer);
  if (rowbuffer) free(rowbuffer);
  if (audiobuffer) free(audiobuffer);

  return 0;
}
