/* oqtexport_audio.c
 * Copyright (C) 2002-2003 OpenQuicktime Team and University of Southampton
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Decode a track in a movie to an audio file.
 *
 *
 * $Id: oqtexport_audio.c,v 1.2 2003/02/20 17:10:07 nhumfrey Exp $
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include <audiofile.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"

int audio_rate, audio_bits;
int num_channels, sample_size, quiet=0;


#define BARWIDTH		(60)
#define PROGRAM_NAME	"oqtexport_audio"
#define FRAMES			4096

/* printaudioinfo.c */
void printfileinfo(char *filename);


/* decompression 
 * -------------
 */
 
static int decode_audio (oqt_t* infile, int qttrack, AFfilehandle outfile,
	int outtrack, AFframecount totalFrameCount)
{
	oqt_int64_t pos;
	oqt_int8_t* buffer;
	int buffer_size = FRAMES * sample_size * num_channels;
	int i, lastpercent = 0;
	int frames_decoded, frames_written;
	
	//fprintf(stderr, "buffer_size=%d, FRAMES=%d, sample_size=%d, num_channels=%d.\n",
	//			buffer_size, FRAMES, sample_size, num_channels);
	
	buffer = malloc(buffer_size);
	if (!buffer) {
		fprintf(stderr, "%s: malloc error in decompress_audio.\n", PROGRAM_NAME);
		return 1;
	}

	
	// Seek to the audio track
	oqt_set_audio_position(infile, qttrack, 0);

	for (pos=1;pos<totalFrameCount;) {	
		frames_decoded = oqt_decode_audio(infile, qttrack, buffer, buffer_size );
		if (frames_decoded <= 0) {
			fprintf(stderr, "\rFailed to decode chunk at sample %ld/%ld.\n",
			(long)pos, (long)totalFrameCount);
			break;
		}
		
		frames_written = afWriteFrames(outfile, outtrack, buffer, frames_decoded);
		if (frames_decoded <= 0) {
			fprintf(stderr, "\rFailed to write chunk to file at sample %ld/%ld.\n",
			(long)pos, (long)totalFrameCount);
			break;
		}
		// *** Check this value ? ***
		

		// Update the status bar
		if(!quiet) {
			int bardonelen = ((float)pos/totalFrameCount)*BARWIDTH;
			int barpercent = ((float)pos/totalFrameCount)*100.0;
			if (lastpercent != barpercent) {
				fprintf(stderr, "\r[");
				for(i=0;i<BARWIDTH;++i) {
					if (i<bardonelen) fprintf(stderr, ".");
					else fprintf(stderr, " ");
				}
				fprintf(stderr, "]   %3d%%", barpercent);
				lastpercent = barpercent;
			}
		}
		
		
		pos += frames_decoded;
	}
	
	free(buffer);
	
	fprintf(stderr, "\n");
	return 0;
} 




static void help(void) 
{ 
	fprintf(stderr,"%s: [options] infile outfile\n", PROGRAM_NAME);
	fprintf(stderr,"options:\n");
	fprintf(stderr,"     -h, --help          This message\n");
	fprintf(stderr,"     -m, --mono          Output as mono\n");
	fprintf(stderr,"     -q, --quiet         No progress information\n");
	fprintf(stderr,"     -t, --track <track> Select track to decode\n");
	fprintf(stderr,"     -f, --format        Format of output file (default aiff)\n");
	fprintf(stderr,"         aiff       Audio Interchange File Format\n");
	fprintf(stderr,"         aifc       AIFF-C File Format\n");
	fprintf(stderr,"         next       NeXT/Sun Format\n");
	fprintf(stderr,"         wave       MS RIFF WAVE Format\n");
#ifdef AF_FILE_BICSF
	fprintf(stderr,"         bics       Berkeley/IRCAM/CARL Sound File Format\n");
#endif
	exit(1);
}

int main(int argc, char** argv)
{
	char* in_filename = NULL;
	char* out_filename = NULL;
	int mono=0, num_tracks, audio_length;
	int outFileFormat = AF_FILE_AIFF;
	
	AFfilehandle	outfile;
	AFfilesetup		outfilesetup;
	AFframecount	totalFrames;
	
	oqt_t* qtfile;
	
	
	// Default to first audio track
	int audio_track=0;
	
	/* parse args */
	for (argc--,argv++;argc && *argv;argc--,argv++) {
		if      (!strcmp(*argv,"-m") || !strcmp(*argv,"--mono")) mono=1;
		else if (!strcmp(*argv,"-q") || !strcmp(*argv,"--quiet") ) quiet=1;
		else if (!strcmp(*argv,"-t") || !strcmp(*argv,"--track")) {
			if (sscanf(*(argv+1), "%d", &audio_track)<1) {
				fprintf(stderr, "%s: Invalid track number %s.\n", PROGRAM_NAME, *(argv+1));
				exit(1);
			}
			--audio_track;
			argc--;argv++;
		}
		else if (!strcmp(*argv,"-f") || !strcmp(*argv,"--format") ) {
			if (argc <= 1)
				help();
			if (!strcmp(*(argv+1), "aiff"))
				outFileFormat = AF_FILE_AIFF;
			else if (!strcmp(*(argv+1), "aifc"))
				outFileFormat = AF_FILE_AIFFC;
			else if (!strcmp(*(argv+1), "wave"))
				outFileFormat = AF_FILE_WAVE;
			else if (!strcmp(*(argv+1), "next"))
				outFileFormat = AF_FILE_NEXTSND;
#ifdef AF_FILE_BICSF
			else if (!strcmp(*(argv+1), "bics"))
				outFileFormat = AF_FILE_BICSF;
#endif
			else
			{
				fprintf(stderr, "%s: Unknown format %s.\n", PROGRAM_NAME, *(argv+1));
				exit(1);
			}
			argc--;argv++;
		}
		else if (!strcmp(*argv,"-?") || !strcmp(*argv,"-h") || !strcmp(*argv,"--help") ) help();
		else if (in_filename && !out_filename) out_filename=*argv;
		else if (!in_filename) in_filename=*argv;
		else {
			fprintf(stderr, "\nUnknown option \"%s\".\n", *argv);
			help();
		}
	}
	
	if (!in_filename) {
		fprintf(stderr, "Error: missing input file parameter.\n");
		help();
	}
	if (!out_filename) {
		fprintf(stderr, "Error: missing output file parameter.\n");
		help();
	}
	
	
	
	/* open QT file and get info */
	qtfile = oqt_open(in_filename);
	if (!qtfile) { 
		fprintf(stderr, "Unable to open file '%s'\n", in_filename); 
		exit(1); 
	}
	
	// Read headers
	if(oqt_read_headers(qtfile)) {
		oqt_close(qtfile);
		fprintf(stderr, "Could not read movie headers from file '%s'\n", in_filename);
		return 1;
	}

	/* Check for audio tracks */
	num_tracks = oqt_get_audio_track_count(qtfile);
	if (num_tracks<1) {
		fprintf(stderr, "Error: No audio tracks available in file.\n");
		exit(1);
	} else if (audio_track<0) {
		fprintf(stderr, "Audio track number should be greater than zero.\n");
	} else if (audio_track>=num_tracks) {
		fprintf(stderr, "Error: selected audio track (%d) is out of range.", audio_track+1);
		fprintf(stderr, " %d audio tracks in file.\n", num_tracks);
	}
	
	// We can only decode if track is supported...
	if (!oqt_supported_audio(qtfile, audio_track)) {
		char* codec_code = oqt_get_audio_compressor(qtfile, audio_track);
		fprintf(stderr, "Error: Audio format of track %d is not supported [%.4s].\n", audio_track+1, codec_code);
		exit(1);
	}
	

	/* Get information about the stream */
	audio_rate = oqt_get_audio_samplerate(qtfile, audio_track);
	audio_bits = oqt_get_audio_bits(qtfile, audio_track);
	audio_length = oqt_get_audio_length(qtfile, audio_track);
	sample_size = audio_bits/8;	// Length in bytes of samples
	num_channels = oqt_get_audio_channels(qtfile, audio_track);
	

	/* Display Info */
	if (!quiet)
		fprintf(stderr, "Audio track %i/%i at %i sps, %i channel(s), length %i s.\n",
		   audio_track+1,num_tracks,audio_rate, num_channels, audio_length/audio_rate);

	

	/* Initialize output audio format parameters. */
	outfilesetup = afNewFileSetup();
	
	afInitFileFormat(outfilesetup, outFileFormat);
	afInitSampleFormat(outfilesetup, AF_DEFAULT_TRACK, AF_SAMPFMT_TWOSCOMP, audio_bits);
	if (mono)	afInitChannels(outfilesetup, AF_DEFAULT_TRACK, 1);
	else		afInitChannels(outfilesetup, AF_DEFAULT_TRACK, num_channels);
	afInitRate(outfilesetup, AF_DEFAULT_TRACK, audio_rate);
	
	/* Open the output file */
	outfile = afOpenFile(out_filename, "w", outfilesetup);
	if (outfile == AF_NULL_FILEHANDLE)
	{
		fprintf(stderr, "Could not open file '%s' for writing.\n", out_filename);
		return 1;
	}
	afFreeFileSetup(outfilesetup);
	
	
	/*	Set the output file's virtual audio format parameters
		to match the format of the quicktime movie    */
	afSetVirtualChannels(outfile, AF_DEFAULT_TRACK, num_channels);
	afSetVirtualSampleFormat(outfile, AF_DEFAULT_TRACK, AF_SAMPFMT_TWOSCOMP, audio_bits);
	
	
	
	/* Start Decoding */
	if (decode_audio(qtfile, audio_track, outfile, AF_DEFAULT_TRACK, audio_length)) {
		fprintf(stderr, "Failed to decode audio.\n");
	} else {
		if (!quiet) fprintf(stderr, "Done.\n");
	}
	
	
	
	// Close the output file
	afCloseFile(outfile);
	
	// Display information about the output file
	if (!quiet)	printfileinfo(out_filename);
	
	
	oqt_close(qtfile);

	return 0;
}
