/* oqtexport_compressed.c
 * Copyright (C) 2002-2003 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: oqtexport_compressed.c,v 1.3 2003/03/15 05:21:04 jhatala Exp $
 */
 
 

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <openquicktime/openquicktime.h>
#include <openquicktime/colormodels.h>

#define PROGRAM_NAME "oqtexport_compressed"
#define AUDIO_TRACK	 0
#define VIDEO_TRACK	 1

/* Globals */
int	quiet = 0;

static void check(int cond)
{
	if(!cond) {
		perror("write failed");
		exit(1);
	}
}

static void extract_audio(oqt_t* infile, int track, FILE *outfile)
{
	oqt_int64_t offset;
	oqt_int32_t total_chunks, chunk, chunk_bytes;
	int result, written, extracted=0;
	char *buffer=NULL;
	
	// get the total number of chunks
	total_chunks = oqt_get_audio_chunks(infile, track);
	if (total_chunks==0) {
		fprintf(stderr, "%s: error, number of chunks in file is unknown.\n", PROGRAM_NAME);
		return;
	}
	
	for(chunk=1; chunk<=total_chunks; ++chunk) {
		
		// Get the length of the current chunk
		chunk_bytes = oqt_get_audio_chunksize(infile, track, chunk);
		if (chunk_bytes==0) {
			fprintf(stderr, "%s: warning chunk %d has no length.\n", PROGRAM_NAME, chunk);
			break;
		}
		
		// Message
		if (!quiet)
		fprintf(stderr, "Reading chunk %d/%d. Size: %d bytes.\n", chunk, total_chunks, chunk_bytes);

		// Allocate a buffer for the chunk
		buffer = realloc(buffer, chunk_bytes);

		// Read the chunk from the file
		result = oqt_set_audio_position_to_chunk(infile, track, chunk);
		if(!result) {
			result = !oqt_read_data(infile->data_file, buffer, chunk_bytes);
		}

		if (result) {
			fprintf(stderr, "%s: failed to read chunk %d: %d.\n", PROGRAM_NAME, chunk, result);
			break;
		}
	
		// Write the chunk to the raw output file
		written = fwrite(buffer, 1, chunk_bytes, outfile);
		if (written!=chunk_bytes)
			fprintf(stderr, "%s: only wrote %d of %d bytes of chunk to file.\n", PROGRAM_NAME, written, chunk_bytes);
			
		extracted+=written;
			
	}

	if (!quiet)
	fprintf(stderr, "Extract and wrote %d bytes to file.\n", extracted);
}



static void extract_video(oqt_t *src, int track, FILE *dst, int frame_lo, int frame_hi)
{
	oqt_int64_t i, frames = oqt_get_video_length(src, track);
	int depth = oqt_get_video_depth(src, track);
	int width = oqt_get_video_width(src, track);
	int height = oqt_get_video_height(src, track);
	oqt_uint8_t* buffer = (oqt_uint8_t*)malloc(4*width*height);
	int size;
	
	for(i = frame_lo; i <= frame_hi; ++i) {
		size = oqt_get_video_framesize(src, track, i);
		oqt_set_video_position(src, track, i);
		if (!quiet)
			fprintf(stderr, "extract_compressed: frame %lld/%lld length=%d.\n", i, frames, size);
		oqt_read_data(src->data_file, buffer, size);
		
		check( fwrite(buffer, size, 1, dst) == 1 );
	}
	
	free(buffer);
}


static void extract_video_dir(oqt_t *src, int track, char *dir, char *outfile,
			      int frame_lo, int frame_hi)
{
	oqt_int64_t i, frames = oqt_get_video_length(src, track);
	int depth = oqt_get_video_depth(src, track);
	int width = oqt_get_video_width(src, track);
	int height = oqt_get_video_height(src, track);
	oqt_uint8_t* buffer = (oqt_uint8_t*)malloc(4*width*height);
	char* filename;
	FILE* out;
	
	filename = (char*) malloc(32+strlen(dir)+strlen(outfile));

	for(i = frame_lo; i <= frame_hi; ++i) {
		int size = oqt_get_video_framesize(src, track, i);
		oqt_set_video_position(src, track, i);
		oqt_read_data(src->data_file, buffer, size);
		sprintf(filename, "%s/%.6lld%s", dir, i, outfile);
		
		if (!quiet)
			fprintf(stderr, "%s: frame %lld/%lld length=%d.\n", filename, i, frames, size);
		out = fopen(filename, "wb");
		if(!out) {
			perror("Could not open output file");
		} else {
			check( fwrite(buffer, size, 1, out) == 1 );
			fclose(out);
		}
	}
	
	free(buffer);
	free(filename);
}


static void usage(void) 
{ 
	fprintf(stderr,"%s: [options] infile.mov outfile\n", PROGRAM_NAME);
	fprintf(stderr,"options:\n");
	fprintf(stderr,"     -h         This message\n");
	fprintf(stderr,"     -q         Quiet Mode\n");
	fprintf(stderr,"     -v         Extract a video track (default)\n");
	fprintf(stderr,"     -a         Extract an audio track\n");
	fprintf(stderr,"     -t <track> Select track (1-based)\n");
	fprintf(stderr,"     -d <dir>   Each frame as file into directory (video only)\n");
	fprintf(stderr,"     -f <n>     Only extract specified frame (video only)\n");
	fprintf(stderr,"     -f <n>-<m> Only extract frames n .. m (video only)\n");
	exit(1);
}

int main(int argc, char **argv)
{
	int track = 0;
	char *infile_name = NULL;
	char *outfile_name = NULL;
	oqt_t *src = NULL;
	FILE *dst = NULL;
	int frame_lo = 0;
	int frame_hi = -1;
	int track_type = VIDEO_TRACK;
	char* dir = NULL;

	while(1) {
		int c = getopt(argc, argv, "hHqd:f:vat:");
		if (c == -1) break;

		switch(c) {
			case 'h':
			case 'H':	usage();					break;
			case 'q':	quiet = 1;					break;
			case 'd':	dir = optarg;				break;
			case 'a':	track_type = AUDIO_TRACK;	break;
			case 'v':	track_type = VIDEO_TRACK;	break;
			case 'f': {
			    char *dash = strchr(optarg,'-');
			    if(dash == NULL) {
			      frame_lo = frame_hi = atoi(optarg);
			    } else if(dash == optarg) {
			      frame_lo = 0;
			      frame_hi = atoi(optarg+1);
			    } else {
			      frame_lo = atoi(optarg);
			      frame_hi = atoi(dash+1);
			    }
			  }
			  break;
			case 't':	track = atoi(optarg);		break;
			default:
				fprintf(stderr, "%s: Unknown option specified %c\n", PROGRAM_NAME, c);
				usage();
			break;
		}
	}
	
	if((argc - optind) < 1 || !(infile_name = argv[optind++]) ) {
		fprintf(stderr, "Missing input file parameter.\n");
		usage();
	}
	if((argc - optind) < 1 || !(outfile_name = argv[optind++]) ) {
		fprintf(stderr, "Missing output file parameter.\n");
		usage();
	}
	if(track_type == AUDIO_TRACK &&
	   (frame_lo != 0 || frame_hi != -1 || dir)) {
		fprintf(stderr, "Sorry can't specify a frame number or directory for audio tracks.\n");
		exit(2);
	}


	/* Open the source movie */
	src = oqt_open(infile_name);
	if(!src) {
		fprintf(stderr, "Could not open QT file: %s\n", infile_name);
		exit(3);
	}

	/* Read headers */
	if(oqt_read_headers(src)) {
		oqt_close(src);
		printf("Could not read movie headers from file: %s\n", infile_name);
		exit(4);
	}


	/* Output as a single file ? */
	if (!dir) {
		if (!strcmp(outfile_name, "-")) {
			dst = stdout;
		} else {
			dst = fopen(outfile_name, "wb");
			if(!dst) {
				perror("Could not open output file");
				exit(6);
			}
		}
	}
	

	if (track_type == VIDEO_TRACK) {
		/* Export compressed video data */
	        oqt_int64_t frames;

		/* Check track number is valid */
		if(track < 0 || track >= oqt_get_video_track_count(src)) {
			fprintf(stderr, "Error: invalid video track number: %d\n", track+1);
			exit(5);
		}

		
		frames = oqt_get_video_length(src, track);
		if(frame_lo < 0 || frame_lo >= frames) {
			fprintf(stderr, "Error: Frame number not valid: %d\n", frame_lo);
			exit(11);
		}

		if(frame_hi < 0) {
		        // -1 => last, -2 => the one before, ...
		        frame_hi = frames + frame_hi;
		}
		if(frame_hi < 0 || frame_hi >= frames) {
			fprintf(stderr, "Error: Frame number not valid: %d\n", frame_hi);
			exit(12);
		}
	
		/* Output as a single file ? */
		if (!dir) {
			
			extract_video(src, track, dst, frame_lo, frame_hi);
		
		} else {
			/* Output as a directory */
			if (mkdir(dir, 0755)) {
				perror("Failed to create directory");
				exit(6);
			}
			
			extract_video_dir(src, track, dir, outfile_name, frame_lo, frame_hi);
		}
	} else {
		/* Export compressed audio data */

		/* Check track number is valid */
		if(track < 0 || track >= oqt_get_audio_track_count(src)) {
			fprintf(stderr, "Error: invalid audio track number: %d\n", track+1);
			exit(5);
		}

		/* Sadly the audio track must be Supported to extract the raw
		   audio. This is because we have to be able to calculate the 
		   length of the chunks */
		if (!oqt_supported_audio(src, track)) {
			char* codec_code = oqt_get_audio_compressor(src, track);
			fprintf(stderr, "Error: Audio format of track %d is not supported [%.4s].\n", track+1, codec_code);
			exit(1);
		}
		
		extract_audio(src, track, dst);
	}
			
	if (!quiet) fprintf(stderr, "Done.\n");

	oqt_close(src);
	if (dst) fclose(dst);

	
	return 0;
}


