/* oqtexport_ppm.c
 * Copyright (C) 2002 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: oqtexport_ppm.c,v 1.8 2003/05/15 13:24:43 nhumfrey Exp $
 */
 
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <openquicktime/openquicktime.h>
#include <openquicktime/colormodels.h>

#define PROGRAM_NAME "oqtexport_ppm"


static void check(int cond)
{
	if(!cond) {
		perror("write failed");
		exit(1);
	}
}

static void extract_yuv(oqt_t *src, int track, FILE *dst, int frame_lo, int frame_hi, int y4m, int yuvpgm)
{
	oqt_int64_t i, frames = oqt_get_video_length(src, track);
	int width = oqt_get_video_width(src, track);
	int height = oqt_get_video_height(src, track);
	double rate = oqt_get_video_framerate(src, track);
	oqt_uint8_t* y = (oqt_uint8_t*) malloc(width*height);
	oqt_uint8_t* u = (oqt_uint8_t*) malloc(width*height/4);
	oqt_uint8_t* v = (oqt_uint8_t*) malloc(width*height/4);
	oqt_uint8_t *rows[3];
	rows[0] = y;
	rows[1] = u;
	rows[2] = v;

	if(y4m) {
		int p, q;
		// factorize the frame-rate, make sure
		// we get well-known frame-rates right
		if(fabs(rate - 29.97) < 0.01) {
			p = 30000;
			q = 1001;
		} else if(fabs(rate - 23.97) < 0.01) {
			p = 24000;
			q = 1001;
		} else if(fabs(rate - 59.94) < 0.01) {
			p = 60000;
			q = 1001;
		} else if(rate == (int)rate) {
			p = (int)rate;
			q = 1;
		} else {
			p = (int)(rate * 1000);
			q = 1000;
		}
		fprintf(dst, "YUV4MPEG2 W%d H%d F%d:%d\n",
			width, height, p, q);

	} else if(!yuvpgm) {
		fprintf(stderr, "-w %d -h %d -r %.2f\n", width, height, rate);
	}

	for(i = frame_lo; i <= frame_hi; ++i) {
		if(y4m) {
			fprintf(dst, "FRAME\n");
		} else if(yuvpgm) {
			fprintf(dst, "P5\n%d %d\n255\n", width, (height*3)/2);
		}
		oqt_set_video_position(src, track, i);

		fprintf(stderr, "frame %lld/%lld.\n", i, frames);
		oqt_decode_video(src, track, BC_YUV420P, rows);
		check( fwrite(y, width*height, 1, dst) == 1 );
		check( fwrite(u, width*height/4, 1, dst) == 1 );
		check( fwrite(v, width*height/4, 1, dst) == 1 );
	}
	free(y);
	free(u);
	free(v);
}

static void extract_rgb(oqt_t *src, int track, FILE *dst, int frame_lo, int frame_hi, int simple_ppm)
{
	oqt_int64_t i, frames = oqt_get_video_length(src, track);
	int width = oqt_get_video_width(src, track);
	int height = oqt_get_video_height(src, track);
	int row;
	
	/* Allocate the output buffer */
	oqt_uint8_t** rows = (oqt_uint8_t**) malloc(height*sizeof(oqt_uint8_t*));
	for(row=0;row<height;row++) rows[row] = malloc(width*sizeof(oqt_uint8_t)*3);
	
	fprintf( dst, "P6\n" );
	if(!simple_ppm) {
		fprintf( dst, "# File written by %s.\n", PROGRAM_NAME );
		fprintf( dst, "# Decoded using OpenQuicktime %s\n", oqt_version());
		fprintf( dst, "# Original file: %s\n", src->stream_reference);
		if (frame_lo != 0 || frame_hi != frames-1)
		        fprintf( dst, "# Frames: %d-%d/%lld\n", frame_lo, frame_hi, frames);
		}
	fprintf( dst, "%d %d\n255\n", width, height);
	
	for(i = frame_lo; i <= frame_hi; ++i) {
		
		/* repeat a PPM header for each frame */
		if(i != frame_lo) {
			fprintf( dst, "P6\n" );
			fprintf( dst, "%d %d\n255\n", width, height);
		}

		oqt_set_video_position(src, track, i);

		fprintf(stderr, "frame %lld/%lld.\n", i, frames);
		oqt_decode_video(src, track, BC_RGB888, rows);
		
		/* Output the rows to the file */
		for(row=0;row<height;row++)
			check( fwrite(rows[row], width*3, 1, dst) == 1 );
	}
	
	/* Release the buffer */
	for(row=0;row<height;row++) free(rows[row]);
	free(rows);
}


static void extract_rgb_dir(oqt_t *src, int track, char *dir, char *outfile, int frame_lo, int frame_hi, int simple_ppm)
{
	oqt_int64_t i, frames = oqt_get_video_length(src, track);
	int width = oqt_get_video_width(src, track);
	int height = oqt_get_video_height(src, track);
	char* filename;
	FILE* out;
	int row;
	
	/* Allocate the output buffer */
	oqt_uint8_t** rows = (oqt_uint8_t**) malloc(height*sizeof(oqt_uint8_t*));
	for(row=0;row<height;row++) rows[row] = malloc(width*sizeof(oqt_uint8_t)*3);
	
	filename = (char*) malloc(32+strlen(dir)+strlen(outfile));

	for(i = frame_lo; i <= frame_hi; ++i) {
		oqt_set_video_position(src, track, i);
		oqt_decode_video(src, track, BC_RGB888, rows);
		sprintf(filename, "%s/%s%.6lld.ppm", dir, outfile, i);

		fprintf(stderr, "%s: frame %lld/%lld.\n", filename, i, frames);
		
		/* Open the output file */
		out = fopen(filename, "wb");
		if(!out) {
			perror("Could not open output file");
		} else {
			fprintf( out, "P6\n" );
			if(!simple_ppm) {
				fprintf( out, "# File written by %s.\n", PROGRAM_NAME );
				fprintf( out, "# Decoded using OpenQuicktime %s\n", oqt_version());
				fprintf( out, "# Original file: %s\n", src->stream_reference);
				fprintf( out, "# Frame: %lld/%lld\n", i, frames);
			}
			fprintf( out, "%d %d\n255\n", width, height);
			
			/* Output the rows to the file */
			for(row=0;row<height;row++)
				check( fwrite(rows[row], width*3, 1, out) == 1 );
			fclose(out);
		}
	}
	
	/* Release the buffer */
	for(row=0;row<height;row++) free(rows[row]);
	free(rows);
	free(filename);
}


static void usage(void) 
{ 
	fprintf(stderr,"%s: [options] infile outfile\n", PROGRAM_NAME);
	fprintf(stderr,"options:\n");
	fprintf(stderr,"     -h, --help          This message\n");
	fprintf(stderr,"     -d, --dir <dir>     Each frame as file into directory\n");
	fprintf(stderr,"     -y, --yuv           Extract header-less YUV and\n");
	fprintf(stderr,"                         output \"-w <width> -h <height> -r <rate>\" on stdout\n");
	fprintf(stderr,"     -Y, --YUV4MPEG2     Extract YUV4MPEG2\n");
	fprintf(stderr,"     -r, --rgb           Extract PPM RGB format (P6) [default]\n");
	fprintf(stderr,"     -g                  Extract PGM YUV420P format (P5)\n");
	fprintf(stderr,"     -s, --simple-ppm    Simple PPM headers (for ffmpeg)\n");
	fprintf(stderr,"     -f, --frame <n>     Only extract specified frame\n");
	fprintf(stderr,"     -f, --frames <n>-<m> Only extract frames n .. m\n");
	fprintf(stderr,"     -p, --poster        Only extract poster frame\n");
	fprintf(stderr,"     -t, --track <track> Select track (1-based)\n");
	exit(1);
}

int main(int argc, char **argv)
{
	int comp = 0, yuv = 0, yuv4mpeg = 0, yuvpgm = 0, rgb = 0;
	int simple_ppm = 0;
	int track = 0;
	char *infile = NULL;
	char *outfile = NULL;
	oqt_t *src = NULL;
	FILE *dst = NULL;
	int width = -1;
	int height = -1;
	int frame_lo = 0;
	int frame_hi = -1;
	int poster = 0;
	float rate = 0;
	char* dir = NULL;

	if(argc == 1) {
		usage();
	}

	while(argc > 1 && argv[1]) {
		if(argv[1][0] == '-' && argv[1][1] != 0) {
			char c = argv[1][1];
			if(c == '-') {
				c = argv[1][2];
			}
			switch(c) {
				case 'h' :
					usage(); break;
				case 'y' :
					yuv = 1; break;
				case 'Y' :
					yuv4mpeg = 1; break;
				case 'c' :
					comp = 1; break;
				case 'r' :
					rgb = 1; break;
				case 'g' :
					yuvpgm = 1; break;
				case 's' :
					simple_ppm = 1; break;
				case 'd' :
					dir = argv[2];
					argv += 1;
					argc -= 1;
					break;
				case 'p' :
					poster = 1; break;
				case 'f':
				  if(!argv[2]) {
				    fprintf(stderr, "Expecting a frame number\n");
				  } else {
				    char *dash = strchr(argv[2],'-');
				    if(dash == NULL) {
				      frame_lo = frame_hi = atoi(argv[2]);
				    } else if(dash == argv[2]) {
				      frame_lo = 0;
				      frame_hi = atoi(argv[2]+1);
				    } else {
				      frame_lo = atoi(argv[2]);
				    frame_hi = atoi(dash+1);
				    } 
				    argv += 1;
				    argc -= 1;
				  } break;
				case 't' :
					if(!argv[2] ||
					    argv[2][0] < '0' ||
					    argv[2][0] > '9') {
						fprintf(stderr, "Expecting a track number: '%s'\n", argv[2]);
						exit(2);
					} else {
						track = atoi(argv[2]) - 1;
						argv += 1;
						argc -= 1;
					} break;
				default:
					fprintf(stderr, "Unknown option: '%s'\n", argv[1]);
					exit(2);
			}
		} else if(!infile) {
			infile = argv[1];
		} else if(!outfile) {
			outfile = argv[1];
		} else {
			fprintf(stderr, "Extra parameter: '%s'\n", argv[1]);
			exit(2);
		}
		argv += 1;
		argc -= 1;
	}

	// default: rgb
	if (!comp && !yuv && !yuv4mpeg && !yuvpgm && !rgb) rgb = 1;
	
	if(!infile) {
		fprintf(stderr, "Missing input file parameter\n");
		exit(2);
	}
	if(!outfile) {
		fprintf(stderr, "Missing output file parameter\n");
		exit(2);
	}
	if(yuv + yuv4mpeg + yuvpgm + rgb != 1) {
		fprintf(stderr, "Please decide between yuv and rgb\n");
		exit(2);
	}


	/* Open the source movie */
	src = oqt_open(infile);
	if(!src) {
		fprintf(stderr, "Could not open QT file: %s\n", infile);
		exit(3);
	}

	/* Read headers */
	if(oqt_read_headers(src)) {
		oqt_close(src);
		printf("Could not read movie headers from file: %s\n", infile);
		exit(4);
	}

	/* Check track number is valid */
	if(track < 0 || track >= oqt_get_video_track_count(src)) {
		fprintf(stderr, "Error: invalid track number: %d\n", track+1);
		exit(5);
	}
	
	/* Ensure codec is supported */
	if (!comp && !oqt_supported_video(src, track)) {
		fprintf(stderr, "Error: video compressor not supported [%.4s].\n", 
		oqt_get_video_compressor(src, track));
		exit(5);
	}

	/* Did they ask for the poster frame ? */
	if (poster) {
		frame_lo = frame_hi = oqt_get_video_poster(src, track);
		fprintf(stderr, "Poster frame is number %d.\n", frame_lo);

	} else { /* Verify specified frame numbers */
	        oqt_int64_t frames  = oqt_get_video_length(src, track);

		if(frame_lo < 0 || frame_lo >= frames) {
			fprintf(stderr, "Error: Frame number not valid: %d\n", frame_lo);
			exit(11);
		}

		if(frame_hi < 0) {
		        // -1 => last, -2 => the one before, ...
		        frame_hi = frames + frame_hi;
		}
		if(frame_hi < 0 || frame_hi >= frames) {
			fprintf(stderr, "Error: Frame number not valid: %d\n", frame_hi);
			exit(12);
		}
	}
	
	/* Output as a single file ? */
	if (!dir) {
		if (!strcmp(outfile, "-")) {
			dst = stdout;
		} else {
			dst = fopen(outfile, "wb");
			if(!dst) {
				perror("Could not open output file");
				exit(6);
			}
		}
		
		if (yuv || yuv4mpeg || yuvpgm) {
			extract_yuv(src, track, dst, frame_lo, frame_hi, yuv4mpeg, yuvpgm);
		} else if (rgb) {
			extract_rgb(src, track, dst, frame_lo, frame_hi, simple_ppm);
		}
	
		fclose(dst);
	} else {
		/* Output as a directory */
		if (mkdir(dir, 0755)) {
			perror("Failed to create directory");
			exit(6);
		}
		
		if (rgb) {
			extract_rgb_dir(src, track, dir, outfile, frame_lo, frame_hi, simple_ppm);
		} else if (yuv) {
			printf("YUV directory output not supported.\n");
		}		
	}	
		
	oqt_close(src);
	
	return 0;
}


