/* oqtimport_dv.c
 * Copyright (C) 2003 OpenQuicktime Team
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: oqtimport_dv.c,v 1.9 2003/05/15 13:26:14 nhumfrey Exp $
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <openquicktime/colormodels.h>
#include <openquicktime/structs.h>
#include <openquicktime/openquicktime.h>
#include <openquicktime/plugin.h>

#include <libdv/dv.h>

#define MIN_FRAME_SIZE 120000
#define MAX_FRAME_SIZE 144000

int main(int argc, char **argv)
{
  int parity = 0;
  oqt_t *in_file;
  oqt_t *out_file;
  dv_decoder_t *decoder;
  int first_time = 1;
  oqt_uint8_t *buffer;
  double fps;
  int xref = 0;
  oqt_uint64_t src_position = 0;
  char *url = NULL;
  int audio_only = 0;
  int video_only = 0;
  int frame_lo = 0;
  int frame_hi = -1;
  int frame;

  if(argc < 3) {
  usage:
    fprintf(stderr, "Usage: oqtimport_dv [options] src.dv dest.mov    (self-containted movie)\n");
    fprintf(stderr, "    or oqtimport_dv -x [options] src.dv dest.mov (external reference)\n");
    fprintf(stderr, "       options:\n");
    fprintf(stderr, "         -v or -a for video or audio only\n");
    fprintf(stderr, "         -f n-m   to import only frames n to m inclusive\n");
    return 1;
  }

  decoder = dv_decoder_new(TRUE, FALSE, FALSE);
  if(!decoder) {
    fprintf(stderr, "dv_decoder_new failed\n");
    return 2;
  }

  while(argc > 2) {
    if(strcmp(argv[1], "-x") == 0) {
      xref = 1;
      argv += 1;
      argc -= 1;
    } else if(strcmp(argv[1], "-a") == 0) {
      audio_only = 1;
      argv += 1;
      argc -= 1;
    } else if(strcmp(argv[1], "-v") == 0) {
      video_only = 1;
      argv += 1;
      argc -= 1;
    } else if(strcmp(argv[1], "-f") == 0) {
      char *ptr;
      argv += 1;
      argc -= 1;
      if(argc <= 2) {
	fprintf(stderr, "-f needs a parameter!\n");
	goto usage;
      }
      frame_lo = atoi(argv[1]);
      ptr = strchr(argv[1], '-');
      if(ptr == NULL) {
	frame_hi = frame_lo;
      } else if(ptr[1] == 0) {
	frame_hi = -1;
      } else {
	frame_hi = atoi(ptr+1);
      }
      argv += 1;
      argc -= 1;
    } else {
      break;
    }
  }

  if(video_only && audio_only) {
    fprintf(stderr, "-a and -v are exclusive!\n");
    goto usage;
  }

  if(argc != 3) goto usage;

  if(xref) {
    url = malloc(strlen(argv[1]) + strlen("file:") + 1);
    strcpy(url, "file:");
    strcat(url, argv[1]);
  }


  in_file = oqt_open_file(argv[1], OQT_MODE_READ);
  if(!in_file) {
    perror(argv[1]);
    return 2;
  }

  out_file = oqt_open_file(argv[2], OQT_MODE_WRITE);
  if(!out_file) {
    fprintf(stderr, "failed to open '%s' for writing\n", argv[2]);
    return 2;
  }

  oqt_write_moviedata_init(out_file);

  buffer = malloc(MAX_FRAME_SIZE);
  
  for(frame = 0; frame_hi < 0 || frame <= frame_hi; frame++) {

    if(first_time) {
      first_time = 0;

      // read in the header
      if(in_file->oqt_read(in_file, buffer, MIN_FRAME_SIZE) != 1) {
	break;
      }

      // have libdv come up with stream properties
      if(dv_parse_header(decoder, buffer) < 0) {
	fprintf(stderr, "dv_parse_header failed\n");
	return 3;
      }

      fps = (1.0*decoder->audio->frequency) / decoder->audio->samples_this_frame;
      // round if near to a well known value
      if(fabs(fps - 29.97) < 0.01) fps = 29.97;
      else if(fabs(fps - 24.97) < 0.01) fps = 24.97;
      else if(fabs(fps - 30) < 0.01) fps = 30;
      else if(fabs(fps - 25) < 0.01) fps = 25;

      if(!audio_only) {
	oqt_add_video_track(out_file, decoder->width, decoder->height,
			    fps, "dvc ");
      }

      if(!video_only) {
	oqt_add_audio_track(out_file, decoder->audio->num_channels,
			    decoder->audio->frequency, 16, "dvca");
      }

      if(decoder->frame_size > MAX_FRAME_SIZE) {
	fprintf(stderr, "unexpected frame size: %d\n", (int)decoder->frame_size);
	return 3;
      }

      // finish reading the first frame
      if(decoder->frame_size > MIN_FRAME_SIZE) {
	if(in_file->oqt_read(in_file, buffer+MIN_FRAME_SIZE,
			     decoder->frame_size-MIN_FRAME_SIZE) != 1) {
	  break;
	}
      }

    } else {
      // the second and later frames
      if(in_file->oqt_read(in_file, buffer, decoder->frame_size) != 1) {
	break;
      }
    }

    // audio and video share the same data
    if(xref) {
      if(frame >= frame_lo) {
	if(!audio_only) {
	  oqt_write_video_reference(out_file, 0, url, src_position, decoder->frame_size, 1);
	}
	if(!video_only) {
	  oqt_write_audio_reference(out_file, 0, url, src_position, decoder->frame_size, decoder->audio->samples_this_frame);
	}
      }

      src_position += decoder->frame_size;

    } else if(frame >= frame_lo) {

      if(audio_only) {
	oqt_write_video_data(out_file, 0, buffer, decoder->frame_size, 1);
      } else if(video_only) {
	oqt_write_audio_data(out_file, 0, buffer, decoder->frame_size,
			     decoder->audio->samples_this_frame, 1, NULL);
      } else {
	oqt_int64_t pos = oqt_get_position(out_file);
	oqt_write_video_data(out_file, 0, buffer, decoder->frame_size, 1);
	oqt_set_position(out_file, pos);
	oqt_write_audio_data(out_file, 0, buffer, decoder->frame_size,
			     decoder->audio->samples_this_frame, 1, NULL);
      }
    }
  }

  oqt_write_moviedata_end(out_file);
  oqt_close(out_file);
  dv_decoder_free(decoder);
  in_file->oqt_close(in_file);
  free(buffer);
  return 0;
}
