/* oqtimport_mash.c
 * Copyright (C) 2003 OpenQuicktime Team & University of Southampton
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * oqtimport_mash by Nicholas Humfrey <njh@ecs.soton.ac.uk>
 *  Convert OpenMash archives to quicktime movie file
 *  Written for use by the CoAKTing Project
 *   - http://www.aktors.org/coakting/
 *
 * $Id: oqtimport_mash.c,v 1.2 2003/04/20 00:32:53 nhumfrey Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"


#define PROGRAM_NAME	"oqtimport_mash"
#define DATA_VERSION	"MDAT1.0"
#define INDEX_VERSION	"MIDX1.0"

#define PROTO_SRM		"SRM"
#define PROTO_RTP		"RTP"
#define NO_TRACK	  	-1


struct FileHeader {
	char version[16];	// DATA_VERSION & INDEX_VERSION
	char protocol[8];	// "RTP" or "SRM"
	char media[32];		// the media: "audio", "video", "mediaboard"...
	char cname[128];	// username@host.domain
	char name[128];		// the user's name

  // all of the following values are stored in network form, so require a
  //   ntohl() call when reading from the file
	oqt_uint32_t start_sec;    // start timestamp
	oqt_uint32_t start_usec;
	oqt_uint32_t end_sec;      // end timestamp
	oqt_uint32_t end_usec;

	oqt_uint32_t privateLen;	// length of any following private header
};

struct IndexRecord {
	oqt_uint32_t sentTS_sec; // sender timestamp: depends on the media stream
	oqt_uint32_t sentTS_usec;
	oqt_uint32_t recvTS_sec; // timestamp when the packet was rcvd by the
	oqt_uint32_t recvTS_usec;// recorder
	oqt_uint32_t seqno;      // packet sequence number
	oqt_uint32_t filePointer;// location of packet in the data file
};

struct RTPHeader {
	oqt_uint32_t	length;		/* Length of the RTP Packet */
	oqt_uint32_t	unknown;	/* ?? */

	/* Start of RTP Packet */
	unsigned version : 2;
	unsigned padding : 1;
	unsigned extension : 1;
	unsigned csrc : 4;
	unsigned marker : 1;
	unsigned payload : 7;

	oqt_uint16_t	seq_num;
	oqt_uint32_t	ts;
	oqt_uint32_t	ssrc;
};

struct H261RTPHeader {
	unsigned sbit : 3;
	unsigned ebit : 3;
	unsigned i : 1;
	unsigned v : 1;
	unsigned gobn : 4;
	unsigned mbap : 5;
	unsigned quant : 5;
	signed hmvd : 5;
	signed vmvd : 5;
};	


struct H261Head {
	unsigned psc : 20;	/* Picture Start code */
	unsigned tr : 5;	/* Temporal Reference */
	
	/* Type inforamtion */
	unsigned ss : 1;	/* Split screen indicator */
	unsigned dc : 1;	/* Document camera indicator */
	unsigned fpr : 1;	/* Freeze picture release */
	unsigned sf : 1;	/* Source Format (QCIF/CIF) */
	unsigned si : 1;	/* Still Image mode */
	unsigned spare : 1;	/* Unused */
	
	unsigned pei : 1;	/* Extra insertion Information */
	
	unsigned gbsc : 16;	/* 0x0001 */
	unsigned gn : 4;	/* Group number */
	unsigned gquant : 5;/* Quantizer Info */
	unsigned gei : 1;	/* Extra insertion Information */
	
};




static void fileheader_ntoh(struct FileHeader *fh) {
	fh->start_sec = ntohl(fh->start_sec);
	fh->start_usec = ntohl(fh->start_usec);
	fh->end_sec = ntohl(fh->end_usec);
	fh->privateLen = ntohl(fh->privateLen);
}

static void indexrecord_ntoh(struct IndexRecord *ir) {
	ir->sentTS_sec = ntohl(ir->sentTS_sec);
	ir->sentTS_usec = ntohl(ir->sentTS_usec);
	ir->recvTS_sec = ntohl(ir->recvTS_sec);
	ir->recvTS_usec = ntohl(ir->recvTS_usec);
	ir->seqno = ntohl(ir->seqno);
	ir->filePointer = ntohl(ir->filePointer);
}

static void rtpheader_ntoh(struct RTPHeader *rtph) {
	rtph->length = ntohl(rtph->length);
	rtph->unknown = ntohl(rtph->unknown);
	rtph->seq_num = ntohs(rtph->seq_num);
	rtph->ts = ntohl(rtph->ts);
	rtph->ssrc = ntohl(rtph->ssrc);
}

static void display_rtp_headers( struct IndexRecord *ir, struct RTPHeader *rtph ) {
	//static int i=20;
	
	//if (i && i--) {
		fprintf(stdout, "Index sentTS: %d.%d\n", ir->sentTS_sec, ir->sentTS_usec);
		fprintf(stdout, "Index recvTS: %d.%d\n", ir->recvTS_sec, ir->recvTS_usec);
		fprintf(stdout, "Index seqno: 0x%x\n", ir->seqno);
		fprintf(stdout, "Index filePointer: 0x%8.8X\n", ir->filePointer);
		
		fprintf(stdout, "RTP Length (Bytes): %d (0x%X)\n", rtph->length, rtph->length);
		fprintf(stdout, "RTP Unknown(?): %d (0x%X)\n", rtph->unknown, rtph->unknown);

		fprintf(stdout, "RTP Version: %d\n", rtph->version);
		fprintf(stdout, "RTP Padding: %d\n", rtph->padding);
		fprintf(stdout, "RTP Extension: %d\n", rtph->extension);
		fprintf(stdout, "RTP CSRC: %d\n", rtph->csrc);

		fprintf(stdout, "RTP Marker: %d\n", rtph->marker);
		fprintf(stdout, "RTP Payload: %d\n", rtph->payload);
		fprintf(stdout, "RTP Seq Num: 0x%x\n", rtph->seq_num);
		fprintf(stdout, "RTP TS: %u\n", rtph->ts);
		fprintf(stdout, "RTP SSRC: %u\n", rtph->ssrc);
		fprintf(stdout, "---------------------\n");
	//}
}

static void display_index_headers( struct FileHeader *fh, const char* file ) {

	fprintf(stdout, "%s Version: %s\n", file, fh->version);
	fprintf(stdout, "%s Protocol: %s\n", file, fh->protocol);
	fprintf(stdout, "%s Media: %s\n", file, fh->media);
	fprintf(stdout, "%s Cname: %s\n", file, fh->cname);
	fprintf(stdout, "%s Name: %s\n", file, fh->name);
	
	fprintf(stdout, "%s Start: %d.%d\n", file, fh->start_sec, fh->start_usec);
	fprintf(stdout, "%s End: %d.%d\n", file, fh->end_sec, fh->end_usec);

	fprintf(stdout, "%s privateLen: %d\n", file, fh->privateLen);
	fprintf(stdout, "-------------------------\n");

}

static void display_h261_rtp_headers( struct H261RTPHeader* head ) {
	fprintf(stdout, "H261 Start bit pos: %d\n", head->sbit);
	fprintf(stdout, "H261 End bit pos: %d\n", head->ebit);
	fprintf(stdout, "H261 Intra Frame: %d\n", head->i);
	fprintf(stdout, "H261 Motion Vector: %d\n", head->v);
	
	fprintf(stdout, "H261 GOB Num: %d\n", head->gobn);
	fprintf(stdout, "H261 MBAP: %d\n", head->mbap);

	fprintf(stdout, "H261 Quantizer: %d\n", head->quant);

	fprintf(stdout, "H261 HMVD: %d\n", head->hmvd);
	fprintf(stdout, "H261 VMVD: %d\n", head->vmvd);


	fprintf(stdout, "-------------------------\n");

}



static void display_h261_head( unsigned char * buffer ) {
	struct H261Head * head = (struct H261Head *) buffer;
	
	//if (head->psc != 0x00010) {
	
		/* Picture Header Layer */
		fprintf(stdout, "H261 Picture Start Code: %5.5x\n", head->psc);
		fprintf(stdout, "H261 Temporal Reference: %x\n", head->tr);
		fprintf(stdout, "H261 Split Screen: %x\n", head->ss);
		fprintf(stdout, "H261 Document Camera: %x\n", head->dc);
		fprintf(stdout, "H261 Freeze Picture Release: %x\n", head->fpr);
		fprintf(stdout, "H261 Source Format: %x\n", head->sf);
		fprintf(stdout, "H261 Still Image: %x\n", !head->si);
		fprintf(stdout, "H261 Spare: %x\n", head->spare);
		fprintf(stdout, "H261 Extra Insertion Info: %x\n", head->pei);
		
		/* Group of Blocks Layer */
		if (head->pei==0) {
			fprintf(stdout, "H261 GOB Start Code: %4.4x\n", head->gbsc);
			fprintf(stdout, "H261 GOB Group Number: %d\n", head->gn);
			fprintf(stdout, "H261 GEI: %d\n", head->gei);
		}
		
		fprintf(stdout, "-------------------------\n");
	
	//}

}

static int process_entries(oqt_t* out_file, FILE* idx_file, FILE* data_file) {
	int payload = -1;
	int track = NO_TRACK;
	unsigned char* buffer = NULL;
	int buffer_size = 0, buffer_used=0;
	int audio_chunk_bytes = 0, audio_chunk_samples = 0;
	oqt_uint16_t last_seq = -1;

	while(!feof(idx_file)) {
		struct IndexRecord ir;
		struct RTPHeader rtph;
		int data_length;
		
		/* Read index entry from index file */
		fread(&ir, sizeof(ir), 1, idx_file);
		indexrecord_ntoh(&ir);
		
		/* Seek to record in data file and read it in */
		if (fseek(data_file, ir.filePointer, SEEK_SET))
			perror("fseek ir.filePointer");
		if (!fread(&rtph, sizeof(rtph), 1, data_file)) {
			perror("fread rtphead");
		}
		rtpheader_ntoh(&rtph);
		
		
	#ifdef DEBUG
		//display_rtp_headers( &ir, &rtph );
	#endif
	
		/* Check we arn't missing a packet */
		if (rtph.seq_num != last_seq+1 && last_seq!=-1) {
			fprintf(stderr, "Missing packet. Jumped from %d to %d.\n", last_seq, rtph.seq_num);
		}
		
		/* Check the payload */
		if (payload == -1) payload = rtph.payload;
		if (payload != rtph.payload) {
			fprintf(stdout, "Error: payload changed part way through stream.\n");
			return 1;
		}
		
		/* Skip the CSRC list - each entry is 32bits */
		fseek(data_file, rtph.csrc * 4, SEEK_CUR);
		
		/* Skip RTP Extension */
		if (rtph.extension) {
			fprintf(stderr, "**** Unhandled RTP Extension ****\n");
		}
		
		/* Calculate the length of the payload */
		data_length = rtph.length - (ftell(data_file)-ir.filePointer-8);
		
		
		/* Copy data to the movie */
		switch (payload) {
			case 0:
				// PT=0	NAME=PCMU	TYPE=A	HZ=8000	CHANNELS=1
				if (track==NO_TRACK)
				track = oqt_add_audio_track(out_file, 1, 8000, 16, "ulaw");
				audio_chunk_bytes = 11025;
				audio_chunk_samples = 11025;
			break;
			
			case 3:
				// PT=3	NAME=GSM	TYPE=A	HZ=8000	CHANNELS=1
				if (track==NO_TRACK)
				track = oqt_add_audio_track(out_file, 1, 8000, 16, "agsm");
				audio_chunk_bytes = 858;
				audio_chunk_samples = 4160;
			break;
				
			case 8:
				// PT=8	NAME=PCMA	TYPE=A	HZ=8000	CHANNELS=1
				if (track==NO_TRACK)
				track = oqt_add_audio_track(out_file, 1, 8000, 16, "alaw");
				audio_chunk_bytes = 11025;
				audio_chunk_samples = 11025;
			break;
			
			case 10:
				// PT=10	NAME=L16	TYPE=A	HZ=44100	CHANNELS=2
				if (track==NO_TRACK)
				track = oqt_add_audio_track(out_file, 2, 44100, 16, "twos");
				audio_chunk_bytes = 44100;
				audio_chunk_samples = 11025;
			break;
				
			case 11:
				// PT=11	NAME=L16	TYPE=A	HZ=44100	CHANNELS=1
				if (track==NO_TRACK)
				track = oqt_add_audio_track(out_file, 1, 44100, 16, "twos");
				audio_chunk_bytes = 22050;
				audio_chunk_samples = 11025;
			break;

			case 31: {
				// PT=31	NAME=H261	TYPE=V	HZ=90000
				struct H261RTPHeader h261_rtp_head;
				static int ts=0;
				static int last_end_bits;
				
				//fprintf(stderr, "TS Diff: %d\n", rtph.ts-ts);
				fread(&h261_rtp_head, sizeof(h261_rtp_head), 1, data_file);
				
				if (h261_rtp_head.i) {
					static int warned = 0;
					if (!warned) {
						fprintf(stderr, "Warning: This stream doesn't contain key frames.\n");
						warned=1;
					}
				}
				
				/* No buffer yet ? */
				if (!buffer) {
					buffer_size = 90096;
					buffer_used = 0;
					buffer = malloc(buffer_size);
				}
				
				
				
								
				/* Add the data to the buffer */
				data_length-=4;
				
				/* Check there is space in the buffer */
				if (data_length>(buffer_size-buffer_used)) {
					fprintf(stderr, "Error, there isn't room in buffer for the data.\n");
					return 1;
				}
				
				
				/* Merge the first byte with the last of the other ? */
				if (h261_rtp_head.sbit && buffer_used) {
					unsigned char last_byte, first_byte;

					fread(&first_byte, 1, 1, data_file);
					last_byte = buffer[buffer_used-1];
					//buffer[buffer_used] = first_byte;
					//buffer_used++;
					--data_length;
					
					if (last_byte != first_byte)
					fprintf(stderr, "%d: Last ebit=%d (%2.2x), this sbit=%d (%2.2x)\n",
							rtph.seq_num, last_end_bits, last_byte, h261_rtp_head.sbit, first_byte);
				}
				
				fread(buffer+buffer_used, data_length, 1, data_file);
				buffer_used += data_length;
			
				//display_h261_rtp_headers( &h261_rtp_head );
				//display_h261_head( buffer );

				/* Is marker set ? */
				if (rtph.marker) {
					struct H261Head *head = (struct H261Head *)buffer;
					
					/* Check if the frame is valid */			
					if (head->psc == 0x00010 && head->gbsc == 0x0001) {
						if (track==NO_TRACK) {
							if (head->sf==0)
									track = oqt_add_video_track(out_file, 176, 144, 5.0, "h261");	// QCIF
							else 	track = oqt_add_video_track(out_file, 352, 288, 5.0, "h261");	// CIF
							oqt_set_video_depth(out_file, track, 24);
						}
					
						oqt_write_video_data(out_file, track, buffer, buffer_used, 0);
					
						//display_h261_rtp_headers( &h261_rtp_head );
					} else {
						fprintf(stderr, "Frame of video %d looks invalid - discarding (length %d).\n", rtph.seq_num, buffer_used);
					}
					
					
					buffer_used=0;
				}
				
				//
				ts = rtph.ts;
				last_end_bits = h261_rtp_head.ebit;
			} break;

			default:
				fprintf(stdout, "Error: unsupported payload - %d.\n", payload);
				return 1;
			break;
		}
		
		
		
		
		/* Can process audio in a similar way for all payloads */
		if (audio_chunk_bytes) {
			int read_size=0;
		
			/* Is the buffer big enough ? */
			if (buffer_size < audio_chunk_bytes) {
				buffer_size = audio_chunk_bytes;
				buffer = realloc(buffer, buffer_size);
				buffer_used = 0;
			}
		
			/* Do we have more data than will fit in the buffer ? */
			read_size = data_length;
			if (buffer_size-buffer_used<read_size) {
				read_size = buffer_size-buffer_used;
			}
			
			/* Read it in from the data file */
			if (!fread(buffer+buffer_used, read_size, 1, data_file)) {
				perror("Payload read failed");
				return 1;
			}
			buffer_used += read_size;
			
			/* Do we have enough to write a chunk ? */
			if (buffer_used == audio_chunk_bytes) {
				oqt_write_audio_data(out_file, track, buffer, audio_chunk_bytes, audio_chunk_samples, 1, NULL);
				buffer_used = 0;
			}
			
			/* Write the extra data to the buffer */
			if (read_size < data_length) {
				int extra_size = data_length-read_size;
				fread(buffer+buffer_used, extra_size, 1, data_file);
				buffer_used += extra_size;
			}
		}

		last_seq = rtph.seq_num;
	}
	
	
	/* Write any remaining audio data */
	if (audio_chunk_bytes) {
		// This isn't really the best way to do it...
		int chunk_samples = ((double)audio_chunk_samples/audio_chunk_bytes) * buffer_used;
		oqt_write_audio_data(out_file, track, buffer, buffer_used, chunk_samples, 1, NULL);
		buffer_used=0;
	}
	


	// Free the buffer
	if (buffer) {
		free(buffer);
		buffer = NULL;
		buffer_size = 0;
	}
	
	return 0;
}


static int process_stream(oqt_t* out_file, char* id, char* session, char* data_filename, char* index_filename)
{
	struct FileHeader ifh, dfh;
	FILE* idx_file=NULL;
	FILE* data_file=NULL;
	static int added_metadata = 0;
	int result=0;
	
	fprintf(stderr, "Stream ID: %s\n", id);
	fprintf(stderr, "Session: %s\n", session);


	/* Open the index file */
	idx_file = fopen(index_filename, "r");
	if (!idx_file) {
		fprintf(stderr, "Error: failed to open index file: %s\n", index_filename);
		return 1;
	}
	
	/* Read header from index file */
	fread(&ifh, sizeof(ifh), 1, idx_file);
	fileheader_ntoh(&ifh);
	if (strcmp(ifh.version, INDEX_VERSION) != 0) {
		fprintf(stdout, "Unsupported index file version %s.\n", ifh.version);
		fclose(idx_file);
		return 1;
	}
	
	/* Check it is either audio or video */
	if (strcmp(ifh.media, "audio")&&strcmp(ifh.media, "video")) {
		fprintf(stdout, "Unsupported media type: %s.\n", ifh.media);
		fclose(idx_file);
		return 1;
	}
	
	

	/* Open the data file */
	data_file = fopen(data_filename, "r");
	if (!idx_file) {
		fprintf(stderr, "Error: failed to open data file: %s\n", data_filename);
		return 1;
	}

	/* Read header from data file */
	fread(&dfh, sizeof(dfh), 1, data_file);
	fileheader_ntoh(&dfh);
	if (strcmp(dfh.version, DATA_VERSION) != 0) {
		fprintf(stdout, "Unsupported date file version %s.\n", dfh.version);
		fclose(idx_file);
		fclose(data_file);
		return 1;
	}
	
	/* Skip the private headers */
	fseek(data_file, dfh.privateLen, SEEK_CUR);
	fseek(idx_file, ifh.privateLen, SEEK_CUR);
	
	
	/* Add metadata to the movie 
	   ***FIXME: This should be done per track not for the movie
	*/
	if (!added_metadata) {
		time_t time = ifh.start_sec;
		char *time_str = ctime(&time);
	
		/* Set Author */
		if(strlen(ifh.name))
		oqt_set_info_value(out_file, OQT_INFO_AUTHOR, ifh.name);
		
		/* Could do with somewhere better than this */
		if(strlen(ifh.cname))
		oqt_set_info_value(out_file, OQT_INFO_COMMENT, ifh.cname);
	
		/* Set Creation date */
		time_str[strlen(time_str)-1]=0;
		oqt_set_info_value(out_file, OQT_INFO_CREATION_DATE, time_str);
		free( time_str );
		added_metadata = 1;
		
		/* Set Original Source */
		oqt_set_info_value(out_file, OQT_INFO_ORIG_SOURCE, data_filename);
		oqt_set_info_value(out_file, OQT_INFO_ORIG_FORMAT, "Converted from OpenMash archive using " PROGRAM_NAME ".");
	}
	
	#ifdef DEBUG
		display_index_headers(&ifh, "Index File");
		display_index_headers(&dfh, "Data File");
	#endif
	
	
	
	/* Process the entries in the index file */
	result = process_entries( out_file, idx_file, data_file );
	

	// Close the files
	fclose(idx_file);
	fclose(data_file);

	return result;
}

static char* split_cat_entry_pair(char* pair) {
	int i;
	char* value=NULL;
	
	for(i=0; i<strlen(pair); i++) {
		if (pair[i]=='=') {
			pair[i]= 0;
				
			/* Is the equals sign the last char ? */
			if (i+1==strlen(pair)) return NULL;
			else return &pair[i+1];
		}
	}
	
	return NULL;
}


static int read_cat_entries(FILE* cat_file, oqt_t* out_file) {
	char line[255];
	int line_num=0;
	int i, in_stream=0;
	char *value=NULL, *id=NULL, *session=NULL;
	char *datafile=NULL, *indexfile=NULL;
	
	/* Read lines */
	while(fgets(&line[0], sizeof(line), cat_file)) {
		char* ptr= &line[0];
		line_num++;
	
		/* Trim the end of the line */
		for(i=strlen(line)-1;i>=0;i--) {
			if (isspace(line[i])) line[i] = 0;
			else break;
		}
		
		/* Ignore blank lines */
		if (strlen(line)==0) continue;
		
		if (strcmp(ptr, "START_STREAM")==0) {
			if (!in_stream) in_stream=1;
			else {
				fprintf(stderr, "Line %d: Error, got 'START_STREAM' again.\n", line_num);
				return 1;
			}
			continue;
		} else if (!in_stream) {
			fprintf(stderr, "Line %d: Expected 'START_STREAM' and found '%s'.\n", line_num, ptr);
			return 1;
		}
		
		if (strcmp(ptr, "END_STREAM")==0) {
			/* Check we have all the values */
			if (!id) {
				fprintf(stderr, "Line %d: Error, didn't get 'id' for stream definition.\n", line_num);
				return 1;
			}
			if (!session) {
				fprintf(stderr, "Line %d: Error, didn't get 'session' for stream definition.\n", line_num);
				return 1;
			}
			if (!datafile) {
				fprintf(stderr, "Line %d: Error, didn't get 'datafile' for stream definition.\n", line_num);
				return 1;
			}
			if (!indexfile) {
				fprintf(stderr, "Line %d: Error, didn't get 'indexfile' for stream definition.\n", line_num);
				return 1;
			}
			
			/* Process the stream */
			process_stream(out_file, id, session, datafile, indexfile);

			
			/* Free the strings */
			free(id); id=NULL;
			free(session); session=NULL;
			free(datafile); datafile=NULL;
			free(indexfile); indexfile=NULL;
			
			in_stream=0;
			continue;
		}
	
		/* Check for tab at start of line */
		if (*ptr!='\t') {
			fprintf(stderr, "Line %d: Expected tab at start of line inside stream definition.\n", line_num);
			return 1;
		} else {
			ptr++;
		}
		
		/* Sperarte field and value */
		value = split_cat_entry_pair( ptr );
		if (!value) {
			fprintf(stderr, "Line %d: Name/value pair is invalid.\n", line_num);
			return 1;
		}
		
		if (strcmp(ptr, "id")==0) { id = strdup(value); }
		else if (strcmp(ptr, "session")==0) { session = strdup(value); }
		else if (strcmp(ptr, "datafile")==0) { datafile = strdup(value); }
		else if (strcmp(ptr, "indexfile")==0) { indexfile = strdup(value); }
		else {
			fprintf(stderr, "Line %d: Unknown field '%s'.\n", line_num, ptr);
		}
	}
	
	return 0;
}


static void usage(void) 
{ 
	fprintf(stderr,"%s: [options] input_file.ctg output_file.mov\n", PROGRAM_NAME);
	fprintf(stderr,"options:\n");
	fprintf(stderr,"     -h, --help          This message\n");
	exit(1);
}


int main(int argc, char **argv)
{
	oqt_t* out_file=NULL;
	FILE* cat_file=NULL;
	char* in_filename = NULL;
	char* out_filename = NULL;
	char buf[1024];
	char *ptr = &buf[0];
	
	
	/*cat_file = fopen("frame.dat", "r");
	fread(&buf[0], 100, 1, cat_file);
	display_h261_head( &buf[0] );
	return 0;*/
	
	// parse args 
	for (argc--,argv++;argc && *argv;argc--,argv++) {
		if (!strcmp(*argv,"-?") || !strcmp(*argv,"-h") || !strcmp(*argv,"--help") ) usage();
		//else if (!strcmp(*argv,"-i") || !strcmp(*argv,"--inc") )  gIncremental=1;
		//else if (!strcmp(*argv,"-b") || !strcmp(*argv,"--binc") ) gIncremental=2;
		//else if (!strcmp(*argv,"-x") || !strcmp(*argv,"--hex") )  gHex=1;
		//else if (!strcmp(*argv,"-t") || !strcmp(*argv,"--track")) {
		else if (!in_filename) in_filename=*argv;
		else if (!out_filename) out_filename=*argv;
		else {
			fprintf(stderr, "\nUnknown option \"%s\".\n", *argv);
			usage();
		}
	}
	
	if (!in_filename) {
		fprintf(stderr, "Error: missing input file parameter.\n");
		usage();
	}
	
	if (!out_filename) {
		fprintf(stderr, "Error: missing output file parameter.\n");
		usage();
	}

	
	/* Open the input (catalogue) file */
	cat_file = fopen(in_filename, "r");
	if (!cat_file) {
		fprintf(stderr, "Error: failed to open file for reading: %s\n", in_filename);
		exit(1);
	}
	
	/* Open the output (movie) file */
	out_file = oqt_open_file(out_filename, OQT_MODE_WRITE);
	if (!out_file) {
		fprintf(stderr, "Error: failed to open file for writing: %s\n", out_filename);
		exit(1);
	}
	
	
	/* Read the entries from the catalogue file */
	oqt_write_moviedata_init(out_file);
	read_cat_entries(cat_file, out_file);
	oqt_write_moviedata_end(out_file);
	
	
	/* Close files */
	oqt_close(out_file);
	fclose(cat_file);
	

	return 0;
}
