/* oqtplayer.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OQTPlayer, a free video player using the QuickTime 
 * library.
 *
 * OQTPlayer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OQTPlayer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: oqtplayer.c,v 1.27 2003/01/03 20:18:27 antoinemine Exp $
 */


/* OQTPlayer - A very simple video player showing QuickTime video using
 * SDL and OpenQuicktime. Multithreaded.
 *
 * Copyright (c) Antoine Mine' 2001
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include <SDL.h>
#include <SDL_byteorder.h>
#include <SDL_thread.h>
#include <SDL_mutex.h>


#include <openquicktime/openquicktime.h>
#include <openquicktime/structs.h>
#include <openquicktime/colormodels.h>

SDL_Surface *screen = NULL;       /* on-screen video surface */
SDL_PixelFormat *pformat;         /* color format used on screen */

int n_audio, n_video, nb_channel;
int video_width, video_height;
int scaled_width, scaled_height;
int audio_rate, audio_length, audio_bits;
int video_length, chunk_size;
double video_rate;
oqt_t* qtfile;

int debug = 0;

volatile int fullscreen;
int screen_width, screen_height;
int pos_x, pos_y;

volatile int want_to_quit = 0;  /* signal the decompression threads to stop */

SDL_Event end_event; /* this event is send at the end of the video */


#define PLAY_DEBUG(s) if (debug) { s; fflush(stdout); }


/* frame buffer 
 * -----------
 */

#define NFRAMES 40

typedef struct {
  volatile int frame;    /* frame number (or -1 for corrupted) */
  SDL_Surface*  surf;    /* for RGB/RGB playing */
  SDL_Overlay*  yuv;     /* for YUV playing */
} frame;

frame frames[NFRAMES];

volatile int frame_shown = -1;   /* frame beeing shown */
volatile int frame_computed = 0; /* frame beeing computed */
volatile int frame_seek = 0;     /* decompression must be preceeded by seek */

typedef struct
{
  int     BytesPerPixel;  /* 0 for YUV */
  oqt_uint32_t  Rmask;    /* for YUV, put the format instead  */
  oqt_uint32_t  Gmask;
  oqt_uint32_t  Bmask;
  int     OQTmodel;
  char*   name;
} cmodelinfo;

cmodelinfo* cmodel = NULL;


/* audio buffer 
 * ------------
 */

#define NCHUNKS  68

#define SAMPLES 1024

typedef struct {
  volatile int chunk;    /* chunk number (or -1 for corrupted) */
  oqt_int16_t* stream;   /* data */
} chunk;
chunk chunks[NCHUNKS];

volatile int chunk_heard = -1;   /* audio chunk beeing heard */
volatile int chunk_computed = 0; /* audio chunk beeing computed */
volatile int chunk_seek = 0;     /* decompression must be preceeded by seek */


SDL_AudioSpec in;

/* Timing
 * ------
 */

double delay;                            /* normal delay between two frames */
volatile oqt_uint32_t date_frame;        /* when to show the next frame */
oqt_uint32_t last_date, lastlast_date;   /* keep for monitoring fps */

volatile int video_dropping; /* video decompression */
volatile int video_pausing, audio_pausing;  /* used to halt temporarily decompression */

SDL_TimerID timer;

#define NDROP 7

const int drop_pattern[NDROP][NDROP] =
  { {0,0,0,0,0,0,0}, /* 0/7 */
    {1,0,0,0,0,0,0}, /* 1/7 */
    {1,0,0,1,0,0,0}, /* 2/7 */
    {1,0,0,1,0,1,0}, /* 3/7 */
    {1,0,1,1,0,1,0}, /* 4/7 */
    {1,0,1,1,0,1,1}, /* 5/7 */
    {1,1,1,1,0,1,1}  /* 6/7 */
  };

double show_limit =  1.5;    /* up to show_limit frames late are acceptable */
double drop_limit = 15.;     /* up to show_limit frames late => dropping */


/* Instrumentation
 * ---------------
 */
int inst_decomp = 0;
int inst_shown = 0;
int inst_dropped = 0;
int inst_late = 0;
int inst_very_late = 0;
int inst_seek = 0;
oqt_uint32_t inst_begin, inst_end;
volatile oqt_uint32_t inst_cursor; /* when the mouse cursor should disapear */


/* decompression threads 
 * ---------------------
 */

SDL_Thread* video_decompress_thread;   
SDL_Thread* audio_decompress_thread;

SDL_mutex *video_compute_mutex, *audio_compute_mutex;
SDL_cond *video_next_cond, *audio_next_cond;
SDL_cond *video_computed_cond, *audio_computed_cond;


/* Print an error and quit */

static void sdlerror(void)
{
  fprintf(stderr,"SDL error: %s\n",SDL_GetError());
  SDL_Quit();
  exit(1);
}

static void customerror(const char* s)
{
  fprintf(stderr,"Error: %s\n",s);
  SDL_Quit();
  exit(1);
}


/* slider 
 * ------
 */

oqt_uint32_t color_black, color_white;

static void update_slider (int frame)
{
  SDL_Rect rect;
  if (frame<0) frame = 0;
  if (frame>=video_length) frame=video_length-1;

  rect.x = 0;
  rect.w = screen_width;
  rect.y = screen_height-20;
  rect.h = 20;
  SDL_FillRect(screen,&rect,color_black);

  if (fullscreen && SDL_GetTicks()>=inst_cursor) {
    /* mouse cursor and slider not shown */
    SDL_ShowCursor(SDL_DISABLE);
  }
  else {
    /* mouse cursor and slider shown */
    rect.y = screen_height-12;
    rect.h = 4;
    SDL_FillRect(screen,&rect,color_white);  
    
    rect.y = screen_height-20;
    rect.h = 20;
    rect.x = ((double)screen_width*((double)frame)/((double)video_length-1))-2;
    rect.w = 5;
    SDL_FillRect(screen,&rect,color_white); 
    SDL_ShowCursor(SDL_ENABLE);
  }

  SDL_UpdateRect(screen, 0, screen_height-20, screen_width, 20);
}

/* decompression 
 * -------------
 */

static int video_decompress_start (void* arg)
{
  unsigned char** decompress_row;
  int i;
  int dropping;
  int computed;

  decompress_row = (unsigned char**) malloc (video_height*sizeof(char*));
  if (!decompress_row) customerror("malloc error in video_decompress_start.");

  SDL_mutexP (video_compute_mutex);
  for (;;) {
    /* wait till the frame buffer is no longer full */
    PLAY_DEBUG(printf(" wait before computing frame %i in %i\n",frame_computed,frame_computed%NFRAMES));
    while ((frame_computed-frame_shown>NFRAMES-1 || 
	    video_pausing ||
	    frame_computed>=video_length) && !want_to_quit)
      SDL_CondWait(video_next_cond,video_compute_mutex); 

    /* quit */
    if (want_to_quit) break;
 
    if (frame_computed>=video_length) {
      SDL_CondSignal(video_computed_cond);
      continue;
    }

    /* seek */
    if (frame_seek) {
      frame_seek = 0;
      oqt_set_video_position (qtfile, n_video, frame_computed);
    }

    /* decompress the frame */
    dropping =   /* outside the mutex, video_dropping can change! */
      drop_pattern
      [(video_dropping>=NDROP)?(NDROP-1):video_dropping]
      [frame_computed%NDROP];
    if (frame_computed<frame_shown) dropping = 1;
    computed = frame_computed; /* can also change outside mutex */
    frames[computed%NFRAMES].frame = -1;
    SDL_mutexV (video_compute_mutex);
    PLAY_DEBUG(printf(" computing frame %i in %i\n",computed,computed%NFRAMES));
    inst_decomp++;
    if (n_video>=0) {

      /* dropping */
      if (dropping) 
	oqt_decode_video (qtfile, n_video, BC_NONE, NULL);

      else {

	switch (cmodel->OQTmodel) {
	  
        /* YUV */
	case BC_YUV420P:
	  if (SDL_LockYUVOverlay(frames[computed%NFRAMES].yuv)) sdlerror();
	  if (cmodel->Rmask==SDL_YV12_OVERLAY) {
	    /* Y V U */
	    decompress_row[0] = 
	      (unsigned char*)frames[computed%NFRAMES].yuv->pixels[0];
	    decompress_row[2] = 
	      (unsigned char*)frames[computed%NFRAMES].yuv->pixels[1];
	    decompress_row[1] = 
	      (unsigned char*)frames[computed%NFRAMES].yuv->pixels[2];	  
	  } else {
	    /* Y U V */
	    decompress_row[0] = 
	      (unsigned char*)frames[computed%NFRAMES].yuv->pixels[0];
	    decompress_row[1] = 
	      (unsigned char*)frames[computed%NFRAMES].yuv->pixels[1];
	    decompress_row[2] = 
	      (unsigned char*)frames[computed%NFRAMES].yuv->pixels[2];	  
	  }
	  oqt_decode_video (qtfile, n_video, cmodel->OQTmodel, decompress_row);
	  SDL_UnlockYUVOverlay(frames[computed%NFRAMES].yuv);
	  break;
	  
	  /* RGB */
	default:
	  if (SDL_LockSurface(frames[computed%NFRAMES].surf)) sdlerror();
	  for (i=0;i<video_height;i++) {
	    decompress_row[i] = (unsigned char*)frames[computed%NFRAMES].surf->pixels + i*video_width*cmodel->BytesPerPixel; }
	  oqt_decode_video (qtfile, n_video, cmodel->OQTmodel, decompress_row);
	  SDL_UnlockSurface(frames[computed%NFRAMES].surf);
	  
	}
      }
    }
    PLAY_DEBUG(printf(" computed frame  %i\n",computed));
    SDL_mutexP (video_compute_mutex);
    frames[computed%NFRAMES].frame = dropping ? -1 : computed;
    if (!frame_seek) frame_computed++;
   
    /* signal there a new frame available in the buffer */
    SDL_CondSignal(video_computed_cond);
  }
  SDL_mutexV (video_compute_mutex);
  return 0;
}

static int audio_decompress_start (void* arg)
{
  int i;
  int computed, decoded;

  SDL_mutexP (audio_compute_mutex);
  for (;;) {
    PLAY_DEBUG(printf(" wait before computing chunk %i in %i\n",chunk_computed,chunk_computed%NCHUNKS));
    while ((chunk_computed-chunk_heard>NCHUNKS-1 || 
	    chunk_computed*SAMPLES>=audio_length || 
	    audio_pausing) && !want_to_quit)
      /* wait till the audio buffer is no longer full */
      SDL_CondWait(audio_next_cond,audio_compute_mutex);
  
    /* quit */
    if (want_to_quit) break;
    
    if (chunk_computed*SAMPLES>=audio_length) {
      SDL_CondSignal(audio_computed_cond);
      continue;
    }
    
    /* seek */
    if (chunk_seek) {
      oqt_set_audio_position (qtfile, n_audio, chunk_computed*SAMPLES);
      chunk_seek = 0;
    }

    /* decompress a chunk: SAMPLES samples */
    computed = chunk_computed; /* can change outside the mutex */
    chunks[computed%NCHUNKS].chunk=-1;
    SDL_mutexV (audio_compute_mutex);
    PLAY_DEBUG(printf(" computing chunk %i in %i\n",computed,computed%NCHUNKS));

    decoded = oqt_decode_audio(qtfile, n_audio, 
			       (unsigned char*)chunks[computed%NCHUNKS].stream,
			       chunk_size);
    
    PLAY_DEBUG(printf(" decoded %d/%d samples\n",decoded,SAMPLES));
    SDL_mutexP (audio_compute_mutex);
    chunks[computed%NCHUNKS].chunk=computed;
    if (!chunk_seek) chunk_computed++;
   
    /* signal there a new chunk available in the buffer */
    SDL_CondSignal(audio_computed_cond);
  }
  SDL_mutexV (audio_compute_mutex);
  return 0;
}

/* Draw the current frame
 */

static void show_frame(int frame)
{
  if (n_video>=0 && frame>=0 && frame<video_length) {
    SDL_Rect rect;
    SDL_Rect src;
    rect.x = pos_x>=0 ? pos_x : 0;
    rect.y = pos_y>=0 ? pos_y : 0;
    rect.w = scaled_width<=screen_width ? scaled_width : screen_width;
    rect.h = scaled_height<=screen_height ? scaled_height : screen_height;
    src.x = 0;
    src.y = 0;
    src.w = scaled_width<=screen_width ? scaled_width : screen_width;
    src.h = scaled_height<=screen_height ? scaled_height : screen_height;

    switch (cmodel->BytesPerPixel) {

    /* YUV */
    case 0:
      if (SDL_DisplayYUVOverlay(frames[frame%NFRAMES].yuv,&rect)) sdlerror();
      update_slider(frame);
      break;

    /* RGB */      
    default:
      if (SDL_BlitSurface(frames[frame%NFRAMES].surf,&src,screen,&rect)) sdlerror();
      SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
      update_slider(frame);
      break;
      
    }
  } 
  else {
    update_slider(frame);
  }
}


/* Timers 
 * ------
 */

/* Timer asks for a new video frame */

/* if there is audio:
   - date_frame is set by the audio_tick procedure to synchronize,
   - if curtime > date_frame, drop frames (decompression AND showing
not every frame)
   - if curtime >> date_frame, pause decompression (so the next case is
willing to occur)
   - if there is no frame available, seek to the next interframe, set
frame_shown and date_frame accordingly
   - if the frame buffer is near empty, then also drop some frames

  if there is no audio, do as above without ever pausing decompression nor
skipping to a next keyframe
 */
  
static oqt_uint32_t video_tick(oqt_uint32_t id, void* arg)
{
  oqt_uint32_t cur = SDL_GetTicks();
  oqt_uint32_t d = delay;

  PLAY_DEBUG(printf("video tick, date=%i\n",cur));

  SDL_mutexP (video_compute_mutex);

  if (n_audio>=0) SDL_PauseAudio(0);

  if (!date_frame) date_frame = cur;
  video_pausing = 0;

  /* end of video */
  
  if (frame_shown+1>=video_length) {
    SDL_PushEvent(&end_event);
    SDL_CondSignal (video_next_cond);
  }

  /* audio only */

  else if (n_video<0) {
    frame_shown++;
    PLAY_DEBUG(printf(" showing frame %i (audio only)\n",frame_shown));
    show_frame(frame_shown);
    cur = SDL_GetTicks();
    date_frame += delay;
    d = (cur < date_frame) ? (date_frame - cur) : 1;
    if (d>delay*show_limit) d = delay*show_limit;
  }

  /* video, and there's a frame ready for us */

  else if (frame_computed>frame_shown+1) {

    /* too early ! */
    if (cur<date_frame-show_limit*delay);

    else {

      /* show the frame if available and not dropping */
      if (frames[(frame_shown+1)%NFRAMES].frame!=frame_shown+1 ||
	  drop_pattern
	  [(video_dropping>=NDROP)?(NDROP-1):video_dropping]
	  [(frame_shown+1)%NDROP]) {
	inst_dropped++;
	PLAY_DEBUG(printf(" frame %i in %i dropped (contains %i)\n",frame_shown+1,(frame_shown+1)%NFRAMES,frames[(frame_shown+1)%NFRAMES].frame));
      }
      else {
	int frame = frame_shown+1;
	PLAY_DEBUG(printf(" showing frame %i in %i\n",frame,frame%NFRAMES));
	SDL_mutexV (video_compute_mutex);
	show_frame(frame);
	SDL_mutexP (video_compute_mutex);
      }
 
      frame_shown++;
      inst_shown++;
     
      /* we are in time */
      if (cur <= date_frame+show_limit*delay) {
	if (frame_computed > frame_shown+NFRAMES/3 ||
	    frame_computed >= video_length) video_dropping = 0;
	else video_dropping = 1 + ((double)NDROP-1.)*
	       ((double)(frame_shown+NFRAMES/3-frame_computed))/
	       ((double)NFRAMES/3.+1.);
	if (video_dropping>=NDROP) video_dropping = NDROP-1;
	PLAY_DEBUG(printf(" in time: cur=%i date_frame=%i (dropping %i/%i)\n",cur,date_frame,video_dropping,NDROP));
      }
      
      /* we are a little late: just try to drop frames */
      else /*if (cur <= date_frame+drop_limit*delay || n_audio<0) */{ 
	video_dropping = 1 + ((double)NDROP-1.)*
	  ((double)(cur-date_frame-delay*show_limit))/
	  ((double)(delay*drop_limit-delay*show_limit)+1.);
	if (video_dropping>=NDROP) video_dropping = NDROP-1;
	PLAY_DEBUG(printf(" late: cur=%i date_frame=%i (dropping %i/%i)\n",cur,date_frame,video_dropping,NDROP));
	inst_late++;
      }
#if 0      
      /* we are in trouble: pause video decoding */
      else {
	PLAY_DEBUG(printf(" very late: cur=%i date_frame=%i\n",cur,date_frame));
	video_pausing = 1;
	inst_very_late++;
      }
#endif
      /* show fps */
      if (last_date && lastlast_date && last_date!=lastlast_date) {
	double lfps = 1000./((double)last_date-(double)lastlast_date+1.);
	PLAY_DEBUG(printf(" latest=%2.2f fps\n",(double)lfps));
	fflush(stdout);
      } 
      else
	PLAY_DEBUG(printf(" no fps info\n"));
      lastlast_date = last_date;
      last_date = cur;
    } 

    cur = SDL_GetTicks();
    date_frame += delay;
    d = (cur < date_frame) ? (date_frame - cur) : 1;
    if (d>delay*show_limit) d = delay*show_limit;    
  }

  /* buffer empty => seek to next keyframe */
  else if (n_audio>=0) {
    int key = frame_shown+2;
    while (1) {
      key = oqt_get_video_keyframe_after(qtfile, n_video, key+1);
      if (key < 0 || key>=video_length || 
	  date_frame+(key-frame_shown)*delay>=cur+drop_limit*delay) break;
    }
    /* no more keyframe => block at end of video */
    if (key<0 || key>=video_length) key=video_length+1;
    date_frame += (key-frame_shown)*delay;
    d = (date_frame>cur) ? (date_frame-cur) : 1;
    /* date_frame += delay; */
    printf("No frame available at date=%i shown=%i computed=%i skip to interframe=%i date=%i\n",cur,frame_shown,frame_computed,key,date_frame);
    fflush(stdout);
    frame_computed = key-1;
    frame_shown = frame_computed-1;
    frame_seek = 1;
    video_dropping = NDROP/2;
    inst_seek++;
  } else

    /* buffer empty and no audio => no skip to next keyframe */
    {
      frame_shown += drop_limit;
      d = delay*drop_limit;
      date_frame = cur+d;
      video_dropping = NDROP/2;
    }

  SDL_CondSignal (video_next_cond);
  SDL_mutexV (video_compute_mutex);
  return d;
}



/* SDL asks for a new audio frame */

/*
   - sets date_frame so that video is synchronized with audio
   - if no more chunk, pause video/audio showing for a little time
*/

static void audio_tick(void *userdata, oqt_uint8_t *stream, int len)
{
  PLAY_DEBUG(printf("audio tick, date=%i\n",SDL_GetTicks()));

  if (len!=chunk_size) 
    customerror("invalid buffer size in audio_tick.");
  
  SDL_mutexP (audio_compute_mutex);

  /* end of video */
  if ((chunk_heard+1)*SAMPLES>=audio_length) {
    memset (stream,0,len);
  }
  
  /* play chunk */
  else if (chunk_computed>chunk_heard+1) {
    chunk_heard++;
    if (chunks[chunk_heard%NCHUNKS].chunk==chunk_heard) {
      PLAY_DEBUG(printf(" playing chunk %i in %i\n",chunk_heard,chunk_heard%NCHUNKS));
      memcpy(stream,chunks[chunk_heard%NCHUNKS].stream,len);
    }
    else {
      memset (stream,0,len);
      PLAY_DEBUG(printf(" dropping chunk %i in %i (contains %i)\n",chunk_heard,chunk_heard%NCHUNKS,chunks[chunk_heard%NCHUNKS].chunk));
    }
   /* sync */
    SDL_mutexP (video_compute_mutex);
    date_frame = SDL_GetTicks() + 
      (double)(frame_shown+1.)*delay -
      (double)chunk_heard*(double)SAMPLES*1000./(double)audio_rate;
    SDL_mutexV (video_compute_mutex); 
   } 

  /* no more data in buffer */
  else  {
    int date;

    /* play silence */
    memset(stream,0,len);

    /* pause audio/video playback */
    SDL_mutexP (video_compute_mutex);
    if (timer) {
      SDL_CondWait(video_next_cond,video_compute_mutex);
      SDL_RemoveTimer(timer);
    }
    SDL_PauseAudio(1);
    date = SDL_GetTicks() + drop_limit*delay;
    printf("No chunk available heard=%i computed=%i olddate=%i newdate=%i\n",chunk_heard,chunk_computed,date_frame,date); 
    fflush(stdout);
    date_frame = date;
    video_pausing = 0;
    audio_pausing = 0;
    SDL_CondSignal(video_next_cond);
    if (timer) timer = SDL_AddTimer (delay*drop_limit,video_tick,NULL);
    SDL_mutexV (video_compute_mutex);
  }

  SDL_CondSignal (audio_next_cond);
  SDL_mutexV (audio_compute_mutex);
}


/* stream control 
 * --------------
 */


/* stop playing */
static void stop(void)
{
  /* stop video showing */
  SDL_mutexP (video_compute_mutex);
  if (timer) {
    SDL_CondWait(video_next_cond,video_compute_mutex);
    SDL_RemoveTimer(timer);
  }
  timer = 0;
  video_pausing = 1;
  video_dropping = 0;
  SDL_mutexV (video_compute_mutex);
  last_date = lastlast_date = 0;

  /* stop audio playing */
  if (n_audio>=0) {
    SDL_LockAudio();
    SDL_mutexP (audio_compute_mutex);
    audio_pausing = 1;
    SDL_PauseAudio(1);
    SDL_mutexV (audio_compute_mutex);
    SDL_UnlockAudio();
  }
  }

/* buffer & start playing */
/* try hard to start audio & video synchronized!  */
static void play(void)
{
  stop();

  /* start decoding & buffer a little */
  if (n_audio>=0) {
    SDL_mutexP (audio_compute_mutex);
    audio_pausing = 0;
    while (chunk_computed-chunk_heard<NCHUNKS/4+1 && 
	   chunk_computed*SAMPLES<audio_length) {
      SDL_CondSignal(audio_next_cond);
      SDL_CondWait(audio_computed_cond,audio_compute_mutex);
    }
    SDL_mutexV (audio_compute_mutex);
  }
  
  SDL_mutexP (video_compute_mutex);
  video_pausing = 0;
  video_dropping = 0;
  last_date = lastlast_date = 0;
  if (n_video>=0) {
    while (frame_computed-frame_shown<NFRAMES/4+1 && 
	   frame_computed<video_length) {
      SDL_CondSignal(video_next_cond);
      SDL_CondWait(video_computed_cond,video_compute_mutex);
    }
  }
  SDL_mutexV (video_compute_mutex);

  /* schedule showing & start playing audio */
  date_frame = SDL_GetTicks() + delay;
  if (n_audio>=0) {
    double sample = /* sample corresponding to frame shown */
      ((double)frame_shown+1.)*((double)audio_rate)/((double)video_rate);
    double middle = /* sample approximately corresponding to chunk heard */
      ((double)chunk_heard+1.)*SAMPLES;
    double extra_delay = /* frame should be delayed by extra_delay more */
      ((double)sample-(double)middle)*1000./(double)audio_rate;
    if (extra_delay>=0) date_frame += extra_delay;
    /* else frame_shown -= extra_delay*video_rate/1000.; */
    PLAY_DEBUG(printf(" extra_delay = %i\n",(int)extra_delay));
  }
  PLAY_DEBUG(printf("  play at date=%i in %i\n",date_frame,(int)delay));
  timer = SDL_AddTimer(delay,video_tick,NULL); 
  SDL_PauseAudio(0);
}


/* seek to a frame and show it */
/* you need to call play() to restart playing after seek */
/* if exact=0, then seeks to the nearest keyframe before (faster) */
static void seek(int frame,
		 int exact)
{
  stop();

  SDL_mutexP (audio_compute_mutex);
  SDL_mutexP (video_compute_mutex);
  frame_seek = 1;
  if (n_video<0) frame_computed = frame;
  else {
    frame_computed = oqt_get_video_keyframe_before(qtfile, n_video, frame)-1;
    if(frame_computed < 0)
      frame_computed = 0;
  }
  if (!exact) frame = frame_computed+2;
  frame_shown = frame-1;
  update_slider(frame_shown+1);

  /* audio seeking */
  if (n_audio>=0) {
    int sample = 
      ((double)frame)*((double)audio_rate)/((double)video_rate);
    chunk_seek = 1;
    chunk_computed = sample/SAMPLES;
    chunk_heard = chunk_computed-1;
    audio_pausing = 0;
  }

  PLAY_DEBUG(printf(" seeking to frame %i (keyframe=%i) chunk %i (cur=%i)\n",
	       frame, frame_computed, chunk_computed,SDL_GetTicks()));

  /* buffer one frame and show it immediately */
  if (n_video>=0) {
    video_pausing = 0;
    while (frame_computed<frame_shown+2) {
      SDL_CondSignal(video_next_cond);
      SDL_CondWait(video_computed_cond,video_compute_mutex);
    }
  }
  show_frame(frame_shown+1);
  SDL_mutexV (video_compute_mutex);
  SDL_mutexV (audio_compute_mutex);
}


static void quit(void)
{
  int dummy;
  stop();
  SDL_mutexP (audio_compute_mutex);
  SDL_mutexP (video_compute_mutex);
  want_to_quit = 1;
  SDL_CondSignal (video_next_cond);
  SDL_CondSignal (audio_next_cond);    
  SDL_mutexV (video_compute_mutex);
  SDL_mutexV (audio_compute_mutex);
  if (n_video>=0) SDL_WaitThread (video_decompress_thread, &dummy);
  if (n_audio>=0) SDL_WaitThread (audio_decompress_thread, &dummy);
  /* if (n_video>=0) SDL_KillThread (video_decompress_thread);
     if (n_audio>=0) SDL_KillThread (audio_decompress_thread); */
  oqt_close(qtfile);
  inst_end = SDL_GetTicks();
  printf("\n\nTime %i ms\n",inst_end-inst_begin);
  printf("Frame shown %i, decompressed %i\n",inst_shown,inst_decomp);
  if (inst_shown)
    printf("Dropped %i (%i%%), late %i (%i%%), very late %i (%i%%)\n",
	   inst_dropped, inst_dropped*100/inst_shown,
	   inst_late, inst_late*100/inst_shown,
	   inst_very_late, inst_very_late*100/inst_shown);
  printf("Forced seek %i\n",inst_seek);
  printf("End at frame %i\n",frame_shown+1);
  SDL_Quit();
  exit(0);
}


/* resize or toggle fullscreen/windowed mode */
/* beware: width and height do not take into account the 20 pixels in height
   needed for the seek bar 
*/
/* you need to call play() to restart playback after resizing */
/* setting fullscale=1 in YUV, fullscreen mode scales to full-screen, ignoring
   width and height parameters 
*/
static void resize(int width,
		   int height,
		   int full,     /* fullscreen */
		   int fullscale /* scale to whole screen if possible */)
{
  int i;
  stop();
  SDL_mutexP (audio_compute_mutex);
  SDL_mutexP (video_compute_mutex);
  
  fullscreen = full;
  if (n_video<0) {
    fullscreen = 0;
    if (!width) width = video_width;
    scaled_width = width;
    scaled_height = height = video_height;
  }
  else if (cmodel->BytesPerPixel || (fullscale && fullscreen)) {
    scaled_width = video_width;
    scaled_height = video_height;
  }
  else {
    /* keep apect ratio */
    scaled_width = width;
    scaled_height = video_height*scaled_width/video_width;
    if (scaled_height>height) {
      scaled_height = height;
      scaled_width = video_width*scaled_height/video_height;
    }
  }
  
  /* fullscreen mode */
  if (fullscreen) {
    SDL_Rect** r;
    /* find the screen with lower resolution that is bigger than the video */
    r = SDL_ListModes(pformat,SDL_FULLSCREEN);
    screen_width=0;
    screen_height=0;
    for (;*r;r++)
      if ( (*r)->w>=scaled_width && (*r)->h>=scaled_height+20 &&
	   ( ( ((*r)->w<screen_width || (*r)->h<screen_height) ) || 
	     !screen_width ) ) {
	screen_width = (*r)->w;
	screen_height = (*r)->h;
      }

    if (!screen_width) {
      fprintf(stderr,"Unable to find a fullscreen allowing at least %ix%i.\n",scaled_width,scaled_height+20);
      exit(1);
    }
    printf("Using fullscreen mode %ix%i.\n",screen_width,screen_height);
    screen = SDL_SetVideoMode(screen_width, screen_height,
			      cmodel->BytesPerPixel ? cmodel->BytesPerPixel*8 :
			      pformat->BitsPerPixel,
			      SDL_FULLSCREEN);
    if (!screen) {
      fprintf(stderr,"Unable to set video mode: %s.\n",SDL_GetError());
      SDL_Quit();
      exit(1);
    } 

    /* re-scale to cover entire screen */
    if (!cmodel->BytesPerPixel && fullscale) {
      /* keep apect ratio */
      scaled_width = screen_width;
      scaled_height = video_height*scaled_width/video_width;
      if (scaled_height>screen_height-20) {
	scaled_height = screen_height-20;
	scaled_width = video_width*scaled_height/video_height;
      } 
    }
  }

  /* windowed mode */
  else {
    screen_width = width;
    screen_height = height+20;
    screen = SDL_SetVideoMode(screen_width, screen_height,
			      cmodel->BytesPerPixel ? cmodel->BytesPerPixel*8 :
			      pformat->BitsPerPixel, 
			      SDL_RESIZABLE);
    if (!screen) {
      fprintf(stderr,"Unable to set video mode: %s.\n",SDL_GetError());
      SDL_Quit();
      exit(1);
    } 
  }
  
  /* center */
  pos_x = (screen_width-scaled_width)/2;
  pos_y = (screen_height-scaled_height-20)/2;

  color_black = SDL_MapRGB (screen->format,0,0,0);
  color_white = SDL_MapRGB (screen->format,196,196,196);
  
  /* create (or recreate) surfaces for image buffer 
     if in RGB mode and surfaces are already there, we need not create them
     in YUV mode, we need to recreate them, as they are attached to screen
   */
  if (n_video>=0)
    for (i=0;i<NFRAMES;i++){
      
      switch (cmodel->BytesPerPixel) {

	/* YUV */
      case 0:
	{ 
	  SDL_Overlay *old = frames[i].yuv;

	  frames[i].yuv = SDL_CreateYUVOverlay(video_width, video_height, 
					       cmodel->Rmask, screen);
	  if (!frames[i].yuv) sdlerror();
	  if (frames[i].yuv->pitches[0] != video_width ||
	      frames[i].yuv->pitches[1] != video_width/2 ||
	      frames[i].yuv->pitches[2] != video_width/2 )
	    customerror("YUV impossible: pitches not compatible with OpenQuicktime.");
	  if (old) {
	    /* not first time => we need to recreate overlay because they are
	       attached to screen which may have changed
	       however, we copy the old buffer into the new one so we do not have
	       to re-decompress frames already in the buffer 
	    */
	    if (frames[i].frame!=-1) {
	      SDL_LockYUVOverlay(old);
	      SDL_LockYUVOverlay(frames[i].yuv);
	      memcpy(frames[i].yuv->pixels[0],old->pixels[0],video_width*video_height);
	      memcpy(frames[i].yuv->pixels[1],old->pixels[1],(video_width*video_height)/4);
	      memcpy(frames[i].yuv->pixels[2],old->pixels[2],(video_width*video_height)/4);
	      SDL_UnlockYUVOverlay(old);
	      SDL_UnlockYUVOverlay(frames[i].yuv);
	    }
	    SDL_FreeYUVOverlay(old);
	  }
	  else frames[i].frame = -1;
	} 
	break;

	/* RGB */
      default:
	if (!frames[i].surf) {
	  frames[i].surf =
	    SDL_CreateRGBSurface(SDL_SWSURFACE, 
				 video_width, video_height, 
				 cmodel->BytesPerPixel*8, 
				 cmodel->Rmask, cmodel->Gmask, 
				 cmodel->Bmask, 0);
	  if (!frames[i].surf) sdlerror();
	  frames[i].frame = -1;
	}
      }

    }
  
  
  { /* clear screen */
    SDL_Rect rect;
    rect.x = 0;
    rect.y = 0;
    rect.w = screen_width;
    rect.h = screen_height;
    SDL_FillRect(screen, &rect, color_black);
    SDL_UpdateRect(screen, 0, 0, screen_width, screen_height);
  }
  show_frame(frame_shown);
  SDL_mutexV (video_compute_mutex);
  SDL_mutexV (audio_compute_mutex);
}


/* video mode negotiation */
/* ---------------------- */


cmodelinfo cmodels[] = {
#if SDL_BYTEORDER==SDL_LIL_ENDIAN
  { 2, 0x0000001f, 0x000007e0, 0x0000f800, BC_RGB565,  "16-bit RGB" },
  { 2, 0x0000f800, 0x000007e0, 0x0000001f, BC_BGR565,  "16-bit BGR" }, 
  { 3, 0x000000ff, 0x0000ff00, 0x00ff0000, BC_RGB888,  "24-bit RGB" },
  { 3, 0x00ff0000, 0x0000ff00, 0x000000ff, BC_BGR888,  "24-bit BGR" },
  { 4, 0x000000ff, 0x0000ff00, 0x00ff0000, BC_RGBA8888,"32-bit RGBA" },
  { 4, 0x0000ff00, 0x00ff0000, 0xff000000, BC_ARGB8888,"32-bit ARGB" },
  { 4, 0x00ff0000, 0x0000ff00, 0x000000ff, BC_BGRA8888,"32-bit BGRA" },
  { 4, 0xff000000, 0x00ff0000, 0x0000ff00, BC_ABGR8888,"32-bit ABGR" },
#else
  { 2, 0x00001f00, 0x0000e007, 0x000000f8, BC_RGB565,  "16-bit RGB" },
  { 2, 0x000000f8, 0x0000e007, 0x00001f00, BC_BGR565,  "16-bit BGR" }, 
  { 3, 0x00ff0000, 0x0000ff00, 0x000000ff, BC_RGB888,  "24-bit RGB" },
  { 3, 0x000000ff, 0x0000ff00, 0x00ff0000, BC_BGR888,  "24-bit BGR" },
  { 4, 0xff000000, 0x00ff0000, 0x0000ff00, BC_RGBA8888,"32-bit RGBA" },
  { 4, 0x00ff0000, 0x0000ff00, 0x000000ff, BC_ARGB8888,"32-bit ARGB" },
  { 4, 0x0000ff00, 0x00ff0000, 0xff000000, BC_BGRA8888,"32-bit BGRA" },
  { 4, 0x000000ff, 0x0000ff00, 0x00ff0000, BC_ABGR8888,"32-bit ABGR" },
#endif
  { 0, SDL_YV12_OVERLAY, 0, 0, BC_YUV420P, "planar YVU 420 (YV12)" },
  { 0, SDL_IYUV_OVERLAY, 0, 0, BC_YUV420P, "planar YUV 420 (IYUY/I420)" }
};

/* returns the RGB colormodel corresponding to a SDL_PixelFormat,
   or NULL */
static cmodelinfo *rgb_colormodel(SDL_PixelFormat* p)
{
  int j;
  for (j=0;j<sizeof(cmodels)/sizeof(cmodels[0]);j++)
    if (cmodels[j].BytesPerPixel==p->BytesPerPixel &&
	cmodels[j].Rmask==p->Rmask &&
	cmodels[j].Gmask==p->Gmask &&
	cmodels[j].Bmask==p->Bmask) return cmodels+j;
  return NULL;
}

/* returns the YUV planar colormodel corresponding to SDL YUV format 
   or NULL */
static cmodelinfo *yuv_colormodel(int yuv)
{
  int j;
  for (j=0;j<sizeof(cmodels)/sizeof(cmodels[0]);j++)
    if (cmodels[j].Rmask==yuv) return cmodels+j;
  return NULL;
}

/* returns the colormodel corresponding to an OQT colormodel 
   or NULL */
static cmodelinfo* oqt_colormodel(int oqtmodel)
{
  int j;
  for (j=0;j<sizeof(cmodels)/sizeof(cmodels[0]);j++)
    if (cmodels[j].OQTmodel==oqtmodel) return cmodels+j;
  return NULL;
}


/* main */
/* ---- */

static void help(void) 
{ 
  printf("oqtplayer [options] filename\n");
  printf("short options list (see README):\n");
  printf("  --video i      show video track number i (numbered from 0)\n");
  printf("  --audio i      play audio track number i (numbered from 0)\n");
  printf("  --novideo      do not show the video (equivalent to --video -1)\n");
  printf("  --noaudio      do not play audio (equivalent to --audio -1)\n");
  printf("  --loop         repeat the video in a loop\n");
  printf("  --scale f      scale the video by the factor f (only in YUV format)\n");
  printf("  --fullscreen   fullscreen mode (and auto-scale in YUV format)\n");
  printf("  --rgb          force to use the RGB format of the current screen\n");
  printf("  --yuv          force to use a YUV format\n");
  exit(1);
}


int main(int argc, char** argv)
{
  char* filename = NULL;
  int nb_audio, nb_video;
  int i;
  double scale = 0.;
  int yuv = 0, rgb = 0;
  int fullscreen = 0;
  int repeat = 0;
  int playing = 0;

  n_audio = n_video = 0;

  /* parse args */

  for (argc--,argv++;argc && *argv;argc--,argv++) {
    if (!strcmp(*argv,"--novideo")) n_video = -1;
    else if (!strcmp(*argv,"--noaudio")) n_audio = -1;
    else if (!strcmp(*argv,"--video")) { n_video = atoi(argv[1]); argv++; argc--; }
    else if (!strcmp(*argv,"--audio")) { n_audio = atoi(argv[1]); argv++; argc--; }
    else if (!strcmp(*argv,"--rgb")) rgb = 1;
    else if (!strcmp(*argv,"--yuv")) yuv = 1;
    else if (!strcmp(*argv,"--loop")) repeat = 1;
    else if (!strcmp(*argv,"--rgb888")) cmodel = oqt_colormodel(BC_RGB888);
    else if (!strcmp(*argv,"--bgr888")) cmodel = oqt_colormodel(BC_BGR888);
    else if (!strcmp(*argv,"--rgba8888")) cmodel = oqt_colormodel(BC_RGBA8888);
    else if (!strcmp(*argv,"--argb8888")) cmodel = oqt_colormodel(BC_ARGB8888);
    else if (!strcmp(*argv,"--bgra8888")) cmodel = oqt_colormodel(BC_BGRA8888);
    else if (!strcmp(*argv,"--abgr8888")) cmodel = oqt_colormodel(BC_ABGR8888);
    else if (!strcmp(*argv,"--yuv420")) cmodel = yuv_colormodel(SDL_IYUV_OVERLAY);
    else if (!strcmp(*argv,"--yvu420")) cmodel = yuv_colormodel(SDL_YV12_OVERLAY);
    else if (!strcmp(*argv,"--scale")) 
      if (argc>1 && argv[1]) { scale = atof(argv[1]); argv++; argc--; } else help();
    else if (!strcmp(*argv,"--fullscreen")) fullscreen = 1;
    else if (!strcmp(*argv,"--debug")) debug = 1;
    else if (!strcmp(*argv,"-?") || !strcmp(*argv,"-h") || !strcmp(*argv,"--help") || **argv=='-') help();
    else filename=*argv;
  }

  if (!filename) help();


  /* init SDL */

  if (SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO|SDL_INIT_TIMER)<0) {
    fprintf(stderr,"Unable to initialize SDL: %s\n",SDL_GetError());
    exit(1);
  }

  pformat = SDL_GetVideoInfo()->vfmt;


  /* open QT file and get infos */

  qtfile = oqt_open(filename);
  if (!qtfile) { 
    fprintf(stderr, "Unable to open file '%s'\n", filename); 
    exit(1); 
  }

  if(oqt_read_headers(qtfile)) {
  	oqt_close(qtfile);
  	printf("Could not read movie headers in file: %s\n", filename);
    exit(1); 
  }

  nb_audio = oqt_get_audio_track_count(qtfile);
  nb_video = oqt_get_video_track_count(qtfile);
  printf("\n%i video track(s)\n%i audio track(s)\n", nb_video, nb_audio);

  printf("\n");

  if (n_video>=0 && n_video<nb_video &&
      oqt_supported_video(qtfile, n_video)) {
    video_width = oqt_get_video_width(qtfile, n_video);
    video_height = oqt_get_video_height(qtfile, n_video);
    video_rate = oqt_get_video_framerate(qtfile, n_video);
    video_length = oqt_get_video_length(qtfile, n_video);
    printf("Video track %i/%i, %ix%i at %f fps, length %.0f s (%i frames).\n",
	   n_video+1, nb_video, video_width, video_height, 
	   video_rate, video_length/video_rate, video_length);

    if (!cmodel) {

      if (yuv) { /* try YUV */
	cmodel = yuv_colormodel(SDL_YV12_OVERLAY);
      }

      else if (rgb) { /* try RGB that matches SDL preferred pixel format */ 
	cmodel = rgb_colormodel(pformat);
      }

      else /* try codec-preferred */ {
	cmodel = oqt_colormodel(oqt_preferred_colormodel(qtfile, n_video));
	if (cmodel)
	  printf("Using preferred colormodel %s.\n", cmodel->name);
	else {
	  printf("Preferred colormodel not supported, trying standard ones.\n");
	  cmodel = yuv_colormodel(SDL_YV12_OVERLAY);
	  if (!cmodel) cmodel = rgb_colormodel(pformat);
	}
      }      
    }

    if (!cmodel) { /* Last hope... */
      printf("Failing to find a colormodel matching the options, reverting to RGB\n");
      cmodel = oqt_colormodel (BC_RGB888);
    }

    printf("Using colormodel %s, %sscaling supported.\n", 
	   cmodel->name, cmodel->BytesPerPixel ? "no " : "");
    if (!oqt_reads_colormodel(qtfile, n_video, cmodel->OQTmodel))
      printf("Colormodel not supported by the codec: we will use conversion.\n");
    
    if (cmodel->BytesPerPixel && scale) {
      printf("Warning: no scaling supported with this video format (try YUV).\n");
      scale = 0.;
    }
    
  } else {
    printf("No video track supported.\n");
    n_video = -1;
    cmodel = oqt_colormodel (BC_RGB888);
    fullscreen = 0;
    video_width = 260, video_height = 0;
  }

  if (n_audio>=0 && n_audio<nb_audio &&
      oqt_supported_audio(qtfile, n_audio)) {
    audio_rate = oqt_get_audio_samplerate(qtfile, n_audio);
    audio_bits = oqt_get_audio_bits(qtfile, n_audio);
    audio_length = oqt_get_audio_length(qtfile, n_audio);
    nb_channel = oqt_get_audio_channels(qtfile, n_audio);
    printf("Audio track %i/%i at %i sps, %i channel(s), length %.0f s (%i samples).\n",
	   n_audio+1,nb_audio,audio_rate, nb_channel, 
	   (float)audio_length/(float)audio_rate, audio_length);
  }
  else {
    printf("No audio track supported.\n");
    n_audio = -1;
  }

  if (n_audio>=0 && n_video<0) {
    video_rate = 25;
    video_length = (video_rate*audio_length)/audio_rate;
  }

  if (n_audio<0 && n_video<0) {
    printf("No audio or video supported.\n");
    return 0;
  }


  /* init audio & audio buffer */
  
  if (n_audio>=0) {
    in.freq = audio_rate;
    
    if (audio_bits==8)		in.format = AUDIO_S8;
    else if (audio_bits==16)	in.format = AUDIO_S16SYS;
    else {
      fprintf(stderr,"Unsupported sample size: %d(must be 8/16).\n", audio_bits);
      SDL_Quit();
      exit(1);
    }
    
    in.channels = nb_channel;
    in.samples = SAMPLES;
    in.callback = audio_tick;
    if (SDL_OpenAudio (&in,NULL)==-1)  {
      fprintf(stderr,"Unable to open audio.\n");
      SDL_Quit();
      exit(1);
    } 
    
    chunk_size = SAMPLES*(audio_bits/8)*nb_channel;
    for (i=0;i<NCHUNKS;i++){
      chunks[i].stream = (oqt_int16_t*) malloc(chunk_size);
      if (!chunks[i].stream) sdlerror();
      chunks[i].chunk = -1;
    }
    
  }
  
  /* init video buffer
     extra init & video mode setup are done in resize()
   */

  for (i=0;i<NFRAMES;i++){
    frames[i].yuv = NULL;
    frames[i].surf = NULL;
    frames[i].frame = -1;
  }
  
  SDL_WM_SetCaption("OQTPlayer", NULL);
  

  /* setup end event */

  end_event.type = SDL_USEREVENT;


  /* create threads */
  
  video_pausing = 1;
  audio_pausing = 1;
  audio_compute_mutex = SDL_CreateMutex();
  audio_next_cond = SDL_CreateCond();
  audio_computed_cond = SDL_CreateCond();
  video_compute_mutex = SDL_CreateMutex();
  video_next_cond = SDL_CreateCond();
  video_computed_cond = SDL_CreateCond();

  if (n_video>=0) {
    video_decompress_thread = SDL_CreateThread(video_decompress_start, NULL);
    if (!video_decompress_thread) sdlerror();
  }
  if (n_audio>=0) {
    audio_decompress_thread = SDL_CreateThread(audio_decompress_start, NULL);
    if (!audio_decompress_thread) sdlerror();
  }

  inst_begin = SDL_GetTicks();
  inst_cursor = inst_begin;

  /* off we go! */

  delay = 1000.0/((double)video_rate);
  timer = 0;
  playing = 1;
  if (fullscreen)
    if (scale) resize(video_width*scale, video_height*scale, 1, 0);
    else resize(0, 0, 1, 1);
  else 
    if (scale) resize(video_width*scale, video_height*scale, 0, 0);
    else resize(video_width, video_height, 0, 0);
  play();


  /* SDL event loop */

  for (;;) {
    SDL_Event event;
    
    SDL_WaitEvent(&event);
    
    switch (event.type) {

    case SDL_KEYDOWN:
      switch( event.key.keysym.sym ) {
      case SDLK_p:
	playing = 1;
	play();
	break;
	
      case SDLK_s:
	playing = 0;
	stop();
	break;

      case SDLK_SPACE:
	if (playing) { playing = 0; stop(); }
	else { playing = 1; play(); }
	break;

      case SDLK_1:
	resize(video_width,video_height,0,0);
	if (playing) play();
	break;
	
      case SDLK_2:
	resize(video_width*2,video_height*2,0,0);
	if (playing) play();
	break;
	
      case SDLK_3:
	resize(video_width*3,video_height*3,0,0);
	if (playing) play();
	break;
	
      case SDLK_4:
	resize(video_width*4,video_height*4,0,0);
	if (playing) play();
	break;
	
      case SDLK_0:
	resize(0,0,1,1);
	if (playing) play();
	break;	

      case SDLK_HOME:
	seek(0,0);
	if (playing) play();
	break;
	
      case SDLK_ESCAPE:
	quit();
	break;
	
      default:
	/* do nothing */
	break;
      }
      break;
      
    case SDL_MOUSEBUTTONUP:
      if (playing) play();
      break;
      
    case SDL_MOUSEBUTTONDOWN:
      if (event.motion.y>=screen_height-20 &&
	  event.motion.x<=screen_width)
	seek((double)(event.motion.x)*(double)video_length/
	     (double)screen_width,
	     event.button.button==SDL_BUTTON_LEFT);
      break;

    case SDL_MOUSEMOTION:
      inst_cursor = SDL_GetTicks() + 3000;
      if (!timer) update_slider(frame_shown);
      break;
      
    case SDL_USEREVENT:
      if (repeat) { playing = 1; seek(0,1); play(); }
      else { playing = 0; stop(); }
      break;

    case SDL_VIDEORESIZE:
      if (!fullscreen) {
	 resize(event.resize.w,event.resize.h-20,0,0);
	 if (playing) play();
      }
      break;
      

    case SDL_QUIT:
      quit();
      break;
      
    default:
      /* do nothing */
      break;
    }
  }
  
  quit();
  return 0;
}
