/* dechunk.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: dechunk.c,v 1.12 2003/02/13 02:01:37 jhatala Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"

/* RGB uncompressed.  Allows alpha */
#define QUICKTIME_RAW  "raw "


static void usage(void)
{
	printf("usage: dechunk [-f framerate] <input movie> <output prefix>\n");
	printf("	Movies containing rgb frames are written as ppm images.\n");
	exit(1);
}

static int match_32(const char *input, const char *output)
{
	if(input[0] == output[0] &&
		input[1] == output[1] &&
		input[2] == output[2] &&
		input[3] == output[3])
		return 1;
	else 
		return 0;
}


int main(int argc, char *argv[])
{
	oqt_t *file;
	FILE *output;
	int i, j;
	oqt_uint64_t length;
	char string[1024], *prefix = 0, *input = 0;
	char *data = 0;
	int bytes = 0, old_bytes = 0;
	float output_rate = 0;
	float input_rate;
	oqt_uint64_t input_frame;
	oqt_uint64_t new_length;
	int width, height;
	int rgb_to_ppm = 0;

	if(argc < 3)
	{
		usage();
	}

	for(i = 1, j = 0; i < argc; i++)
	{
		if(!strcmp(argv[i], "-f"))
		{
			if(i + 1 < argc)
			{
				output_rate = atof(argv[++i]);
			}
			else
				usage();
		}
		else
		if(j == 0)
		{
			input = argv[i];
			j++;
		}
		else
		if(j == 1)
		{
			prefix = argv[i];
			j++;
		}
	}

	if(!prefix || !input) usage();

	if(!(file = oqt_open_file(input, OQT_MODE_READ)))
	{
		perror("Open failed\n");
		return 1;
	}
	
	
	// Read headers
	if(oqt_read_headers(file)) {
		oqt_close(file);
		printf("Could not read movie headers.\n");
		return 1;
	}
	
	if(!oqt_get_video_track_count(file))
	{
		printf("No video tracks.\n");
		return 1;
	}
	
	if(match_32(oqt_get_video_compressor(file, 0), QUICKTIME_RAW))
	{
		printf("Converting to ppm.\n");
		rgb_to_ppm = 1;
	}

	length = oqt_get_video_length(file, 0);
	input_rate = oqt_get_video_framerate(file, 0);
	if(!output_rate) output_rate = input_rate;
	new_length = output_rate / input_rate * length;
	width = oqt_get_video_width(file, 0);
	height = oqt_get_video_height(file, 0);

	for(i = 0; i < new_length; i++)
	{
/* Get output file */
		sprintf(string, "%s%06d", prefix, i);
		if(!(output = fopen(string, "wb")))
		{
			perror("Open failed");
			return 1;
		}

/* Get input frame */
		input_frame = (oqt_uint64_t)(input_rate / output_rate * i);
		bytes = oqt_get_video_framesize(file, 0, input_frame);

		if(data)
		{
			if(bytes > old_bytes) { free(data); data = 0; }
		}

		if(!data)
		{
			old_bytes = bytes;
			data = malloc(bytes);
		}

		oqt_set_video_position(file, 0, input_frame);
		oqt_read_data(file->data_file, data, bytes);
		if(rgb_to_ppm)
		{
			fprintf(output, "P6\n%d %d\n%d\n", width, height, 0xff);
		}

		if(!fwrite(data, bytes, 1, output))
		{
			perror("write failed");
		}
		fclose(output);
	}

	oqt_close(file);
	return 0;
}
