/* oqtchunkinfo.c
 * Copyright (C) 2002 OpenQuicktime Team & University of Southampton
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * oqtchunkinfo by Nicholas Humfrey <njh@ecs.soton.ac.uk>
 *  I wrote this utility in order to analyse and understand
 *  the size and boundaries of quicktime audio chunks.
 *
 * $Id: oqtchunkinfo.c,v 1.7 2003/03/04 11:15:53 nhumfrey Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"


#define PROGRAM_NAME	"oqtchunkinfo"

static int gIncremental = 0;
static int gHex = 0;


/* In trak.c */
oqt_int64_t oqt_chunk_to_offset(oqt_trak_t *trak, oqt_int32_t chunk);



static void usage(void) 
{ 
	fprintf(stderr,"%s: [options] filename\n", PROGRAM_NAME);
	fprintf(stderr,"options:\n");
	fprintf(stderr,"     -h, --help          This message\n");
	fprintf(stderr,"     -x, --hex           Display values as hex\n");
	fprintf(stderr,"     -i, --inc           Incremental mode\n");
	fprintf(stderr,"     -b, --binc          Block incremental mode\n");
	fprintf(stderr,"     -t, --track <track> Select track to extract\n");
	exit(1);
}


static oqt_int32_t calc_chunk_frames(
	oqt_trak_t *trak,
	oqt_int32_t chunk)
{
	oqt_stsc_table_t *stsc_table = trak->mdia.minf.stbl.stsc.table;
	int stsc_entries = trak->mdia.minf.stbl.stsc.total_entries;
	int i, entry;
	
	for(entry=0,i=1;i<=chunk;i++) {
		if ((entry+1)<stsc_entries && stsc_table[entry+1].chunk==i) entry++;
		
		if (i==chunk) break;
		if (entry>=stsc_entries) {
			fprintf(stderr, "chunk_frames: ran out of stsc entries (%d).\n", entry);
			break;
		}
	}
	
	return stsc_table[entry].samples;
}

static void display_version2_info(oqt_stsd_table_t *stsd_tab)
{
	if (stsd_tab->version>=1) {
		printf("Samples per packet:	%d\n", stsd_tab->samplesPerPacket);
		printf("Bytes per packet:	%d\n", stsd_tab->bytesPerPacket);
		printf("Bytes per frame:	%d\n", stsd_tab->bytesPerFrame);
		printf("Bytes per sample:	%d\n", stsd_tab->bytesPerSample);
		printf("----------------------------------\n");
	}
}

static void display_chunks(oqt_t* file, int track)
{
	oqt_audio_map_t *atrack = &(file->atracks[track]);
	const oqt_codec_info_t* codec;
	oqt_int64_t total_chunks;
	int audio_rate, audio_bits, audio_length;
	int chunk, audio_channels;
	
	int chunk_bytes=0, chunk_frames=0;
	int last_chunk_frames=0;
	oqt_int64_t chunk_offset=0, chunk_length=0;
	
	
	// Get some information about the track
	audio_rate = oqt_get_audio_samplerate(file, track);
	audio_bits = oqt_get_audio_bits(file, track);
	audio_length = oqt_get_audio_length(file, track);
	audio_channels = oqt_get_audio_channels(file, track);
	codec = oqt_get_audio_codec(file, track);
	
	
	// Number of chunks in the track
	total_chunks = oqt_get_audio_chunks(file, track);
	if (total_chunks==0) {
		fprintf(stderr, "%s: error, number of chunks in file is unknown.\n", PROGRAM_NAME);
		return;
	}


	// Display header
	printf("----------------------------------\n");
	printf("Codec:       %s [%.4s]\n", codec->name ,codec->fourcc);
	printf("Sample Rate: %.3f kHz\n", (float)audio_rate/1000);
	printf("Sample Size: %d bit\n", audio_bits);
	printf("Channels:    %d\n", audio_channels);
	printf("Samples:     %d (%.2f secs)\n", audio_length, (float)audio_length/audio_rate);
	printf("----------------------------------\n");
	display_version2_info( &(atrack->track->mdia.minf.stbl.stsd.table[0]) );
	printf("\n");
	
	printf("Chunk   Offset         Length    Frames    Samples   \n");
	printf("===========================================================\n");

	// Process chunk by chunk
	for(chunk=1; chunk<=total_chunks; ++chunk) {
		int this_chunk_frames = calc_chunk_frames(atrack->track, chunk);
		
		if (gIncremental==0 ||
		   (gIncremental==2 && (this_chunk_frames!=last_chunk_frames)))
		{
			chunk_offset=0; chunk_bytes=0;
			chunk_frames=0; chunk_length=0;
		}
		
		if (this_chunk_frames!=last_chunk_frames)
				printf("*");
		else	printf(" ");
		
		chunk_offset += oqt_chunk_to_offset(atrack->track, chunk);
		chunk_bytes += oqt_get_audio_chunksize(file, track, chunk);
		chunk_frames += this_chunk_frames;
		chunk_length += oqt_get_audio_chunk_samples(file, track, chunk);
		last_chunk_frames=this_chunk_frames;
	
		if (gHex) {
			printf("%.3ld:    0x%.8llx    0x%.6x    0x%.6x    0x%.6llx (%.2fs)",
			(long)chunk, chunk_offset, chunk_bytes, chunk_frames, chunk_length,
			 (float)chunk_length/audio_rate);
		} else {
			printf("%.3ld:    0x%.8llx    %6d    %6d    %6lld (%.2fs)",
			(long)chunk, chunk_offset, chunk_bytes, chunk_frames, chunk_length,
			 (float)chunk_length/audio_rate);
		}
		if (chunk_offset==0) 
			printf(" {zero}\n");
		else printf("\n");
		
	}
	
	printf("===========================================================\n");
}


int main(int argc, char **argv)
{
	char* filename=NULL;
	char* program=*argv;
	oqt_t* file=NULL;
	long track=0,num_tracks;

	// parse args 
	for (argc--,argv++;argc && *argv;argc--,argv++) {
		if (!strcmp(*argv,"-?") || !strcmp(*argv,"-h") || !strcmp(*argv,"--help") ) usage();
		else if (!strcmp(*argv,"-i") || !strcmp(*argv,"--inc") )  gIncremental=1;
		else if (!strcmp(*argv,"-b") || !strcmp(*argv,"--binc") ) gIncremental=2;
		else if (!strcmp(*argv,"-x") || !strcmp(*argv,"--hex") )  gHex=1;
		else if (!strcmp(*argv,"-t") || !strcmp(*argv,"--track")) {
			if (sscanf(*(argv+1), "%ld", &track)<1) {
				fprintf(stderr, "%s: Invalid track number %s.\n", PROGRAM_NAME, *(argv+1));
				exit(1);
			}
			--track;
			argc--;argv++;
		}
		else if (*argv[0] != '-') filename=*argv;
		else {
			fprintf(stderr, "\nUnknown option \"%s\".\n", *argv);
			usage();
		}
	}
	
	if (!filename) {
		fprintf(stderr, "Error: missing input file parameter.\n");
		usage();
	}



	// Open the quicktime file
	file = oqt_open(filename);
	if (!file) {
		fprintf(stderr, "Error: failed to open file: %s\n", filename);
		exit(1);
	}
	
	// Display I/O inforamtion
	printf("Stream Type: %s\n", file->stream_type);
	printf("Reference: %s\n", file->stream_reference);

	// Read headers
	if(oqt_read_headers(file)) {
		oqt_close(file);
		printf("Error: Could not read movie headers.\n");
		exit(1);
	}

	// Check for audio tracks
	num_tracks = oqt_get_audio_track_count(file);
	if (num_tracks<1) {
		fprintf(stderr, "Error: No audio tracks available in file.\n");
		exit(1);
	} else if (track<0) {
		fprintf(stderr, "Audio track number should be greater than zero.\n");
	} else if (track>=num_tracks) {
		fprintf(stderr, "Error: selected audio track (%ld) is out of range.", track+1);
		fprintf(stderr, " %ld audio tracks in file.\n", num_tracks);
	}
	

	// Sadly we can only display supported tracks
	if (!oqt_supported_audio(file, track)) {
		char* codec_code = oqt_get_audio_compressor(file, track);
		fprintf(stderr, "Error: Audio format of track %ld is not supported [%.4s].\n", track+1, codec_code);
		exit(1);
	}


	// Display the table
	display_chunks(file, track);

	// Close the file again
	oqt_close(file);

	return 0;
}


