/* qtcodecs.c
 * Copyright (C) 2002 OpenQuicktime Teams, University of Southampton
 * qtcodecs by Nicholas Humfrey <njh@ecs.soton.ac.uk>
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: qtcodecs.c,v 1.14 2002/10/28 13:51:12 nj_humfrey Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"

char long_version = 0;
char display_hex = 0;


static void print_char(const char chr, int *hex) {
	if (*hex) {
		printf("%02X", chr);
	} else if ((chr >= 'A' && chr <= 'Z') || // A-Z 
		(chr >= 'a' && chr <= 'z') || // a-z
		(chr >= '0' && chr <= '9') || // 0-9
		 chr == '-' || chr == '_'  ||
		 chr == '+' || chr == '.'  ||
		 chr == ' ')
	{
		 printf("%c", chr);
	} else {
		 printf(" 0x%02X", chr);
		 *hex=1;
	}
}


static void print_fourcc(const char* fourcc) {
	int hex=0;

	// Are any of the characters 
	print_char(fourcc[0], &hex);
	print_char(fourcc[1], &hex);
	print_char(fourcc[2], &hex);
	print_char(fourcc[3], &hex);
}


static void display_direction( const oqt_codec_info_t* info ) {
	int encodes=0, decodes=0;

	printf("    Direction: ");

	if (info->type == OQT_CODEC_TYPE_VIDEO) {
		oqt_video_codec_t *codec = (oqt_video_codec_t*)info;
		if (codec->decode)	decodes=1;
		if (codec->encode)	encodes=1;
	} else if (info->type == OQT_CODEC_TYPE_AUDIO) {
		oqt_audio_codec_t *codec = (oqt_audio_codec_t*)info;
		if (codec->decode)	decodes=1;
		if (codec->encode)	encodes=1;
	}
	
	if (encodes && decodes)	printf("Both\n");
	else if (encodes) 		printf("Encodes\n");
	else if (decodes) 		printf("Decodes\n");
	else					printf("None\n");
}

static void list_codecs(const oqt_codec_info_t** list, int count) 
{
	const oqt_codec_info_t* info;
	int i;
	
	if (count<1)
		printf("<None Found>\n");
	else {
		for(i=0; i<count; i++) {
			info = list[i];
			if (!info) {
				printf("Error: codec pointer is null.\n");
				continue;
			}
		
			printf("%2.2d) %s [", i+1, info->name);
			print_fourcc(info->fourcc);
			printf("]\n");
			
			if (long_version) {
				const oqt_codec_param_t* params;
				int num_params;
				
				// Does the codec support encoding/decoding ?
				display_direction( info );
			
				// These shouldn't actually ever be null
				// but we won't take the risk...
				if (info->version)
					printf("    Version: %s\n", info->version);
				if (info->description)
					printf("    Description: %s\n", info->description);
				if (info->url)
					printf("    Homepage: %s\n", info->url);
				if (info->copyright)
					printf("    Copyright: %s\n", info->copyright);
				if (info->license)
					printf("    License: %s\n", info->license);
				if (info->module_author)
					printf("    Module Author: %s\n", info->module_author);
				if (info->codec_author)
					printf("    Codec Author: %s\n", info->codec_author);
					
				// Dumping hex ?
				if (display_hex) {
					int j,started=0;
				
					printf("    Hex: 0x%.8x\n", *((int*)info->fourcc) );
					
					if (info->parent) 
					printf("    Parent: 0x%.8x\n", *((int*)info->parent->fourcc) );
					
					// Loop, displaying children
					for(j=0; j<count; j++) {
						const oqt_codec_info_t* child = list[j];
						if (child->parent==info) {
							if (started) printf(", ");
							else printf("    Children: ");
							printf("0x%.8x", *((int*)child->fourcc) );
							started=1;
						}
					}
					if (started) printf("\n");
					
				} else {
					// Was is loaded by a seperate plugin ?
					if (info->parent && info->parent->name)
						printf("    - Loaded by the '%s' plugin.\n", info->parent->name);
					if (info->has_children)
						printf("    - This plugin has child codecs.\n");
				}
					
				// List Parameters
				params = oqt_list_codec_params(info, &num_params);
				
				if (num_params==0 || params==NULL) {
					printf("    - This plugin has no parameters.\n");
				} else {
					int p;
					
					for(p=0; p<num_params; p++) {
						printf("     \"%s\" [", params[p].name);
						switch(params[p].type) {
							case OQT_PARAMETER_STRING: printf("String"); break;
							case OQT_PARAMETER_BOOLEAN: printf("Boolean"); break;
							case OQT_PARAMETER_INTEGER: printf("Integer"); break;
							case OQT_PARAMETER_UNSIGNED_INTEGER: printf("Unsigned"); break;
							case OQT_PARAMETER_DOUBLE: printf("Double"); break;
							default:	printf("Unknown"); break;
						}
						printf("] - %s\n", params[p].desc);
						
						// *** Insert code for displaying the default, max, min value ***
					}
					
				}
			}
			
			// Sanity checks
			if (info->has_children && !info->plugin_handle)
				printf("    Warning: has children but isn't a plugin.\n");
			if (info->parent && info->plugin_handle)
				printf("    Warning: has parent and is a plugin.\n");
		}
	}


}

static void usage(const char* name) {
	printf("usage: %s [options]\n", name);
	printf("     -h        This message\n");
	printf("     -l        Long view\n");
	printf("     -ll       Long view with Hex Codes\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	int i, total_vcodecs, total_acodecs;
	const oqt_codec_info_t** vcodecs;
	const oqt_codec_info_t** acodecs;
	
	for(i=1;i<argc;i++) {
		if (!strcmp(argv[i], "-l"))
			long_version=1;
		else if (!strcmp(argv[i], "-ll")) {
			long_version=1;
			display_hex=1;
		}
		else usage(argv[0]);
	}


	printf("OpenQuicktime v%s.\n", oqt_version());
	
	// How many are there ?
	total_acodecs = oqt_count_audio_codecs();
	total_vcodecs = oqt_count_video_codecs();
	
	// Get pointer to the list
	acodecs = oqt_list_audio_codecs();
	vcodecs = oqt_list_video_codecs();
	
	
	printf("\nAudio Codecs (%d)\n", total_acodecs);
	printf("============\n");
	list_codecs(acodecs, total_acodecs);
	

	printf("\nVideo Codecs (%d)\n", total_vcodecs);
	printf("============\n");
	list_codecs(vcodecs, total_vcodecs);

	printf("\n");
	return 0;
}


