/* qtinfo.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams, and Elliot Lee
 * Based upon Qtinfo by Elliot Lee <sopwith@redhat.com>
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: qtinfo.c,v 1.18 2002/12/15 01:51:59 nj_humfrey Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"


int long_display = 0;

static void file_info(char *filename);


static void usage(char *prog) {
	fprintf(stderr, "Usage: %s [options] filename...\n", prog);
	fprintf(stderr, "       -l  Display longer output\n");
	fprintf(stderr, "       -h  This message\n");
	exit(1);
}



static void
display_file_annotations(oqt_t* file) {
	int i,n = oqt_get_info_count(file);
	oqt_udta_t* info = oqt_get_info_list(file);
	
	if (!long_display) printf("  %d annotations.\n", n);
	for(i=0;i<n;i++) {
		const char* name = oqt_get_info_name(info[i].code);
		if (!long_display) printf("  ");
		
		if (name)	printf("  %s:", name);
		else		printf("  %.4s:", info[i].code);

		if (info[i].value)	printf(" %s\n", info[i].value);
		else printf(" \n");
	}
	
}

static void
file_info(char *filename)
{
	oqt_t* qtfile;
	int i, n;
	
	// Open the file
	qtfile = oqt_open(filename);
	if(!qtfile) {
		fprintf(stderr, "Couldn't open file %s.\n", filename);
		return;
	}
	
	// Display I/O inforamtion
	printf("\nStream Type: %s\n", qtfile->stream_type);
	printf(  "Reference: %s\n", qtfile->stream_reference);
	
	
	// Read headers
	if(oqt_read_headers(qtfile)) {
		oqt_close(qtfile);
		printf("Could not read movie headers.\n");
		return;
	}
	
	
	display_file_annotations(qtfile);
	
	
	// Loop through all the audio tracks
	n = oqt_get_audio_track_count(qtfile);
	if (!long_display) printf("  %d audio tracks.\n", n);
	for(i = 0; i < n; i++) {
		char* codec_code = oqt_get_audio_compressor(qtfile, i);
		oqt_int64_t audio_length = oqt_get_audio_length(qtfile, i);
		int audio_rate = oqt_get_audio_samplerate(qtfile, i);
	
		if (long_display) {
			printf("  Audio Track %d.\n", i);
			printf("    Channels: %d\n", oqt_get_audio_channels(qtfile, i));
			printf("    Bits: %d\n", oqt_get_audio_bits(qtfile, i));
			printf("    Sample Rate: %d\n", audio_rate);
			printf("    Length: %lld\n", audio_length);
			printf("    Duration: %.2f\n", (float)audio_length/audio_rate);
			printf("    Hex: 0x%.8x\n", *((int*)codec_code));
			printf("    Signature: %.4s\n", codec_code);
		} else {
			printf("    %d channels. %d bits. sample rate %d. length %lld. duration %.2f secs.\n",
				oqt_get_audio_channels(qtfile, i),
				oqt_get_audio_bits(qtfile, i),
				audio_rate, audio_length,
				(float)audio_length/audio_rate);
		}
		
		
		// Is the audio codec supported ?
		if (oqt_supported_audio(qtfile, i)) {
			const oqt_codec_info_t* codec_info = oqt_get_audio_codec(qtfile, i);
			if (long_display) {
				printf("    Supported: Yes\n");
				printf("    Codec Name: %s\n", codec_info->name);
				printf("    Codec Version: %s\n", codec_info->version);
			} else {
				printf("    Supported using '%s' codec [%.4s] version %s.\n",
					codec_info->name,
					codec_code,
					codec_info->version);
			}
		} else {
			if (long_display)	printf("    Supported: No\n");
			else printf("    Compressor not supported [%.4s].\n", codec_code);
		}
	}
	
	
	
	// Loop through all the Video tracks
	n = oqt_get_video_track_count(qtfile);
	if (!long_display) printf("  %d video tracks.\n", n);
	for(i = 0; i < n; i++) {
		char* codec_code = oqt_get_video_compressor(qtfile, i);
		float video_rate = oqt_get_video_framerate(qtfile, i);
		oqt_int64_t video_length = oqt_get_video_length(qtfile, i);

		if (long_display) {
			printf("  Video Track %d.\n", i);
			printf("    Width: %d\n", oqt_get_video_width(qtfile, i));
			printf("    Height: %d\n", oqt_get_video_height(qtfile, i));
			printf("    Depth: %d\n", oqt_get_video_depth(qtfile, i));
			printf("    Rate: %f\n", video_rate);
			printf("    Length: %lld\n", video_length);
			printf("    Duration: %.2f\n", video_length/video_rate);
			printf("    Hex: 0x%.8x\n", *((int*)codec_code));
			printf("    Signature: %.4s\n", codec_code);
		} else {
			printf("    %dx%d. depth %d. rate %f. length %lld. duration %.2f secs.\n",
				oqt_get_video_width(qtfile, i),
				oqt_get_video_height(qtfile, i),
				oqt_get_video_depth(qtfile, i),
				video_rate, video_length,
				video_length/video_rate);
		}
		
		
		// Is the video codec supported ?
		if (oqt_supported_video(qtfile, i)) {
			const oqt_codec_info_t* codec_info = oqt_get_video_codec(qtfile, i);
			
			// Is the specific track supported ?
			if (oqt_reads_colormodel(qtfile, i,
			    oqt_preferred_colormodel(qtfile, i))) {
			    if (long_display)	printf("    Supported: Yes\n");
			    else				printf("    Supported");
			} else {
			    if (long_display)	printf("    Supported: No\n");
			    else				printf("    Track unsupported");
			}
			
			if (long_display) {
				printf("    Codec Name: %s\n", codec_info->name);
				printf("    Codec Version: %s\n", codec_info->version);
			} else {
				printf(" using '%s' codec [%.4s] version %s.\n",
					codec_info->name,
					codec_code,
					codec_info->version);
			}
		} else {
			if (long_display)	printf("    Supported: No\n");
			else printf("    Compressor not supported [%.4s].\n", codec_code);
		}
	}

	
	// Close the file
	oqt_close(qtfile);
}



int main(int argc, char *argv[])
{
	int i, got_file=0;

	for(i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			if (!strcmp("-l", argv[i])) {
				long_display = 1;
			} else {
				if (strcmp("-h", argv[i]))
					fprintf(stderr, "%s: Unknown option (%s).\n", argv[0], argv[i]);
				usage(argv[0]);
			}
		} else {
			file_info(argv[i]);
			got_file=1;
		}
	}
	
	if (!got_file) {
		fprintf(stderr, "%s: Not enough parameters.\n", argv[0]);
		usage(argv[0]);
	}

	return 0;
}
